@@ -62,11 +62,11 @@ bool TelEngine::isE164(const char* str)
 }
 
 static unsigned int s_callid = 0;
-static Mutex s_callidMutex;
+static Mutex s_callidMutex(false,"CallID");
 
 // this is to protect against two threads trying to (dis)connect a pair
 //  of call endpoints at the same time
-static Mutex s_mutex(true);
+static Mutex s_mutex(true,"CallEndpoint");
 
 CallEndpoint::CallEndpoint(const char* id)
     : m_peer(0), m_id(id), m_mutex(0)
@@ -82,8 +82,8 @@ void CallEndpoint::destroyed()
 	Debug(DebugAll,"Endpoint at %p type '%s' refcount=%d",e,e->name().c_str(),e->refcount());
     }
 #endif
-    disconnect(true,0,true);
-    m_data.clear();
+    disconnect(true,0,true,0);
+    clearEndpoint();
 }
 
 Mutex& CallEndpoint::commonMutex()
@@ -144,6 +144,7 @@ bool CallEndpoint::connect(CallEndpoint* peer, const char* reason, bool notify)
 
     m_peer = peer;
     peer->setPeer(this,reason,notify);
+    setDisconnect(0);
     connected(reason);
 
 #if 0
@@ -153,14 +154,14 @@ bool CallEndpoint::connect(CallEndpoint* peer, const char* reason, bool notify)
     return true;
 }
 
-bool CallEndpoint::disconnect(bool final, const char* reason, bool notify)
+bool CallEndpoint::disconnect(bool final, const char* reason, bool notify, const NamedList* params)
 {
     if (!m_peer)
 	return false;
     DDebug(DebugAll,"CallEndpoint '%s' disconnecting peer %p from [%p]",m_id.c_str(),m_peer,this);
 
     Lock lock(s_mutex,5000000);
-    if (!lock.mutex()) {
+    if (!lock.locked()) {
 	Debug(DebugFail,"Call disconnect failed - deadlock on call endpoint mutex!");
 	Engine::restart(0);
 	return false;
@@ -178,7 +179,7 @@ bool CallEndpoint::disconnect(bool final, const char* reason, bool notify)
 	e->disconnect();
     }
 
-    temp->setPeer(0,reason,notify);
+    temp->setPeer(0,reason,notify,params);
     if (final)
 	disconnected(true,reason);
     lock.drop();
@@ -186,13 +187,17 @@ bool CallEndpoint::disconnect(bool final, const char* reason, bool notify)
     return deref();
 }
 
-void CallEndpoint::setPeer(CallEndpoint* peer, const char* reason, bool notify)
+void CallEndpoint::setPeer(CallEndpoint* peer, const char* reason, bool notify, const NamedList* params)
 {
     m_peer = peer;
-    if (m_peer)
+    if (m_peer) {
+	setDisconnect(0);
 	connected(reason);
-    else if (notify)
+    }
+    else if (notify) {
+	setDisconnect(params);
 	disconnected(false,reason);
+    }
 }
 
 DataEndpoint* CallEndpoint::getEndpoint(const char* type) const
@@ -239,6 +244,7 @@ void CallEndpoint::clearEndpoint(const char* type)
 	    DataEndpoint* e = static_cast<DataEndpoint*>(l->get());
 	    DDebug(DebugAll,"Endpoint at %p type '%s' peer %p",e,e->name().c_str(),e->getPeer());
 	    e->disconnect();
+	    e->clearCall(this);
 	}
 	m_data.clear();
     }
@@ -247,6 +253,7 @@ void CallEndpoint::clearEndpoint(const char* type)
 	if (dat) {
 	    m_data.remove(dat,false);
 	    dat->disconnect();
+	    dat->clearCall(this);
 	    dat->destruct();
 	}
     }
@@ -278,20 +285,30 @@ DataConsumer* CallEndpoint::getConsumer(const char* type) const
     return dat ? dat->getConsumer() : 0;
 }
 
+bool CallEndpoint::clearData(DataNode* node, const char* type)
+{
+    if (null(type) || !node)
+	return false;
+    Lock mylock(DataEndpoint::commonMutex());
+    RefPointer<DataEndpoint> dat = getEndpoint(type);
+    return dat && dat->clearData(node);
+}
+
+
 Channel::Channel(Driver* driver, const char* id, bool outgoing)
     : CallEndpoint(id),
-      m_driver(driver), m_outgoing(outgoing),
+      m_parameters(""), m_driver(driver), m_outgoing(outgoing),
       m_timeout(0), m_maxcall(0),
-      m_sequence(0), m_answered(false)
+      m_dtmfTime(0), m_dtmfSeq(0), m_answered(false)
 {
     init();
 }
 
 Channel::Channel(Driver& driver, const char* id, bool outgoing)
     : CallEndpoint(id),
-      m_driver(&driver), m_outgoing(outgoing),
+      m_parameters(""), m_driver(&driver), m_outgoing(outgoing),
       m_timeout(0), m_maxcall(0),
-      m_sequence(0), m_answered(false)
+      m_dtmfTime(0), m_dtmfSeq(0), m_answered(false)
 {
     init();
 }
@@ -377,9 +394,11 @@ void Channel::zeroRefs()
 
 void Channel::connected(const char* reason)
 {
+    CallEndpoint::connected(reason);
     Channel* peer = YOBJECT(Channel,getPeer());
     if (peer && peer->billid() && m_billid.null())
 	m_billid = peer->billid();
+    m_lastPeerId = getPeerId();
 }
 
 void Channel::disconnected(bool final, const char* reason)
@@ -387,15 +406,21 @@ void Channel::disconnected(bool final, const char* reason)
     if (final || Engine::exiting())
 	return;
     // last chance to get reconnected to something
-    Message* m = message("chan.disconnected");
+    Message* m = getDisconnect(reason);
     m_targetid.clear();
     // we will remain referenced until the message is destroyed
     m->userData(this);
-    if (reason)
-	m->setParam("reason",reason);
     Engine::enqueue(m);
 }
 
+void Channel::setDisconnect(const NamedList* params)
+{
+    DDebug(this,DebugInfo,"setDisconnect(%p) [%p]",params,this);
+    m_parameters.clearParams();
+    if (params)
+	m_parameters.copyParams(*params);
+}
+
 void Channel::setId(const char* newId)
 {
     debugName(0);
@@ -403,6 +428,16 @@ void Channel::setId(const char* newId)
     debugName(id());
 }
 
+Message* Channel::getDisconnect(const char* reason)
+{
+    Message* msg = new Message("chan.disconnected");
+    msg->copyParams(m_parameters);
+    complete(*msg);
+    if (reason)
+	msg->setParam("reason",reason);
+    return msg;
+}
+
 void Channel::status(const char* newstat)
 {
     m_status = newstat;
@@ -421,6 +456,13 @@ void Channel::setMaxcall(const Message* msg)
 	maxcall(Time::now() + tout*(u_int64_t)1000);
     else
 	maxcall(0);
+    if (msg) {
+	tout = msg->getIntValue("timeout",-1);
+	if (tout > 0)
+	    timeout(Time::now() + tout*(u_int64_t)1000);
+	else if (tout == 0)
+	    timeout(0);
+    }
 }
 
 void Channel::complete(Message& msg, bool minimal) const
@@ -428,11 +470,6 @@ void Channel::complete(Message& msg, bool minimal) const
     msg.setParam("id",id());
     if (m_driver)
 	msg.setParam("module",m_driver->name());
-    if ((msg == "chan.dtmf") && !msg.getParam("sequence")) {
-	// need to add sequence number used to detect reorders
-	Lock lock(mutex());
-	msg.addParam("sequence",String(m_sequence++));
-    }
 
     if (minimal)
 	return;
@@ -447,6 +484,8 @@ void Channel::complete(Message& msg, bool minimal) const
 	msg.setParam("billid",m_billid);
     if (getPeer())
 	msg.setParam("peerid",getPeer()->id());
+    if (m_lastPeerId)
+	msg.setParam("lastpeerid",m_lastPeerId);
     msg.setParam("answered",String::boolText(m_answered));
 }
 
@@ -551,7 +590,7 @@ bool Channel::msgMasquerade(Message& msg)
     if (msg == "call.answered") {
 	Debug(this,DebugInfo,"Masquerading answer operation [%p]",this);
 	m_maxcall = 0;
-	status("answered");
+	m_status = "answered";
     }
     else if (msg == "call.progress") {
 	Debug(this,DebugInfo,"Masquerading progress operation [%p]",this);
@@ -561,6 +600,14 @@ bool Channel::msgMasquerade(Message& msg)
 	Debug(this,DebugInfo,"Masquerading ringing operation [%p]",this);
 	status("ringing");
     }
+    else if (msg == "chan.dtmf") {
+	// add sequence, stop the message if it was a disallowed DTMF duplicate
+	if (dtmfSequence(msg) && m_driver && !m_driver->m_dtmfDups) {
+	    Debug(this,DebugNote,"Stopping duplicate '%s' DTMF '%s' [%p]",
+		msg.getValue("detected"),msg.getValue("text"),this);
+	    return true;
+	}
+    }
     return false;
 }
 
@@ -575,6 +622,18 @@ void Channel::msgStatus(Message& msg)
     msg.retValue() << "name=" << id() << ",type=channel;" << par << "\r\n";
 }
 
+// Control message handler that is invoked only for messages to this channel
+// Find a data endpoint to process it
+bool Channel::msgControl(Message& msg)
+{
+    for (ObjList* o = m_data.skipNull(); o; o = o->skipNext()) {
+	DataEndpoint* dep = static_cast<DataEndpoint*>(o->get());
+	if (dep->control(msg))
+	    return true;
+    }
+    return false;
+}
+
 void Channel::statusParams(String& str)
 {
     if (m_driver)
@@ -677,6 +736,44 @@ void Channel::callRejected(const char* error, const char* reason, const Message*
     status("rejected");
 }
 
+bool Channel::dtmfSequence(Message& msg)
+{
+    if ((msg != "chan.dtmf") || msg.getParam("sequence"))
+	return false;
+    bool duplicate = false;
+    const String* detected = msg.getParam("detected");
+    const String* text = msg.getParam("text");
+    Lock lock(mutex());
+    unsigned int seq = m_dtmfSeq;
+    if (text && detected &&
+	(*text == m_dtmfText) && (*detected != m_dtmfDetected) &&
+	(msg.msgTime() < m_dtmfTime))
+	duplicate = true;
+    else {
+	seq = ++m_dtmfSeq;
+	m_dtmfTime = msg.msgTime() + 4000000;
+	m_dtmfText = text;
+	m_dtmfDetected = detected;
+    }
+    // need to add sequence number used to detect reorders
+    msg.addParam("sequence",String(seq));
+    msg.addParam("duplicate",String::boolText(duplicate));
+    return duplicate;
+}
+
+bool Channel::dtmfEnqueue(Message* msg)
+{
+    if (!msg)
+	return false;
+    if (dtmfSequence(*msg) && m_driver && !m_driver->m_dtmfDups) {
+	Debug(this,DebugNote,"Dropping duplicate '%s' DTMF '%s' [%p]",
+	    msg->getValue("detected"),msg->getValue("text"),this);
+	TelEngine::destruct(msg);
+	return false;
+    }
+    return Engine::enqueue(msg);
+}
+
 bool Channel::dtmfInband(const char* tone)
 {
     if (null(tone))
@@ -750,6 +847,9 @@ TokenDict Module::s_messages[] = {
     { "chan.masquerade", Module::Masquerade },
     { "chan.locate",     Module::Locate },
     { "chan.transfer",   Module::Transfer },
+    { "chan.control",	 Module::Control },
+    { "msg.route",       Module::ImRoute },
+    { "msg.execute",     Module::ImExecute },
     { 0, 0 }
 };
 
@@ -762,8 +862,8 @@ const char* Module::messageName(int id)
     return lookup(id,s_messages);
 }
 
-Module::Module(const char* name, const char* type)
-    : Plugin(name), Mutex(true),
+Module::Module(const char* name, const char* type, bool earlyInit)
+    : Plugin(name,earlyInit), Mutex(true,"Module"),
       m_init(false), m_relays(0), m_name(name), m_type(type), m_changed(0)
 {
     debugName(m_name);
@@ -781,9 +881,9 @@ void* Module::getObject(const String& name) const
     return Plugin::getObject(name);
 }
 
-bool Module::installRelay(const char* name, int id, unsigned priority)
+bool Module::installRelay(int id, const char* name, unsigned priority)
 {
-    if (!(id && name))
+    if (!(id && name && priority))
 	return false;
 
     Lock lock(this);
@@ -791,18 +891,68 @@ bool Module::installRelay(const char* name, int id, unsigned priority)
 	return true;
     m_relays |= id;
 
-    Engine::install(new MessageRelay(name,this,id,priority));
+    MessageRelay* relay = new MessageRelay(name,this,id,priority);
+    m_relayList.append(relay)->setDelete(false);
+    Engine::install(relay);
     return true;
 }
 
 bool Module::installRelay(int id, unsigned priority)
 {
-    return installRelay(messageName(id),id,priority);
+    return installRelay(id,messageName(id),priority);
 }
 
 bool Module::installRelay(const char* name, unsigned priority)
 {
-    return installRelay(name,lookup(name,s_messages),priority);
+    return installRelay(lookup(name,s_messages),name,priority);
+}
+
+bool Module::installRelay(MessageRelay* relay)
+{
+    if (!relay || ((relay->id() & m_relays) != 0) || m_relayList.find(relay))
+	return false;
+    m_relays |= relay->id();
+    m_relayList.append(relay)->setDelete(false);
+    Engine::install(relay);
+    return true;
+}
+
+bool Module::uninstallRelay(MessageRelay* relay, bool delRelay)
+{
+    if (!relay || ((relay->id() & m_relays) == 0) || !m_relayList.remove(relay,false))
+	return false;
+    Engine::uninstall(relay);
+    m_relays &= ~relay->id();
+    if (delRelay)
+	TelEngine::destruct(relay);
+    return true;
+}
+
+bool Module::uninstallRelay(int id, bool delRelay)
+{
+    if ((id & m_relays) == 0)
+	return false;
+    for (ObjList* l = m_relayList.skipNull(); l; l = l->skipNext()) {
+	MessageRelay* r = static_cast<MessageRelay*>(l->get());
+	if (r->id() != id)
+	    continue;
+	Engine::uninstall(r);
+	m_relays &= ~id;
+	if (delRelay)
+	    TelEngine::destruct(r);
+    }
+    return false;
+}
+
+
+bool Module::uninstallRelays()
+{
+    while (MessageRelay* relay = static_cast<MessageRelay*>(m_relayList.remove(false))) {
+	Engine::uninstall(relay);
+	m_relays &= ~relay->id();
+	relay->destruct();
+    }
+    return (0 == m_relays) && (0 == m_relayList.count());
 }
 
 void Module::initialize()
@@ -861,11 +1011,16 @@ bool Module::commandExecute(String& retVal, const String& line)
 
 bool Module::commandComplete(Message& msg, const String& partLine, const String& partWord)
 {
-    if ((partLine == "debug") || (partLine == "status")) {
-	if (partWord.null() || name().startsWith(partWord)) {
-	    msg.retValue().append(name(),"\t");
-	    return false;
-	}
+    if ((partLine == "debug") || (partLine == "status"))
+	itemComplete(msg.retValue(),name(),partWord);
+    return false;
+}
+
+bool Module::itemComplete(String& itemList, const String& item, const String& partWord)
+{
+    if (partWord.null() || item.startsWith(partWord)) {
+	itemList.append(item,"\t");
+	return true;
     }
     return false;
 }
@@ -981,7 +1136,7 @@ Driver::Driver(const char* name, const char* type)
       m_init(false), m_varchan(true),
       m_routing(0), m_routed(0), m_total(0),
       m_nextid(0), m_timeout(0),
-      m_maxroute(0), m_maxchans(0)
+      m_maxroute(0), m_maxchans(0), m_dtmfDups(false)
 {
     m_prefix << name << "/";
 }
@@ -1014,6 +1169,7 @@ void Driver::setup(const char* prefix, bool minimal)
     installRelay(Locate,40);
     installRelay(Drop,60);
     installRelay(Execute,90);
+    installRelay(Control,90);
     if (minimal)
 	return;
     installRelay(Tone);
@@ -1091,18 +1247,21 @@ bool Driver::received(Message &msg, int id)
 	    return exact;
 	}
     }
-    // check if the message was for this driver
-    if (!dest.startsWith(m_prefix))
-	return false;
 
     // handle call.execute which should start a new channel
     if (id == Execute) {
 	if (!canAccept(false))
 	    return false;
-	dest.startSkip(m_prefix,false);
-	return msgExecute(msg,dest);
+	if (dest.startSkip(m_prefix,false) ||
+	    (dest.startSkip("line/",false) && hasLine(msg.getValue("line"))))
+	    return msgExecute(msg,dest);
+	return false;
     }
 
+    // check if the message was for this driver
+    if (!dest.startsWith(m_prefix))
+	return false;
+
     lock();
     RefPointer<Channel> chan = find(dest);
     unlock();
@@ -1116,11 +1275,11 @@ bool Driver::received(Message &msg, int id)
 	    chan->msgStatus(msg);
 	    return true;
 	case Progress:
-	    return chan->isIncoming() && chan->msgProgress(msg);
+	    return chan->isIncoming() && !chan->isAnswered() && chan->msgProgress(msg);
 	case Ringing:
-	    return chan->isIncoming() && chan->msgRinging(msg);
+	    return chan->isIncoming() && !chan->isAnswered() && chan->msgRinging(msg);
 	case Answered:
-	    return chan->isIncoming() && chan->msgAnswered(msg);
+	    return chan->isIncoming() && !chan->isAnswered() && chan->msgAnswered(msg);
 	case Tone:
 	    return chan->msgTone(msg,msg.getValue("text"));
 	case Text:
@@ -1142,6 +1301,8 @@ bool Driver::received(Message &msg, int id)
 	case Locate:
 	    msg.userData(chan);
 	    return true;
+	case Control:
+	    return chan->msgControl(msg);
     }
     return false;
 }
@@ -1186,6 +1347,28 @@ bool Driver::canRoute()
     return true;
 }
 
+bool Driver::hasLine(const String& line) const
+{
+    return false;
+}
+
+bool Driver::msgRoute(Message& msg)
+{
+    String called = msg.getValue("called");
+    if (called.null())
+	return false;
+    String line = msg.getValue("line");
+    if (line.null())
+	line = msg.getValue("account");
+    if (line && hasLine(line)) {
+	// asked to route to a line we have locally
+	msg.setParam("line",line);
+	msg.retValue() = prefix() + called;
+	return true;
+    }
+    return Module::msgRoute(msg);
+}
+
 void Driver::genUpdate(Message& msg)
 {
     msg.addParam("routed",String(m_routed));
@@ -1266,6 +1449,7 @@ void Driver::loadLimits()
     timeout(Engine::config().getIntValue("telephony","timeout"));
     maxRoute(Engine::config().getIntValue("telephony","maxroute"));
     maxChans(Engine::config().getIntValue("telephony","maxchans"));
+    dtmfDups(Engine::config().getBoolValue("telephony","dtmfdups"));
 }
 
 unsigned int Driver::nextid()
