@@ -46,7 +46,7 @@ static const FormatInfo s_formats[] = {
     FormatInfo("ilbc20", 38, 20000),
     FormatInfo("ilbc30", 50, 30000),
 //    FormatInfo("speex", 0),
-//    FormatInfo("g729", 10, 10000),
+    FormatInfo("g729", 10, 10000),
     FormatInfo("plain", 0, 0, "text", 0),
     FormatInfo("raw", 0, 0, "data", 0),
 };
@@ -109,7 +109,7 @@ static TranslatorCaps s_stereoCaps[] = {
     { 0, 0, 0 }
 };
 
-static Mutex s_dataMutex(true);
+static Mutex s_dataMutex(true,"DataEndpoint");
 
 class ThreadedSourcePrivate : public Thread
 {
@@ -128,16 +128,14 @@ protected:
 
     virtual void cleanup()
 	{
-	    RefObject::refMutex().lock();
-	    ThreadedSource* source = m_source;
+	    RefPointer<ThreadedSource> source = m_source;
 	    m_source = 0;
-	    RefObject::refMutex().unlock();
 	    if (source)
 		source->cleanup();
 	}
 
 private:
-    ThreadedSource* m_source;
+    RefPointer<ThreadedSource> m_source;
 };
 
 // slin/alaw/mulaw converter
@@ -146,10 +144,11 @@ class SimpleTranslator : public DataTranslator
 public:
     SimpleTranslator(const DataFormat& sFormat, const DataFormat& dFormat)
 	: DataTranslator(sFormat,dFormat) { }
-    virtual void Consume(const DataBlock& data, unsigned long tStamp)
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 	{
 	    if (!ref())
-		return;
+		return 0;
+	    unsigned long len = 0;
 	    while (getTransSource()) {
 		int nchan = m_format.numChannels();
 		if (nchan != getTransSource()->getFormat().numChannels())
@@ -163,18 +162,19 @@ public:
 		}
 		DataBlock oblock;
 		if (oblock.convert(data, sFmt, dFmt)) {
-		    if (tStamp == (unsigned long)-1) {
+		    if (tStamp == invalidStamp()) {
 			unsigned int delta = data.length();
 			if (delta > oblock.length())
 			    delta = oblock.length();
 			tStamp = m_timestamp + delta;
 		    }
 		    m_timestamp = tStamp;
-		    getTransSource()->Forward(oblock, tStamp);
+		    len = getTransSource()->Forward(oblock, tStamp, flags);
 		}
 		break;
 	    }
 	    deref();
+	    return len;
 	}
 };
 
@@ -188,11 +188,12 @@ public:
 	: DataTranslator(sFormat,dFormat),
 	m_sRate(sFormat.sampleRate()), m_dRate(dFormat.sampleRate())
 	{ }
-    virtual void Consume(const DataBlock& data, unsigned long tStamp)
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 	{
 	    unsigned int n = data.length();
 	    if (!n || (n & 1) || !m_sRate || !m_dRate || !ref())
-		return;
+		return 0;
+	    unsigned long len = 0;
 	    n /= 2;
 	    DataSource* src = getTransSource();
 	    if (src) {
@@ -235,9 +236,10 @@ public:
 		}
 		if (src->timeStamp() != invalidStamp())
 		    delta += src->timeStamp();
-		src->Forward(oblock, delta);
+		len = src->Forward(oblock, delta, flags);
 	    }
 	    deref();
+	    return len;
 	}
 };
 
@@ -251,11 +253,12 @@ public:
 	: DataTranslator(sFormat,dFormat),
 	m_sChans(sFormat.numChannels()), m_dChans(dFormat.numChannels())
 	{ }
-    virtual void Consume(const DataBlock& data, unsigned long tStamp)
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 	{
 	    unsigned int n = data.length();
 	    if (!n || (n & 1) || !ref())
-		return;
+		return 0;
+	    unsigned long len = 0;
 	    n /= 2;
 	    if (getTransSource()) {
 		short* s = (short*) data.data();
@@ -286,9 +289,10 @@ public:
 			*d++ = v;
 		    }
 		}
-		getTransSource()->Forward(oblock, tStamp);
+		len = getTransSource()->Forward(oblock, tStamp, flags);
 	    }
 	    deref();
+	    return len;
 	}
 };
 
@@ -453,18 +457,19 @@ void* DataConsumer::getObject(const String& name) const
     return DataNode::getObject(name);
 }
 
-void DataConsumer::Consume(const DataBlock& data, unsigned long tStamp, DataSource* source)
+unsigned long DataConsumer::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags, DataSource* source)
 {
     if (source == m_override)
 	tStamp += m_overrideTsDelta;
     else if (m_override || (source != m_source))
-	return;
+	return 0;
     else
 	tStamp += m_regularTsDelta;
     u_int64_t tsTime = Time::now();
-    Consume(data,tStamp);
+    unsigned long len = Consume(data,tStamp,flags);
     m_timestamp = tStamp;
     m_lastTsTime = tsTime;
+    return len;
 }
 
 bool DataConsumer::synchronize(DataSource* source)
@@ -504,13 +509,27 @@ bool DataConsumer::synchronize(DataSource* source)
 }
 
 
-void DataSource::Forward(const DataBlock& data, unsigned long tStamp)
+bool DataSource::valid() const
 {
-    Lock lock(m_mutex,100000);
+    Lock mylock(const_cast<DataSource*>(this));
+    if (!m_translator)	
+	return true;
+    // this is a translator's source - check if we have at least one valid consumer
+    for (ObjList* l = m_consumers.skipNull(); l; l=l->skipNext()) {
+	DataConsumer* c = static_cast<DataConsumer *>(l->get());
+	if (c->valid())
+	    return true;
+    }
+    return false;
+}
+
+unsigned long DataSource::Forward(const DataBlock& data, unsigned long tStamp, unsigned long flags)
+{
+    Lock mylock(this,100000);
     // we DON'T refcount here, we rely on the mutex to keep us safe
-    if (!(lock.mutex() && alive())) {
+    if (!(mylock.locked() && alive())) {
 	DDebug(DebugInfo,"Forwarding on a dead DataSource! [%p]",this);
-	return;
+	return 0;
     }
 
     // try to evaluate amount of samples in this packet
@@ -526,13 +545,31 @@ void DataSource::Forward(const DataBlock& data, unsigned long tStamp)
 	    m_timestamp,nSamp,this);
 	tStamp = m_timestamp + nSamp;
     }
-    ObjList *l = m_consumers.skipNull();
-    for (; l; l=l->skipNext()) {
-	DataConsumer *c = static_cast<DataConsumer *>(l->get());
-	c->Consume(data,tStamp,this);
+    unsigned long len = invalidStamp();
+    bool empty = true;
+    ObjList* l = m_consumers.skipNull();
+    while (l) {
+	DataConsumer* c = static_cast<DataConsumer *>(l->get());
+	unsigned long ll = c->Consume(data,tStamp,flags,this);
+	if (ll || c->valid()) {
+	    // get the minimum data amount forwarded to all consumers
+	    if (len > ll)
+		len = ll;
+	    l = l->skipNext();
+	    empty = false;
+	}
+	else {
+	    DDebug(DebugInfo,"Consumer %p becomes invalid [%p]",c,this);
+	    detachInternal(c);
+	    // do not advance in list, we just removed the current element
+	    l = l->skipNull();
+	}
     }
+    if (empty)
+	len = 0;
     m_timestamp = tStamp;
     m_nextStamp = nSamp ? (tStamp + nSamp) : invalidStamp();
+    return len;
 }
 
 bool DataSource::attach(DataConsumer* consumer, bool override)
@@ -545,16 +582,12 @@ bool DataSource::attach(DataConsumer* consumer, bool override)
 	this,(override ? " as override" : ""),consumer);
     if (!(consumer && consumer->ref()))
 	return false;
-    Lock lock(m_mutex);
-    if (override) {
-	if (consumer->m_override)
-	    consumer->m_override->detach(consumer);
-	consumer->m_override = this;
-    }
-    else {
-	if (consumer->m_source)
-	    consumer->m_source->detach(consumer);
-	consumer->m_source = this;
+    Lock mylock(this);
+    DataSource*& src = override ? consumer->m_override : consumer->m_source;
+    if (src != this) {
+	if (src)
+	    src->detach(consumer);
+	src = this;
     }
     consumer->synchronize(this);
     m_consumers.append(consumer);
@@ -571,9 +604,9 @@ bool DataSource::detach(DataConsumer* consumer)
     }
     DDebug(DebugAll,"DataSource [%p] detaching consumer [%p]",this,consumer);
     // lock the source to prevent races with the Forward method
-    m_mutex.lock();
+    lock();
     bool ok = detachInternal(consumer);
-    m_mutex.unlock();
+    unlock();
     deref();
     return ok;
 }
@@ -604,16 +637,16 @@ void DataSource::destroyed()
 void DataSource::clear()
 {
     // keep the source locked to prevent races with the Forward method
-    m_mutex.lock();
+    lock();
     while (detachInternal(static_cast<DataConsumer*>(m_consumers.get())))
 	;
-    m_mutex.unlock();
+    unlock();
 }
 
 void DataSource::synchronize(unsigned long tStamp)
 {
-    Lock lock(m_mutex,100000);
-    if (!(lock.mutex() && alive())) {
+    Lock mylock(this,100000);
+    if (!(mylock.locked() && alive())) {
 	DDebug(DebugInfo,"Synchronizing on a dead DataSource! [%p]",this);
 	return;
     }
@@ -750,6 +783,7 @@ bool DataEndpoint::disconnect()
     DataEndpoint* temp = m_peer;
     m_peer = 0;
     temp->m_peer = 0;
+    lock.drop();
     temp->deref();
     return deref();
 }
@@ -790,7 +824,7 @@ void DataEndpoint::setSource(DataSource* source)
 	ObjList* l = m_sniffers.skipNull();
 	for (; l; l = l->skipNext())
 	    DataTranslator::detachChain(temp,static_cast<DataConsumer*>(l->get()));
-	temp->deref();
+	temp->attached(false);
     }
     if (source) {
 	source->ref();
@@ -803,14 +837,15 @@ void DataEndpoint::setSource(DataSource* source)
 	ObjList* l = m_sniffers.skipNull();
 	for (; l; l = l->skipNext())
 	    DataTranslator::attachChain(source,static_cast<DataConsumer*>(l->get()));
+	source->attached(true);
     }
     m_source = source;
-    if (c1)
-	c1->deref();
-    if (c2)
-	c2->deref();
     if (m_callRecord)
 	m_callRecord->deref();
+    lock.drop();
+    TelEngine::destruct(temp);
+    TelEngine::destruct(c1);
+    TelEngine::destruct(c2);
 }
 
 void DataEndpoint::setConsumer(DataConsumer* consumer)
@@ -831,11 +866,14 @@ void DataEndpoint::setConsumer(DataConsumer* consumer)
 	    consumer = 0;
     }
     m_consumer = consumer;
-    if (temp) {
-	if (source)
-	    DataTranslator::detachChain(source,temp);
-	temp->deref();
-    }
+    if (source && temp)
+	DataTranslator::detachChain(source,temp);
+    if (temp)
+	temp->attached(false);
+    if (consumer)
+	consumer->attached(true);
+    lock.drop();
+    TelEngine::destruct(temp);
 }
 
 void DataEndpoint::setPeerRecord(DataConsumer* consumer)
@@ -856,11 +894,14 @@ void DataEndpoint::setPeerRecord(DataConsumer* consumer)
 	    consumer = 0;
     }
     m_peerRecord = consumer;
-    if (temp) {
-	if (source)
-	    DataTranslator::detachChain(source,temp);
-	temp->deref();
-    }
+    if (source && temp)
+	DataTranslator::detachChain(source,temp);
+    if (temp)
+	temp->attached(false);
+    if (consumer)
+	consumer->attached(true);
+    lock.drop();
+    TelEngine::destruct(temp);
 }
 
 void DataEndpoint::setCallRecord(DataConsumer* consumer)
@@ -880,11 +921,14 @@ void DataEndpoint::setCallRecord(DataConsumer* consumer)
 	    consumer = 0;
     }
     m_callRecord = consumer;
-    if (temp) {
-	if (m_source)
-	    DataTranslator::detachChain(m_source,temp);
-	temp->deref();
-    }
+    if (temp && m_source)
+	DataTranslator::detachChain(m_source,temp);
+    if (temp)
+	temp->attached(false);
+    if (consumer)
+	consumer->attached(true);
+    lock.drop();
+    TelEngine::destruct(temp);
 }
 
 bool DataEndpoint::addSniffer(DataConsumer* sniffer)
@@ -901,6 +945,7 @@ bool DataEndpoint::addSniffer(DataConsumer* sniffer)
     m_sniffers.append(sniffer);
     if (m_source)
 	DataTranslator::attachChain(m_source,sniffer);
+    sniffer->attached(true);
     return true;
 }
 
@@ -915,6 +960,7 @@ bool DataEndpoint::delSniffer(DataConsumer* sniffer)
 	return false;
     if (m_source)
 	DataTranslator::detachChain(m_source,sniffer);
+    sniffer->attached(false);
     sniffer->deref();
     return true;
 }
@@ -930,24 +976,56 @@ void DataEndpoint::clearSniffers()
 	    sniffer,m_source,this);
 	if (m_source)
 	    DataTranslator::detachChain(m_source,sniffer);
+	sniffer->attached(false);
 	sniffer->deref();
     }
 }
 
+bool DataEndpoint::clearData(DataNode* node)
+{
+    if (!node)
+	return false;
+    Lock lock(s_dataMutex);
+    bool ok = delSniffer(static_cast<DataConsumer*>(node));
+    if (m_callRecord == node) {
+	setCallRecord();
+	ok = true;
+    }
+    if (m_peerRecord == node) {
+	setPeerRecord();
+	ok = true;
+    }
+    if (m_consumer == node) {
+	setConsumer();
+	ok = true;
+    }
+    if (m_source == node) {
+	setSource();
+	ok = true;
+    }
+    return ok;
+}
+
+// Change source(s) or consumer(s)
+bool DataEndpoint::control(NamedList& params)
+{
+    return (m_source && m_source->control(params)) ||
+	(m_consumer && m_consumer->control(params)) ||
+	(m_peerRecord && m_peerRecord->control(params)) ||
+	(m_callRecord && m_callRecord->control(params));
+}
+
 
 void ThreadedSource::destroyed()
 {
-    if (m_asyncDelete && m_thread)
-	Debug(DebugFail,"ThreadedSource destroyed holding thread %p [%p]",m_thread,this);
-    m_asyncDelete = false;
     if (m_thread)
-	stop();
+	Debug(DebugFail,"ThreadedSource destroyed holding thread %p [%p]",m_thread,this);
     DataSource::destroyed();
 }
 
 bool ThreadedSource::start(const char* name, Thread::Priority prio)
 {
-    Lock lock(mutex());
+    Lock mylock(this);
     if (!m_thread) {
 	ThreadedSourcePrivate* thread = new ThreadedSourcePrivate(this,name,prio);
 	if (thread->startup()) {
@@ -962,43 +1040,21 @@ bool ThreadedSource::start(const char* name, Thread::Priority prio)
 
 void ThreadedSource::stop()
 {
-    Lock lock(mutex());
-    if (!m_thread)
-	return;
-    RefObject::refMutex().lock();
+    Lock mylock(this);
     ThreadedSourcePrivate* tmp = m_thread;
     m_thread = 0;
-    if (tmp) {
-	if (tmp->m_source == this)
-	    tmp->m_source = 0;
-	else
-	    tmp = 0;
-    }
-    RefObject::refMutex().unlock();
-    if (tmp)
-	delete tmp;
+    if (!tmp || tmp->running())
+	return;
+    Debug(DebugInfo,"ThreadedSource deleting stopped thread %p [%p]",tmp,this);
+    mylock.drop();
+    delete tmp;
 }
 
 void ThreadedSource::cleanup()
 {
-    Lock lock(RefObject::refMutex());
+    lock();
     m_thread = 0;
-    if (m_asyncDelete && !alive()) {
-	lock.drop();
-	zeroRefs();
-    }
-}
-
-bool ThreadedSource::zeroRefsTest()
-{
-    // let the data thread destroy us if possible
-    if (m_asyncDelete && m_thread && m_thread->running()) {
-	m_thread = 0;
-	return false;
-    }
-    // if async not possible make sure we are set up for synchronous destruction
-    m_asyncDelete = false;
-    return DataSource::zeroRefsTest();
+    unlock();
 }
 
 Thread* ThreadedSource::thread() const
@@ -1008,10 +1064,23 @@ Thread* ThreadedSource::thread() const
 
 bool ThreadedSource::running() const
 {
-    Lock lock(RefObject::refMutex());
+    Lock mylock(const_cast<ThreadedSource*>(this));
     return m_thread && m_thread->running();
 }
 
+bool ThreadedSource::looping(bool runConsumers) const
+{
+    Lock mylock(const_cast<ThreadedSource*>(this));
+    if (runConsumers) {
+	if (!(alive() && m_consumers.count()))
+	    return false;
+    }
+    else if (refcount() <= 1)
+	return false;
+    return m_thread && !m_thread->check(false) &&
+	m_thread->isCurrent() && !Engine::exiting();
+}
+
 
 DataTranslator::DataTranslator(const char* sFormat, const char* dFormat)
     : DataConsumer(sFormat)
@@ -1073,7 +1142,7 @@ bool DataTranslator::synchronize(DataSource* source)
     return true;
 }
 
-Mutex DataTranslator::s_mutex(true);
+Mutex DataTranslator::s_mutex(true,"DataTranslator");
 ObjList DataTranslator::s_factories;
 unsigned int DataTranslator::s_maxChain = 3;
 static ObjList s_compose;
@@ -1421,9 +1490,16 @@ DataTranslator* DataTranslator::create(const DataFormat& sFormat, const DataForm
 bool DataTranslator::attachChain(DataSource* source, DataConsumer* consumer, bool override)
 {
     XDebug(DebugInfo,"DataTranslator::attachChain [%p] '%s' -> [%p] '%s'",
-	source,source->getFormat().c_str(),consumer,consumer->getFormat().c_str());
-    if (!source || !consumer || !source->getFormat() || !consumer->getFormat())
+	source,(source ? source->getFormat().c_str() : ""),
+	consumer,(consumer ? consumer->getFormat().c_str() : ""));
+    if (!source || !consumer)
 	return false;
+    if (!source->getFormat() || !consumer->getFormat()) {
+	Debug(DebugInfo,"DataTranslator::attachChain [%p] '%s' -> [%p] '%s' not possible",
+	source,source->getFormat().c_str(),
+	consumer,consumer->getFormat().c_str());
+	return false;
+    }
 
     bool retv = false;
     // first attempt to connect directly, changing format if possible
