@@ -30,12 +30,15 @@
 #define RTLD_NOW 0
 #define dlopen(name,flags) LoadLibrary(name)
 #define dlclose !FreeLibrary
+#define dlsym GetProcAddress
 #define dlerror() "LoadLibrary error"
 #define YSERV_RUN 1
 #define YSERV_INS 2
 #define YSERV_DEL 4
 #define PATH_SEP "\\"
+#ifndef CFG_DIR
 #define CFG_DIR "Yate"
+#endif
 
 #ifndef SHGetSpecialFolderPath
 __declspec(dllimport) BOOL WINAPI SHGetSpecialFolderPathA(HWND,LPSTR,INT,BOOL);
@@ -50,7 +53,9 @@ __declspec(dllimport) BOOL WINAPI SHGetSpecialFolderPathA(HWND,LPSTR,INT,BOOL);
 #include <sys/resource.h>
 typedef void* HMODULE;
 #define PATH_SEP "/"
+#ifndef CFG_DIR
 #define CFG_DIR ".yate"
+#endif
 
 #endif
 
@@ -65,6 +70,10 @@ typedef void* HMODULE;
 #include <sys/stat.h>
 #include <fcntl.h>
 
+#ifdef HAVE_PRCTL
+#include <sys/prctl.h>
+#endif
+
 #include <assert.h>
 
 namespace TelEngine {
@@ -73,7 +82,7 @@ class EnginePrivate : public Thread
 {
 public:
     EnginePrivate()
-	: Thread("EnginePrivate")
+	: Thread("Engine Worker")
 	{ count++; }
     ~EnginePrivate()
 	{ count--; }
@@ -81,6 +90,14 @@ public:
     static int count;
 };
 
+class EngineCommand : public MessageHandler
+{
+public:
+    EngineCommand() : MessageHandler("engine.command") { }
+    virtual bool received(Message &msg);
+    static void doCompletion(Message &msg, const String& partLine, const String& partWord);
+};
+
 };
 
 using namespace TelEngine;
@@ -88,16 +105,35 @@ using namespace TelEngine;
 #ifndef MOD_PATH
 #define MOD_PATH "." PATH_SEP "modules"
 #endif
+#ifndef SHR_PATH
+#define SHR_PATH "." PATH_SEP "share"
+#endif
 #ifndef CFG_PATH
 #define CFG_PATH "." PATH_SEP "conf.d"
 #endif
 #define DLL_SUFFIX ".yate"
 #define CFG_SUFFIX ".conf"
 
+
+// Supervisor control constants
+
+// Maximum the child's sanity pool can grow
 #define MAX_SANITY 5
+// Initial sanity buffer, allow some init time for the child
 #define INIT_SANITY 30
+// Minimum (and initial) delay until supervisor restarts child
+#define RUNDELAY_MIN 1000000
+// Maximum delay until supervisor restarts child, allow system to breathe
+#define RUNDELAY_MAX 60000000
+// Amount we decerease delay towards minimum each time child proves sanity
+#define RUNDELAY_DEC 20000
+// Size of log relay buffer in bytes
 #define MAX_LOGBUFF 4096
 
+#ifndef HOST_NAME_MAX
+#define HOST_NAME_MAX 255
+#endif
+
 static u_int64_t s_nextinit = 0;
 static u_int64_t s_restarts = 0;
 static bool s_makeworker = true;
@@ -105,6 +141,7 @@ static bool s_keepclosing = false;
 static bool s_nounload = false;
 static int s_super_handle = -1;
 static bool s_localsymbol = false;
+static bool s_logtruncate = false;
 
 static void sighandler(int signal)
 {
@@ -133,40 +170,66 @@ static void sighandler(int signal)
     }
 }
 
-String Engine::s_cfgpath(CFG_PATH);
+String Engine::s_node;
+String Engine::s_shrpath(SHR_PATH);
 String Engine::s_cfgsuffix(CFG_SUFFIX);
 String Engine::s_modpath(MOD_PATH);
 String Engine::s_modsuffix(DLL_SUFFIX);
 ObjList Engine::s_extramod;
+NamedList Engine::s_params("");
 
 Engine::RunMode Engine::s_mode = Engine::Stopped;
 Engine* Engine::s_self = 0;
 int Engine::s_haltcode = -1;
 int EnginePrivate::count = 0;
+static String s_cfgpath(CFG_PATH);
+static String s_usrpath;
+static bool s_createusr = true;
 static bool s_init = false;
 static bool s_dynplugin = false;
+static Engine::PluginMode s_loadMode = Engine::LoadFail;
 static int s_maxworkers = 10;
 static bool s_debug = true;
 
+#ifdef RLIMIT_CORE
 static bool s_coredump = false;
+#endif
+
+#ifdef RLIMIT_NOFILE
+#ifdef FDSIZE_HACK
+static bool s_numfiles = false;
+#else
+#undef RLIMIT_NOFILE
+#endif
+#endif
+
 static bool s_sigabrt = false;
 static bool s_lateabrt = false;
 static String s_cfgfile;
+static String s_userdir(CFG_DIR);
 static const char* s_logfile = 0;
 static Configuration s_cfg;
 static ObjList plugins;
 static ObjList* s_cmds = 0;
 static unsigned int s_runid = 0;
 
-class SLib : public GenObject
+static EngineCheck* s_engineCheck = 0;
+
+void EngineCheck::setChecker(EngineCheck* ptr)
+{
+    s_engineCheck = ptr;
+}
+
+class SLib : public String
 {
 public:
     virtual ~SLib();
-    static SLib* load(const char* file, bool local);
+    static SLib* load(const char* file, bool local, bool nounload);
+    bool unload(bool doNow);
 private:
-    SLib(HMODULE handle, const char* file);
-    const char* m_file;
+    SLib(HMODULE handle, const char* file, bool nounload);
     HMODULE m_handle;
+    bool m_nounload;
 };
 
 class EngineSuperHandler : public MessageHandler
@@ -185,6 +248,44 @@ public:
     virtual bool received(Message &msg);
 };
 
+class EngineHelp : public MessageHandler
+{
+public:
+    EngineHelp() : MessageHandler("engine.help") { }
+    virtual bool received(Message &msg);
+};
+
+
+// helper function to initialize user application data dir
+static void initUsrPath(String& path, const char* newPath = 0)
+{
+    if (path)
+	return;
+    if (TelEngine::null(newPath)) {
+#ifdef _WINDOWS
+	// we force using the ANSI version
+	char szPath[MAX_PATH];
+	if (SHGetSpecialFolderPathA(NULL,szPath,CSIDL_APPDATA,TRUE))
+	    path = szPath;
+#else
+	path = ::getenv("HOME");
+#endif
+	if (path.null()) {
+	    if (Engine::mode() == Engine::Client)
+		Debug(DebugWarn,"Could not get per-user application data path!");
+	    path = s_cfgpath;
+	}
+	if (!path.endsWith(PATH_SEP))
+	    path += PATH_SEP;
+	path += s_userdir;
+    }
+    else
+	path = newPath;
+    if (path.endsWith(PATH_SEP))
+	path = path.substr(0,path.length()-1);
+}
+
+
 bool EngineStatusHandler::received(Message &msg)
 {
     const char *sel = msg.getValue("module");
@@ -201,23 +302,242 @@ bool EngineStatusHandler::received(Message &msg)
     msg.retValue() << ",workers=" << EnginePrivate::count;
     msg.retValue() << ",mutexes=" << Mutex::count();
     msg.retValue() << ",locks=" << Mutex::locks();
+    msg.retValue() << ",semaphores=" << Semaphore::count();
+    msg.retValue() << ",waiting=" << Semaphore::locks();
     msg.retValue() << "\r\n";
     return false;
 }
 
+static char s_cmdsOpt[] = "  module {{load|reload} modulefile|unload modulename|list}\r\n";
+static char s_cmdsMsg[] = "Controls the modules loaded in the Telephony Engine\r\n";
+
+// get the base name of a module file
+static String moduleBase(const String& fname)
+{
+    int sep = fname.rfind(PATH_SEP[0]);
+    if (sep >= 0)
+	sep++;
+    else
+	sep = 0;
+    int len = fname.length() - sep;
+    if (fname.endsWith(Engine::moduleSuffix()))
+	len -= Engine::moduleSuffix().length();
+    return fname.substr(sep,len);
+}
+
+// perform one completion only if match still possible
+static void completeOne(String& ret, const String& str, const char* part)
+{
+    if (part && !str.startsWith(part))
+	return;
+    ret.append(str,"\t");
+}
+
+static void completeOneModule(String& ret, const String& str, const char* part, ObjList& mods, bool reload)
+{
+    SLib* s = static_cast<SLib*>(mods[moduleBase(str)]);
+    if (s) {
+	if (!(reload && s->unload(false)))
+	    return;
+    }
+    else if (reload)
+	return;
+    completeOne(ret,str,part);
+}
+
+// complete a module filename from filesystem
+void completeModule(String& ret, const String& part, ObjList& mods, bool reload, const String& rpath = String::empty())
+{
+    if (part[0] == '.')
+	return;
+    String path = Engine::modulePath();
+    String rdir = rpath;
+    int sep = part.rfind(PATH_SEP[0]);
+    if (sep >= 0)
+	rdir += part.substr(0,sep+1);
+    if (rdir) {
+	if (!path.endsWith(PATH_SEP))
+	    path += PATH_SEP;
+	path += rdir;
+    }
+    if (path.endsWith(PATH_SEP))
+	path = path.substr(0,path.length()-1);
+
+    DDebug(DebugInfo,"completeModule path='%s' rdir='%s'",path.c_str(),rdir.c_str());
+#ifdef _WINDOWS
+    WIN32_FIND_DATA entry;
+    HANDLE hf = ::FindFirstFile(path + PATH_SEP "*",&entry);
+    if (hf == INVALID_HANDLE_VALUE)
+	return;
+    do {
+	XDebug(DebugInfo,"Found dir entry %s",entry.cFileName);
+	if (entry.cFileName[0] == '.')
+	    continue;
+	if ((entry.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) != 0)
+	    continue;
+	if ((entry.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
+	    completeModule(ret,part,mods,reload,rdir + entry.cFileName + PATH_SEP);
+	    continue;
+	}
+	int n = ::strlen(entry.cFileName) - Engine::moduleSuffix().length();
+	if ((n > 0) && !::strcmp(entry.cFileName+n,Engine::moduleSuffix()))
+	    completeOneModule(ret,rdir + entry.cFileName,part,mods,reload);
+    } while (::FindNextFile(hf,&entry));
+    ::FindClose(hf);
+#else
+    DIR *dir = ::opendir(path);
+    if (!dir)
+	return;
+    struct dirent* entry;
+    while ((entry = ::readdir(dir)) != 0) {
+	XDebug(DebugInfo,"Found dir entry %s",entry->d_name);
+	if (entry->d_name[0] == '.')
+	    continue;
+	struct stat stat_buf;
+	if (::stat(path + PATH_SEP + entry->d_name,&stat_buf))
+	    continue;
+	if (S_ISDIR(stat_buf.st_mode)) {
+	    completeModule(ret,part,mods,reload,rdir + entry->d_name + PATH_SEP);
+	    continue;
+	}
+	int n = ::strlen(entry->d_name) - Engine::moduleSuffix().length();
+	if ((n > 0) && !::strcmp(entry->d_name+n,Engine::moduleSuffix()))
+	    completeOneModule(ret,rdir + entry->d_name,part,mods,reload);
+    }
+    ::closedir(dir);
+#endif
+}
+
+// perform command line completion
+void EngineCommand::doCompletion(Message &msg, const String& partLine, const String& partWord)
+{
+    if (partLine.null() || (partLine == "help"))
+	completeOne(msg.retValue(),"module",partWord);
+    else if (partLine == "status")
+	completeOne(msg.retValue(),"engine",partWord);
+    else if (partLine == "module") {
+	completeOne(msg.retValue(),"load",partWord);
+	completeOne(msg.retValue(),"unload",partWord);
+	completeOne(msg.retValue(),"reload",partWord);
+	completeOne(msg.retValue(),"list",partWord);
+    }
+    else if (partLine == "module load")
+	completeModule(msg.retValue(),partWord,Engine::self()->m_libs,false);
+    else if (partLine == "module reload")
+	completeModule(msg.retValue(),partWord,Engine::self()->m_libs,true);
+    else if (partLine == "module unload") {
+	for (ObjList* l = Engine::self()->m_libs.skipNull();l;l = l->skipNext()) {
+	    SLib* s = static_cast<SLib*>(l->get());
+	    if (s->unload(false))
+		completeOne(msg.retValue(),*s,partWord);
+	}
+    }
+}
+
+bool EngineCommand::received(Message &msg)
+{
+    String line = msg.getValue("line");
+    if (line.null()) {
+	doCompletion(msg,msg.getValue("partline"),msg.getValue("partword"));
+	return false;
+    }
+    if (!line.startSkip("module"))
+	return false;
+
+    bool ok = false;
+    int sep = line.find(' ');
+    if (sep > 0) {
+	String cmd = line.substr(0,sep).trimBlanks();
+	String arg = line.substr(sep+1).trimBlanks();
+	if ((cmd == "load") || (cmd == "reload")) {
+	    bool reload = (cmd == "reload");
+	    cmd = moduleBase(arg);
+	    SLib* s = static_cast<SLib*>(Engine::self()->m_libs[cmd]);
+	    if (s) {
+		ok = true;
+		if (reload) {
+		    if (s->unload(true)) {
+			Engine::self()->m_libs.remove(s);
+			ok = false;
+		    }
+		    else
+			msg.retValue() = "Module not unloaded: " + arg + "\r\n";
+		}
+		else
+		    msg.retValue() = "Module is already loaded: " + cmd + "\r\n";
+	    }
+	    if (!ok) {
+		ok = Engine::self()->loadPlugin(Engine::modulePath() + PATH_SEP + arg);
+		// if we loaded it successfully we must initialize
+		if (ok)
+		    Engine::self()->initPlugins();
+	    }
+	}
+	else if (cmd == "unload") {
+	    SLib* s = static_cast<SLib*>(Engine::self()->m_libs[arg]);
+	    if (!s)
+		msg.retValue() = "Module not loaded: " + arg + "\r\n";
+	    else if (s->unload(true)) {
+		Engine::self()->m_libs.remove(s);
+		msg.retValue() = "Unloaded module: " + arg + "\r\n";
+	    }
+	    else
+		msg.retValue() = "Could not unload module: " + arg + "\r\n";
+	    ok = true;
+	}
+    }
+    else if (line == "list") {
+	msg.retValue().clear();
+	for (ObjList* l = Engine::self()->m_libs.skipNull();l;l = l->skipNext()) {
+	    SLib* s = static_cast<SLib*>(l->get());
+	    msg.retValue().append(*s,"\t");
+	    if (s->unload(false))
+		msg.retValue() += "*";
+	}
+	msg.retValue() << "\r\n";
+	ok = true;
+    }
+    if (!ok)
+	msg.retValue() = "Module operation failed: " + line + "\r\n";
+    return true;
+}
+
+
+bool EngineHelp::received(Message &msg)
+{
+    String line = msg.getValue("line");
+    if (line.null()) {
+	msg.retValue() << s_cmdsOpt;
+	return false;
+    }
+    if (line != "module")
+	return false;
+    msg.retValue() << s_cmdsOpt << s_cmdsMsg;
+    return true;
+}
+
+
 void EnginePrivate::run()
 {
     for (;;) {
 	s_makeworker = false;
 	Engine::self()->m_dispatcher.dequeue();
-	msleep(5,true);
+	Thread::idle(true);
     }
 }
 
+
 static bool logFileOpen()
 {
     if (s_logfile) {
-	int fd = ::open(s_logfile,O_WRONLY|O_CREAT|O_APPEND,0640);
+	int flags = O_WRONLY|O_CREAT;
+	if (s_logtruncate) {
+	    s_logtruncate = false;
+	    flags |= O_TRUNC;
+	}
+	else
+	    flags |= O_APPEND;
+	int fd = ::open(s_logfile,flags,0640);
 	if (fd >= 0) {
 	    // Redirect stdout and stderr to the new file
 	    ::fflush(stdout);
@@ -321,7 +641,9 @@ static SERVICE_TABLE_ENTRY dispatchTable[] =
 #define checkPoint()
 
 static bool s_logrotator = false;
-static bool s_runagain = true;
+static volatile bool s_rotatenow = false;
+static volatile bool s_runagain = true;
+static unsigned int s_rundelay = RUNDELAY_MIN;
 static pid_t s_childpid = -1;
 static pid_t s_superpid = -1;
 
@@ -329,11 +651,8 @@ static void superhandler(int signal)
 {
     switch (signal) {
 	case SIGHUP:
-	    if (s_logrotator) {
-		::fprintf(stderr,"Supervisor (%d) closing the log file\n",s_superpid);
-		logFileOpen();
-		::fprintf(stderr,"Supervisor (%d) reopening the log file\n",s_superpid);
-	    }
+	    if (s_logrotator)
+		s_rotatenow = true;
 	    break;
 	case SIGINT:
 	case SIGTERM:
@@ -344,6 +663,16 @@ static void superhandler(int signal)
 	::kill(s_childpid,signal);
 }
 
+static void rotatelogs()
+{
+    if (s_rotatenow) {
+	s_rotatenow = false;
+	::fprintf(stderr,"Supervisor (%d) closing the log file\n",s_superpid);
+	logFileOpen();
+	::fprintf(stderr,"Supervisor (%d) reopening the log file\n",s_superpid);
+    }
+}
+
 static void copystream(int dest, int src)
 {
     for (;;) {
@@ -353,6 +682,7 @@ static void copystream(int dest, int src)
 	    break;
 	::write(dest,buf,rd);
     }
+    rotatelogs();
 }
 
 static int supervise(void)
@@ -388,6 +718,7 @@ static int supervise(void)
 	// reap any children we may have before spawning a new one
 	while (::waitpid(-1,0,WNOHANG) > 0)
 	    ;
+	rotatelogs();
 	s_childpid = ::fork();
 	if (s_childpid < 0) {
 	    int err = errno;
@@ -446,12 +777,17 @@ static int supervise(void)
 		    tmp = MAX_SANITY;
 		if (sanity < tmp)
 		    sanity = tmp;
+		// Decrement inter-run delay each time child proves sanity
+		s_rundelay -= RUNDELAY_DEC;
+		if (s_rundelay < RUNDELAY_MIN)
+		    s_rundelay = RUNDELAY_MIN;
 	    }
 	    else if ((errno != EINTR) && (errno != EAGAIN))
 		break;
 	    // Consume sanity points slightly slower than added
 	    for (int i = 0; i < 12; i++) {
-		copystream(2,logfd[0]);
+		if (s_logrotator)
+		    copystream(2,logfd[0]);
 		::usleep(100000);
 	    }
 	}
@@ -464,25 +800,43 @@ static int supervise(void)
 		::kill(s_childpid,SIGABRT);
 		::usleep(500000);
 	    }
-	    ::kill(s_childpid,SIGKILL);
+	    // Try to kill until it dies or we get a termination signal
+	    while ((s_childpid > 0) && !::kill(s_childpid,SIGKILL)) {
+		if (!s_runagain)
+		    break;
+		::usleep(100000);
+		int status = -1;
+		if (::waitpid(s_childpid,&status,WNOHANG) > 0)
+		    break;
+	    }
 	    s_childpid = -1;
 	}
 	if (s_logrotator) {
 	    copystream(2,logfd[0]);
 	    ::close(logfd[0]);
 	}
-	if (s_runagain)
-	    ::usleep(1000000);
+	if (s_runagain) {
+	    if (s_rundelay > RUNDELAY_MIN)
+		::fprintf(stderr,"Supervisor (%d) delaying child start by %u.%02u seconds\n",
+		    s_superpid,s_rundelay / 1000000,(s_rundelay / 10000) % 100);
+	    ::usleep(s_rundelay);
+	    // Exponential backoff, double run delay each time we restart child
+	    s_rundelay *= 2;
+	    if (s_rundelay > RUNDELAY_MAX)
+		s_rundelay = RUNDELAY_MAX;
+	}
     }
     ::fprintf(stderr,"Supervisor (%d) exiting with code %d\n",s_superpid,retcode);
     return retcode;
 }
 #endif /* _WINDOWS */
 
-SLib::SLib(HMODULE handle, const char* file)
-    : m_handle(handle)
+
+SLib::SLib(HMODULE handle, const char* file, bool nounload)
+    : String(moduleBase(file)), m_handle(handle), m_nounload(nounload)
 {
-    DDebug(DebugAll,"SLib::SLib(%p,'%s') [%p]",handle,file,this);
+    DDebug(DebugAll,"SLib::SLib(%p,'%s',%s) [%p]",
+	handle,file,String::boolText(nounload),this);
     checkPoint();
 }
 
@@ -491,7 +845,7 @@ SLib::~SLib()
 #ifdef DEBUG
     Debugger debug("SLib::~SLib()"," [%p]",this);
 #endif
-    if (s_nounload) {
+    if (s_nounload || m_nounload) {
 #ifdef _WINDOWS
 	typedef void (WINAPI *pFini)(HINSTANCE,DWORD,LPVOID);
 	pFini fini = (pFini)GetProcAddress(m_handle,"_DllMainCRTStartup");
@@ -505,7 +859,7 @@ SLib::~SLib()
 	if (fini)
 	    fini();
 #endif
-	if (fini) {
+	if (fini || m_nounload) {
 	    checkPoint();
 	    return;
 	}
@@ -525,7 +879,7 @@ SLib::~SLib()
     checkPoint();
 }
 
-SLib* SLib::load(const char* file, bool local)
+SLib* SLib::load(const char* file, bool local, bool nounload)
 {
     DDebug(DebugAll,"SLib::load('%s')",file);
     int flags = RTLD_NOW;
@@ -535,18 +889,29 @@ SLib* SLib::load(const char* file, bool local)
 #endif
     HMODULE handle = ::dlopen(file,flags);
     if (handle)
-	return new SLib(handle,file);
+	return new SLib(handle,file,nounload);
 #ifdef _WINDOWS
     Debug(DebugWarn,"LoadLibrary error %u in '%s'",::GetLastError(),file);
 #else
-    Debug(DebugWarn,dlerror());
+    Debug(DebugWarn,"%s",dlerror());
 #endif    
     return 0;
 }
 
+bool SLib::unload(bool unloadNow)
+{
+    typedef bool (*pUnload)(bool);
+    if (m_nounload)
+	return false;
+    pUnload unl = (pUnload)dlsym(m_handle,"_unload");
+    return (unl != 0) && unl(unloadNow);
+}
+
+
 Engine::Engine()
 {
     DDebug(DebugAll,"Engine::Engine() [%p]",this);
+    initUsrPath(s_usrpath);
 }
 
 Engine::~Engine()
@@ -602,9 +967,43 @@ int Engine::run()
     }
 #endif
     SysUsage::init();
+
     s_runid = Time::secNow();
+    if (s_node.trimBlanks().null()) {
+	char hostName[HOST_NAME_MAX+1];
+	if (::gethostname(hostName,sizeof(hostName)))
+	    hostName[0] = '\0';
+	s_node = s_cfg.getValue("general","nodename",hostName);
+	s_node.trimBlanks();
+    }
+    const char *modPath = s_cfg.getValue("general","modpath");
+    if (modPath)
+	s_modpath = modPath;
+    s_maxworkers = s_cfg.getIntValue("general","maxworkers",s_maxworkers);
+    s_restarts = s_cfg.getIntValue("general","restarts");
+    m_dispatcher.warnTime(1000*(u_int64_t)s_cfg.getIntValue("general","warntime"));
+    extraPath(clientMode() ? "client" : "server");
+    extraPath(s_cfg.getValue("general","extrapath"));
+
+    s_params.addParam("version",YATE_VERSION);
+    s_params.addParam("release",YATE_STATUS YATE_RELEASE);
+    s_params.addParam("nodename",s_node);
+    s_params.addParam("runid",String(s_runid));
+    s_params.addParam("configname",s_cfgfile);
+    s_params.addParam("sharedpath",s_shrpath);
+    s_params.addParam("configpath",s_cfgpath);
+    s_params.addParam("usercfgpath",s_usrpath);
+    s_params.addParam("cfgsuffix",s_cfgsuffix);
+    s_params.addParam("modulepath",s_modpath);
+    s_params.addParam("modsuffix",s_modsuffix);
+    s_params.addParam("logfile",s_logfile);
+    s_params.addParam("clientmode",String::boolText(clientMode()));
+    s_params.addParam("supervised",String::boolText(s_super_handle >= 0));
+    s_params.addParam("maxworkers",String(s_maxworkers));
     DDebug(DebugAll,"Engine::run()");
     install(new EngineStatusHandler);
+    install(new EngineCommand);
+    install(new EngineHelp);
     loadPlugins();
     Debug(DebugAll,"Loaded %d plugins",plugins.count());
     if (s_super_handle >= 0) {
@@ -632,7 +1031,8 @@ int Engine::run()
     ::signal(SIGUSR1,sighandler);
     ::signal(SIGUSR2,sighandler);
 #endif
-    Output("Yate engine is initialized and starting up");
+    Output("Yate%s engine is initialized and starting up%s%s",
+	clientMode() ? " client" : "",s_node.null() ? "" : " on " ,s_node.safe());
     while (s_haltcode == -1) {
 	if (s_cmds) {
 	    Output("Executing initial commands");
@@ -752,27 +1152,23 @@ const char* Engine::pathSeparator()
     return PATH_SEP;
 }
 
+const String& Engine::configPath(bool user)
+{
+    if (user) {
+	if (s_createusr) {
+	    // create user data dir on first request
+	    s_createusr = false;
+	    if (::mkdir(s_usrpath,S_IRWXU) == 0)
+		Debug(DebugNote,"Created user data directory: '%s'",s_usrpath.c_str());
+	}
+	return s_usrpath;
+    }
+    return s_cfgpath;
+}
+
 String Engine::configFile(const char* name, bool user)
 {
-    String path;
-    if (user) {
-#ifdef _WINDOWS
-	// we force using the ANSI version
-	char szPath[MAX_PATH];
-	if (SHGetSpecialFolderPathA(NULL,szPath,CSIDL_APPDATA,TRUE))
-	    path = szPath;
-#else
-	path = ::getenv("HOME");
-#endif
-    }
-    if (path.null())
-	path = s_cfgpath;
-    else {
-	if (!path.endsWith(PATH_SEP))
-	    path += PATH_SEP;
-	path += CFG_DIR;
-	::mkdir(path,S_IRWXU);
-    }
+    String path = configPath(user);
     if (!path.endsWith(PATH_SEP))
 	path += PATH_SEP;
     return path + name + s_cfgsuffix;
@@ -790,7 +1186,12 @@ bool Engine::Register(const Plugin* plugin, bool reg)
     if (reg) {
 	if (p)
 	    return false;
-	p = plugins.append(plugin);
+	if (plugin->earlyInit()) {
+	    s_loadMode = LoadEarly;
+	    p = plugins.insert(plugin);
+	}
+	else
+	    p = plugins.append(plugin);
 	p->setDelete(s_dynplugin);
     }
     else if (p)
@@ -798,18 +1199,35 @@ bool Engine::Register(const Plugin* plugin, bool reg)
     return true;
 }
 
-bool Engine::loadPlugin(const char* file, bool local)
+bool Engine::loadPlugin(const char* file, bool local, bool nounload)
 {
     s_dynplugin = false;
-    SLib *lib = SLib::load(file,local);
+    s_loadMode = Engine::LoadLate;
+    SLib *lib = SLib::load(file,local,nounload);
     s_dynplugin = true;
     if (lib) {
-	m_libs.append(lib);
+	switch (s_loadMode) {
+	    case LoadFail:
+		delete lib;
+		return false;
+	    case LoadEarly:
+		// load early - unload late
+		m_libs.append(lib);
+		break;
+	    default:
+		m_libs.insert(lib);
+		break;
+	}
 	return true;
     }
     return false;
 }
 
+void Engine::pluginMode(PluginMode mode)
+{
+    s_loadMode = mode;
+}
+
 bool Engine::loadPluginDir(const String& relPath)
 {
 #ifdef DEBUG
@@ -836,7 +1254,8 @@ bool Engine::loadPluginDir(const String& relPath)
 	int n = ::strlen(entry.cFileName) - s_modsuffix.length();
 	if ((n > 0) && !::strcmp(entry.cFileName+n,s_modsuffix)) {
 	    if (s_cfg.getBoolValue("modules",entry.cFileName,defload))
-		loadPlugin(path + PATH_SEP + entry.cFileName);
+		loadPlugin(path + PATH_SEP + entry.cFileName,false,
+		    s_cfg.getBoolValue("nounload",entry.cFileName));
 	}
     } while (::FindNextFile(hf,&entry));
     ::FindClose(hf);
@@ -853,7 +1272,8 @@ bool Engine::loadPluginDir(const String& relPath)
 	if ((n > 0) && !::strcmp(entry->d_name+n,s_modsuffix)) {
 	    if (s_cfg.getBoolValue("modules",entry->d_name,defload))
 		loadPlugin(path + PATH_SEP + entry->d_name,
-		    s_cfg.getBoolValue("localsym",entry->d_name,s_localsymbol));
+		    s_cfg.getBoolValue("localsym",entry->d_name,s_localsymbol),
+		    s_cfg.getBoolValue("nounload",entry->d_name));
 	}
     }
     ::closedir(dir);
@@ -863,13 +1283,6 @@ bool Engine::loadPluginDir(const String& relPath)
 
 void Engine::loadPlugins()
 {
-    const char *name = s_cfg.getValue("general","modpath");
-    if (name)
-	s_modpath = name;
-    extraPath(s_cfg.getValue("general","extrapath"));
-    s_maxworkers = s_cfg.getIntValue("general","maxworkers",s_maxworkers);
-    s_restarts = s_cfg.getIntValue("general","restarts");
-    m_dispatcher.warnTime(1000*(u_int64_t)s_cfg.getIntValue("general","warntime"));
     NamedList *l = s_cfg.getSection("preload");
     if (l) {
         unsigned int len = l->length();
@@ -927,6 +1340,16 @@ void Engine::extraPath(const String& path)
     s_extramod.append(new String(path));
 }
 
+void Engine::userPath(const String& path)
+{
+    if (path.null())
+	return;
+    if (s_usrpath.null())
+	s_userdir = path;
+    else
+	Debug(DebugWarn,"Engine::userPath('%s') called too late!",path.c_str());
+}
+
 void Engine::halt(unsigned int code)
 {
     if (s_haltcode == -1)
@@ -999,13 +1422,20 @@ static void usage(bool client, FILE* f)
 "   -p filename    Write PID to file\n"
 "   -l filename    Log to file\n"
 "   -n configname  Use specified configuration name (%s)\n"
+"   -e pathname    Path to shared files directory (" SHR_PATH ")\n"
 "   -c pathname    Path to conf files directory (" CFG_PATH ")\n"
+"   -u pathname    Path to user files directory (%s)\n"
 "   -m pathname    Path to modules directory (" MOD_PATH ")\n"
 "   -x relpath     Relative path to extra modules directory (can be repeated)\n"
 "   -w directory   Change working directory\n"
+"   -N nodename    Set the name of this node in a cluster\n"
 #ifdef RLIMIT_CORE
 "   -C             Enable core dumps if possible\n"
 #endif
+#ifdef RLIMIT_NOFILE
+"   -F             Increase the maximum file handle to compiled value\n"
+#endif
+"   -t             Truncate log file, don't append to it\n"
 "   -D[options]    Special debugging options\n"
 "     a            Abort if bugs are encountered\n"
 "     m            Attempt to debug mutex deadlocks\n"
@@ -1032,7 +1462,8 @@ static void usage(bool client, FILE* f)
 "   -s             Supervised, restart if crashes or locks up\n"
 "   -r             Enable rotation of log file (needs -s and -l)\n"
 #endif
-    ,s_cfgfile.safe());
+    ,s_cfgfile.safe()
+    ,s_usrpath.safe());
 }
 
 static void badopt(bool client, char chr, const char* opt)
@@ -1052,7 +1483,7 @@ static void noarg(bool client, const char* opt)
 
 static void version()
 {
-    ::fprintf(stdout,"Yate " YATE_VERSION " " YATE_RELEASE "\n");
+    ::fprintf(stdout,"Yate " YATE_VERSION " " YATE_STATUS YATE_RELEASE "\n");
 }
 
 int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bool fail)
@@ -1068,8 +1499,11 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
     bool colorize = false;
     const char* pidfile = 0;
     const char* workdir = 0;
+    const char* usrpath = 0;
     int debug_level = debugLevel();
 
+    Lockable::startUsingNow();
+
     const char* cfgfile = ::strrchr(argv[0],'/');
     if (!cfgfile)
 	cfgfile = ::strrchr(argv[0],'\\');
@@ -1093,6 +1527,8 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			    continue;
 			}
 			if (!::strcmp(pc,"help")) {
+			    s_mode = mode;
+			    initUsrPath(s_usrpath);
 			    usage(client,stdout);
 			    return 0;
 			}
@@ -1117,10 +1553,13 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			    continue;
 			}
 #endif
+			initUsrPath(s_usrpath);
 			badopt(client,0,argv[i]);
 			return EINVAL;
 			break;
 		    case 'h':
+			s_mode = mode;
+			initUsrPath(s_usrpath);
 			usage(client,stdout);
 			return 0;
 		    case 'v':
@@ -1140,6 +1579,9 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			s_logrotator = true;
 			break;
 #endif
+		    case 't':
+			s_logtruncate = true;
+			break;
 		    case 'p':
 			if (i+1 >= argc) {
 			    noarg(client,argv[i]);
@@ -1164,6 +1606,14 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			pc = 0;
 			cfgfile=argv[++i];
 			break;
+		    case 'e':
+			if (i+1 >= argc) {
+			    noarg(client,argv[i]);
+			    return ENOENT;
+			}
+			pc = 0;
+			s_shrpath=argv[++i];
+			break;
 		    case 'c':
 			if (i+1 >= argc) {
 			    noarg(client,argv[i]);
@@ -1172,6 +1622,14 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			pc = 0;
 			s_cfgpath=argv[++i];
 			break;
+		    case 'u':
+			if (i+1 >= argc) {
+			    noarg(client,argv[i]);
+			    return ENOENT;
+			}
+			pc = 0;
+			usrpath=argv[++i];
+			break;
 		    case 'm':
 			if (i+1 >= argc) {
 			    noarg(client,argv[i]);
@@ -1196,10 +1654,23 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			pc = 0;
 			extraPath(argv[++i]);
 			break;
+		    case 'N':
+			if (i+1 >= argc) {
+			    noarg(client,argv[i]);
+			    return ENOENT;
+			}
+			pc = 0;
+			s_node=argv[++i];
+			break;
 #ifdef RLIMIT_CORE
 		    case 'C':
 			s_coredump = true;
 			break;
+#endif
+#ifdef RLIMIT_NOFILE
+		    case 'F':
+			s_numfiles = true;
+			break;
 #endif
 		    case 'D':
 			while (*++pc) {
@@ -1211,7 +1682,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 				    s_lateabrt = true;
 				    break;
 				case 'm':
-				    Mutex::wait(10000000);
+				    Lockable::wait(10000000);
 				    break;
 #ifdef RTLD_GLOBAL
 				case 'l':
@@ -1246,6 +1717,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 				    tstamp = Debugger::Textual;
 				    break;
 				default:
+				    initUsrPath(s_usrpath);
 				    badopt(client,*pc,argv[i]);
 				    return EINVAL;
 			    }
@@ -1256,6 +1728,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 			version();
 			return 0;
 		    default:
+			initUsrPath(s_usrpath);
 			badopt(client,*pc,argv[i]);
 			return EINVAL;
 		}
@@ -1277,8 +1750,17 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
     if (s_cfgfile.endsWith(".exe") || s_cfgfile.endsWith(".EXE"))
 	s_cfgfile = s_cfgfile.substr(0,s_cfgfile.length()-4);
 
-    if (workdir)
-	::chdir(workdir);
+    initUsrPath(s_usrpath,usrpath);
+
+    if (workdir && ::chdir(workdir)) {
+	int err = errno;
+	::fprintf(stderr,"Could not change working directory to '%s': %s (%d)\n",
+	    workdir,::strerror(err),err);
+	return err;
+    }
+
+    if (s_engineCheck && !s_engineCheck->check(s_cmds))
+	return s_haltcode;
 
 #ifdef _WINDOWS
     if ((mode == Server) && !service)
@@ -1386,6 +1868,11 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 
 #ifdef RLIMIT_CORE
     while (s_coredump) {
+#ifdef HAVE_PRCTL
+#ifdef PR_SET_DUMPABLE
+	prctl(PR_SET_DUMPABLE,1,0,0,0);
+#endif
+#endif
 	struct rlimit lim;
 	if (!::getrlimit(RLIMIT_CORE,&lim)) {
 	    errno = 0;
@@ -1401,6 +1888,24 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 	break;
     }
 #endif
+#ifdef RLIMIT_NOFILE
+    while (s_numfiles) {
+	struct rlimit lim;
+	if (!::getrlimit(RLIMIT_NOFILE,&lim)) {
+	    errno = 0;
+	    if (lim.rlim_cur >= FDSIZE_HACK)
+		break;
+	    lim.rlim_cur = FDSIZE_HACK;
+	    if (lim.rlim_max < FDSIZE_HACK)
+		lim.rlim_max = FDSIZE_HACK;
+	    if (!::setrlimit(RLIMIT_NOFILE,&lim))
+		break;
+	}
+	Debug(DebugWarn,"Could not increase max file handle: %s (%d)",
+	    errno ? strerror(errno) : "hard limit",errno);
+	break;
+    }
+#endif
 
     int retcode = -1;
 #ifndef _WINDOWS
@@ -1424,6 +1929,7 @@ int Engine::main(int argc, const char** argv, const char** env, RunMode mode, bo
 
 void Engine::help(bool client, bool errout)
 {
+    initUsrPath(s_usrpath);
     usage(client, errout ? stderr : stdout);
 }
 
