@@ -24,6 +24,13 @@
 
 using namespace TelEngine;
 
+static const NamedList s_empty("");
+
+const NamedList& NamedList::empty()
+{
+    return s_empty;
+}
+
 NamedList::NamedList(const char* name)
     : String(name)
 {
@@ -39,6 +46,12 @@ NamedList::NamedList(const NamedList& original)
     }
 }
 
+NamedList::NamedList(const char* name, const NamedList& original, const String& prefix)
+    : String(name)
+{
+    copySubParams(original,prefix);
+}
+
 void* NamedList::getObject(const String& name) const
 {
     if (name == "NamedList")
@@ -66,11 +79,11 @@ NamedList& NamedList::setParam(NamedString* param)
 {
     XDebug(DebugInfo,"NamedList::setParam(%p) [\"%s\",\"%s\"]",
         param,param->name().c_str(),param->c_str());
-    NamedString *s = getParam(param->name());
-    if (s) {
-	*s = param->c_str();
-	param->destruct();
-    }
+    if (!param)
+	return *this;
+    ObjList* p = m_params.find(param->name());
+    if (p)
+	p->set(param);
     else
 	m_params.append(param);
     return *this;
@@ -126,10 +139,21 @@ NamedList& NamedList::copyParam(const NamedList& original, const String& name, c
     return *this;
 }
 
+NamedList& NamedList::copyParams(const NamedList& original)
+{
+    XDebug(DebugInfo,"NamedList::copyParams(%p) [%p]",&original,this);
+    for (unsigned int i = 0; i < original.length(); i++) {
+	const NamedString* p = original.getParam(i);
+	if (p)
+	    setParam(p->name(),*p);
+    }
+    return *this;
+}
+
 NamedList& NamedList::copyParams(const NamedList& original, ObjList* list, char childSep)
 {
-    XDebug(DebugInfo,"NamedList::copyParams(%p,%p,'%1s')",
-	&original,list,&childSep);
+    XDebug(DebugInfo,"NamedList::copyParams(%p,%p,'%1s') [%p]",
+	&original,list,&childSep,this);
     for (; list; list = list->next()) {
 	GenObject* obj = list->get();
 	if (!obj)
@@ -144,8 +168,8 @@ NamedList& NamedList::copyParams(const NamedList& original, ObjList* list, char
 
 NamedList& NamedList::copyParams(const NamedList& original, const String& list, char childSep)
 {
-    XDebug(DebugInfo,"NamedList::copyParams(%p,\"%s\",'%1s')",
-	&original,list.c_str(),&childSep);
+    XDebug(DebugInfo,"NamedList::copyParams(%p,\"%s\",'%1s') [%p]",
+	&original,list.c_str(),&childSep,this);
     ObjList* l = list.split(',',false);
     if (l) {
 	copyParams(original,l,childSep);
@@ -154,6 +178,38 @@ NamedList& NamedList::copyParams(const NamedList& original, const String& list,
     return *this;
 }
 
+NamedList& NamedList::copySubParams(const NamedList& original, const String& prefix)
+{
+    XDebug(DebugInfo,"NamedList::copySubParams(%p,\"%s\")",&original,prefix.c_str());
+    if (prefix) {
+	unsigned int offs = prefix.length();
+	unsigned int n = original.length();
+	for (unsigned int i = 0; i < n; i++) {
+	    const NamedString* s = original.getParam(i);
+	    if (s && s->name().startsWith(prefix)) {
+		const char* name = s->name().c_str() + offs;
+		if (*name)
+		    addParam(name,*s);
+	    }
+	}
+    }
+    return *this;
+}
+
+void NamedList::dump(String& str, const char* separator, char quote, bool force) const
+{
+    if (force && str.null())
+	str << separator;
+    str << quote << *this << quote;
+    const ObjList *p = m_params.skipNull();
+    for (; p; p = p->skipNext()) {
+        const NamedString* s = static_cast<const NamedString *>(p->get());
+	String tmp;
+	tmp << quote << s->name() << quote << "=" << quote << *s << quote;
+	str.append(tmp,separator);
+    }
+}
+
 int NamedList::getIndex(const NamedString* param) const
 {
     if (!param)
@@ -195,6 +251,12 @@ NamedString* NamedList::getParam(unsigned int index) const
     return static_cast<NamedString *>(m_params[index]);
 }
 
+const String& NamedList::operator[](const String& name) const
+{
+    const String* s = getParam(name);
+    return s ? *s : String::empty();
+}
+
 const char* NamedList::getValue(const String& name, const char* defvalue) const
 {
     XDebug(DebugInfo,"NamedList::getValue(\"%s\",\"%s\")",name.c_str(),defvalue);
@@ -228,23 +290,44 @@ bool NamedList::getBoolValue(const String& name, bool defvalue) const
 
 int NamedList::replaceParams(String& str, bool sqlEsc, char extraEsc) const
 {
-    int p1;
+    int p1 = 0;
     int cnt = 0;
-    while ((p1 = str.find("${")) >= 0) {
+    while ((p1 = str.find("${",p1)) >= 0) {
 	int p2 = str.find('}',p1+2);
 	if (p2 > 0) {
+	    String def;
 	    String tmp = str.substr(p1+2,p2-p1-2);
 	    tmp.trimBlanks();
-	    DDebug(DebugAll,"NamedList replacing parameter '%s' [%p]",tmp.c_str(),this);
-	    tmp = getValue(tmp);
-	    if (sqlEsc)
-		tmp = tmp.sqlEscape(extraEsc);
-	    str = str.substr(0,p1) + tmp + str.substr(p2+1);
-	    // put a limit to avoid infinite loops
-	    if (++cnt >= 1000) {
-		Debug(DebugWarn,"NamedList reached count %d replacing into '%s' [%p]",cnt,str.c_str(),this);
-		return -1;
+	    int pq = tmp.find('$');
+	    if (pq >= 0) {
+		// param is in ${<name>$<default>} format
+		def = tmp.substr(pq+1).trimBlanks();
+		tmp = tmp.substr(0,pq).trimBlanks();
 	    }
+	    DDebug(DebugAll,"NamedList replacing parameter '%s' [%p]",tmp.c_str(),this);
+	    const String* ns = getParam(tmp);
+	    if (ns) {
+		if (sqlEsc) {
+		    const DataBlock* data = 0;
+		    if (ns->null()) {
+			NamedPointer* np = YOBJECT(NamedPointer,ns);
+			if (np)
+			    data = YOBJECT(DataBlock,np->userData());
+		    }
+		    if (data)
+			tmp = data->sqlEscape(extraEsc);
+		    else
+			tmp = ns->sqlEscape(extraEsc);
+		}
+		else
+		    tmp = *ns;
+	    }
+	    else
+		tmp = def;
+	    str = str.substr(0,p1) + tmp + str.substr(p2+1);
+	    // advance search offset past the string we just replaced
+	    p1 += tmp.length();
+	    cnt++;
 	}
 	else
 	    return -1;
