@@ -24,6 +24,13 @@
 
 using namespace TelEngine;
 
+static const ObjList s_empty;
+
+const ObjList& ObjList::empty()
+{
+    return s_empty;
+}
+
 ObjList::ObjList()
     : m_next(0), m_obj(0), m_delete(true)
 {
@@ -113,7 +120,7 @@ ObjList* ObjList::operator+(int index) const
     return obj;
 }
 
-GenObject* ObjList::operator[](int index) const
+GenObject* ObjList::at(int index) const
 {
     ObjList *obj = operator+(index);
     return obj ? obj->get() : 0;
@@ -148,6 +155,26 @@ ObjList* ObjList::find(const String& str) const
     return const_cast<ObjList*>(n);
 }
 
+int ObjList::index(const GenObject* obj) const
+{
+    if (!obj)
+	return -1;
+    int idx = 0;
+    for (const ObjList* n = this; n; n = n->next(), idx++)
+	if (n->get() == obj)
+	    return idx;
+    return -1;
+}
+
+int ObjList::index(const String& str) const
+{
+    int idx = 0;
+    for (const ObjList* n = this; n; n = n->next(), idx++)
+	if (n->get() && str.matches(n->get()->toString()))
+	    return idx;
+    return -1;
+}
+
 GenObject* ObjList::set(const GenObject* obj, bool delold)
 {
     if (m_obj == obj)
@@ -170,7 +197,9 @@ ObjList* ObjList::insert(const GenObject* obj, bool compact)
 	ObjList *n = new ObjList();
 	n->set(m_obj);
 	set(obj,false);
+	n->m_delete = m_delete;
 	n->m_next = m_next;
+	m_delete = true;
 	m_next = n;
     }
     else
@@ -188,6 +217,8 @@ ObjList* ObjList::append(const GenObject* obj, bool compact)
 	n->m_next = new ObjList();
 	n = n->m_next;
     }
+    else
+	n->m_delete = true;
     n->set(obj);
     return n;
 }
@@ -210,7 +241,10 @@ GenObject* ObjList::remove(bool delobj)
 
     if (delobj && tmp) {
 	XDebug(DebugInfo,"ObjList::remove() deleting %p",tmp);
-	TelEngine::destruct(tmp);
+	// Don't use TelEngine::destruct(): the compiler will call the non-template
+	// function (which doesn't reset the pointer)
+	tmp->destruct();
+	tmp = 0;
     }
     return tmp;
 }
@@ -221,6 +255,12 @@ GenObject* ObjList::remove(GenObject* obj, bool delobj)
     return n ? n->remove(delobj) : 0;
 }
 
+GenObject* ObjList::remove(const String& str, bool delobj)
+{
+    ObjList *n = find(str);
+    return n ? n->remove(delobj) : 0;
+}
+
 void ObjList::clear()
 {
 #ifdef XDEBUG
