@@ -43,6 +43,7 @@
 
 #ifndef _WINDOWS
 
+#include <dirent.h>
 #include <sys/stat.h>
 #include <sys/un.h>
 #define HAS_AF_UNIX
@@ -52,6 +53,8 @@
 
 #include <fcntl.h>
 #include <stdlib.h>
+#include <stdio.h>
+#include <utime.h>
 #endif
 
 #ifndef SHUT_RD
@@ -71,7 +74,37 @@
 
 using namespace TelEngine;
 
-static Mutex s_mutex;
+static Mutex s_mutex(false,"SocketAddr");
+
+
+#ifdef _WINDOWS
+
+// The number of seconds from January 1, 1601 (Windows FILETIME)
+//  to EPOCH January 1, 1970
+#define FILETIME_EPOCH_SEC 11644473600
+
+// Convert from FILETIME (100 nsec units since January 1, 1601)
+//  to time_t (seconds since January 1, 1970)
+static inline unsigned int ftToEpoch(FILETIME& ft)
+{
+    // FILETIME in seconds
+    u_int64_t rval = ((ULARGE_INTEGER*)&ft)->QuadPart / 10000000;
+    // EPOCH time in seconds
+    rval -= FILETIME_EPOCH_SEC;
+    return (unsigned int)rval;
+}
+
+// Convert from time_t (seconds since January 1, 1970)
+//  to FILETIME (100 nsec units since January 1, 1601)
+static void epochToFt(unsigned int secEpoch, FILETIME& ft)
+{
+    u_int64_t time = (secEpoch + FILETIME_EPOCH_SEC) * 10000000;
+    ft.dwLowDateTime = (DWORD)time;
+    ft.dwHighDateTime = (DWORD)(time >> 32);
+}
+
+#endif
+
 
 SocketAddr::SocketAddr(const struct sockaddr* addr, socklen_t len)
     : m_address(0), m_length(0)
@@ -150,7 +183,7 @@ void SocketAddr::assign(const struct sockaddr* addr, socklen_t len)
 #endif
 	}
     }
-    if (addr && (len >= sizeof(struct sockaddr))) {
+    if (addr && (len >= (socklen_t)sizeof(struct sockaddr))) {
 	void* tmp = ::malloc(len);
 	::memcpy(tmp,addr,len);
 	m_address = (struct sockaddr*)tmp;
@@ -185,6 +218,18 @@ bool SocketAddr::host(const String& name)
 	    {
 		in_addr_t a = inet_addr(name);
 		if (a == INADDR_NONE) {
+#ifdef HAVE_GHBN_R
+		    char buf[576];
+		    struct hostent h;
+		    struct hostent* hr = 0;
+		    int errn = 0;
+		    int r = gethostbyname_r(name,&h,buf,sizeof(buf),&hr,&errn);
+		    if (r != ERANGE) {
+			if ((r == 0) && hr && (hr->h_addrtype == AF_INET))
+			    a = *((in_addr_t*)(hr->h_addr_list[0]));
+		    }
+		    else // fallback to single threaded version
+#endif
 		    if (s_mutex.lock(MAX_RESWAIT)) {
 			struct hostent* he = gethostbyname(name);
 			if (he && (he->h_addrtype == AF_INET))
@@ -202,8 +247,16 @@ bool SocketAddr::host(const String& name)
 	    }
 	    break;
 #ifdef AF_INET6
-	// TODO: implement AF_INET6
+	case AF_INET6:
+#ifdef HAVE_PTON
+	    if (inet_pton(family(),name,&((struct sockaddr_in6*)m_address)->sin6_addr) > 0) {
+		stringify();
+		return true;
+	    }
 #endif
+	    // TODO: implement AF_INET6 resolving
+	    break;
+#endif // AF_INET6
 #ifdef HAS_AF_UNIX
 	case AF_UNIX:
 	    if (name.length() >= (UNIX_PATH_MAX-1))
@@ -220,13 +273,31 @@ void SocketAddr::stringify()
 {
     switch (family()) {
 	case AF_INET:
+#ifdef HAVE_NTOP
+	    {
+		char buf[16];
+		buf[0] = '\0';
+		m_host = inet_ntop(family(),&((struct sockaddr_in*)m_address)->sin_addr,
+		    buf,sizeof(buf));
+	    }
+#else
 	    s_mutex.lock();
 	    m_host = inet_ntoa(((struct sockaddr_in*)m_address)->sin_addr);
 	    s_mutex.unlock();
+#endif
 	    break;
 #ifdef AF_INET6
-	// TODO: implement AF_INET6
+	case AF_INET6:
+#ifdef HAVE_NTOP
+	    {
+		char buf[48];
+		buf[0] = '\0';
+		m_host = inet_ntop(family(),&((struct sockaddr_in6*)m_address)->sin6_addr,
+		    buf,sizeof(buf));
+	    }
 #endif
+	    break;
+#endif // AF_INET6
 #ifdef HAS_AF_UNIX
 	case AF_UNIX:
 	    m_host = ((struct sockaddr_un*)m_address)->sun_path;
@@ -341,6 +412,16 @@ bool Stream::setBlocking(bool block)
     return false;
 }
 
+int64_t Stream::length()
+{
+    return 0;
+}
+
+int64_t Stream::seek(SeekPos pos, int64_t offset)
+{
+    return -1;
+}
+
 int Stream::writeData(const char* str)
 {
     if (null(str))
@@ -398,6 +479,71 @@ bool Stream::supportsPairs()
 }
 
 
+int MemoryStream::writeData(const void* buffer, int len)
+{
+    if ((len < 0) || !buffer)
+	return -1;
+    if (!len)
+	return 0;
+    int ovr = m_data.length() - m_offset;
+    if (ovr < 0)
+	ovr = 0;
+    else if (ovr > len)
+	ovr = len;
+    if (ovr > 0) {
+	// overwrite part of the existing data
+	void* dest = m_data.data(m_offset,ovr);
+	if (!dest)
+	    return -1;
+	::memcpy(dest,buffer,ovr);
+	m_offset += ovr;
+	len -= ovr;
+	buffer = static_cast<const char*>(buffer) + ovr;
+    }
+    if (len > 0) {
+	DataBlock tmp(const_cast<void*>(buffer),len,false);
+	m_data += tmp;
+	m_offset += len;
+	tmp.clear(false);
+    }
+    return len + ovr;
+}
+
+int MemoryStream::readData(void* buffer, int len)
+{
+    if ((len <= 0) || !buffer)
+	return -1;
+    if (len + m_offset > m_data.length())
+	len = m_data.length() - m_offset;
+    if (len <= 0)
+	return 0;
+    void* src = m_data.data(m_offset,len);
+    if (!src)
+	return -1;
+    ::memcpy(buffer,src,len);
+    m_offset += len;
+    return len;
+}
+
+int64_t MemoryStream::seek(SeekPos pos, int64_t offset)
+{
+    switch (pos) {
+	case SeekBegin:
+	    break;
+	case SeekEnd:
+	    offset += length();
+	    break;
+	case SeekCurrent:
+	    offset += m_offset;
+	    break;
+    }
+    if ((offset < 0) || (offset > length()))
+	return -1;
+    m_offset = offset;
+    return offset;
+}
+
+
 File::File()
     : m_handle(invalidHandle())
 {
@@ -483,6 +629,8 @@ void File::copyError()
 
 bool File::canRetry() const
 {
+    if (!valid())
+	return false;
     if (!m_error)
 	return true;
     return (m_error == EAGAIN) || (m_error == EINTR)
@@ -499,7 +647,7 @@ bool File::setBlocking(bool block)
 #else
     unsigned long flags = 1;
     flags = ::fcntl(m_handle,F_GETFL);
-    if (flags < 0) {
+    if ((signed long)flags < 0) {
 	copyError();
 	return false;
     }
@@ -517,7 +665,7 @@ bool File::setBlocking(bool block)
 }
 
 bool File::openPath(const char* name, bool canWrite, bool canRead,
-    bool create, bool append, bool binary)
+    bool create, bool append, bool binary, bool pubReadable, bool pubWritable)
 {
     if (!terminate())
 	return false;
@@ -529,13 +677,27 @@ bool File::openPath(const char* name, bool canWrite, bool canRead,
 	access |= GENERIC_WRITE;
     if (canRead)
 	access |= GENERIC_READ;
-    HANDLE h = CreateFile(name,access,0,NULL,create ? OPEN_ALWAYS : OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
+    DWORD createMode;
+    if (create)
+	createMode = (!canRead && !append) ? CREATE_ALWAYS : OPEN_ALWAYS;
+    else
+	createMode = OPEN_EXISTING;
+    DWORD share = 0;
+    if (!canWrite && canRead)
+	share |= FILE_SHARE_READ;
+    HANDLE h = CreateFile(name,access,share,NULL,createMode,FILE_ATTRIBUTE_NORMAL,NULL);
     if (h == invalidHandle()) {
 	copyError();
 	return false;
     }
-    if (append)
-	SetFilePointer(h,0,NULL,FILE_END);
+    // Move file pointer if append. Result might be the same as the error code
+    if (append &&
+	::SetFilePointer(h,0,NULL,FILE_END) == INVALID_SET_FILE_POINTER &&
+	::GetLastError() != NO_ERROR) {
+	copyError();
+	::CloseHandle(h);
+	return false;
+    }
 #else
     int flags = 0;
     if (canWrite)
@@ -546,9 +708,16 @@ bool File::openPath(const char* name, bool canWrite, bool canRead,
 	flags |= O_CREAT;
     if (append)
 	flags |= O_APPEND;
+    else if (!canRead)
+	flags |= O_TRUNC;
     if (binary)
 	flags |= O_BINARY;
-    HANDLE h = ::open(name,flags,S_IRWXU);
+    int mode = S_IRUSR|S_IWUSR;
+    if (pubReadable)
+	mode |= S_IRGRP|S_IROTH;
+    if (pubWritable)
+	mode |= S_IWGRP|S_IWOTH;
+    HANDLE h = ::open(name,flags,mode);
     if (h == invalidHandle()) {
 	copyError();
 	return false;
@@ -559,26 +728,55 @@ bool File::openPath(const char* name, bool canWrite, bool canRead,
     return true;
 }
 
-unsigned int File::length()
+int64_t File::length()
 {
     if (!valid())
 	return 0;
 #ifdef _WINDOWS
-    DWORD sz = GetFileSize(m_handle,NULL);
-    if (sz == (DWORD)-1) {
+    LARGE_INTEGER li;
+    li.LowPart = ::GetFileSize(m_handle,(LPDWORD)(&li.HighPart));
+    if (li.LowPart == INVALID_FILE_SIZE && ::GetLastError() != NO_ERROR) {
 	copyError();
-	return 0;
+	return -1;
     }
-    return sz;
+    return li.QuadPart;
 #else
-    off_t pos = ::lseek(m_handle,0,SEEK_CUR);
-    if (pos == (off_t)-1) {
+    int64_t pos = seek(SeekCurrent);
+    if (pos < 0) {
 	copyError();
 	return 0;
     }
-    off_t len = ::lseek(m_handle,0,SEEK_END);
-    ::lseek(m_handle,pos,SEEK_SET);
-    return (len == (off_t)-1) ? 0 : len;
+    int64_t len = seek(SeekEnd);
+    seek(SeekBegin,pos);
+    return len;
+#endif
+}
+
+// Set the file read/write pointer
+int64_t File::seek(SeekPos pos, int64_t offset)
+{
+    if (!valid())
+	return -1;
+#ifdef _WINDOWS
+    int whence = (pos == SeekBegin) ? FILE_BEGIN : ((pos == SeekEnd) ? FILE_END : FILE_CURRENT);
+    LARGE_INTEGER li;
+    li.QuadPart = offset;
+    li.LowPart = ::SetFilePointer(m_handle,li.LowPart,&li.HighPart,whence);
+    // Check low 32bit value and the last error
+    // It might have the same as the error code
+    if (li.LowPart == INVALID_SET_FILE_POINTER && ::GetLastError() != NO_ERROR) {
+	copyError();
+	return -1;
+    }
+    return li.QuadPart;
+#else
+    int whence = (pos == SeekBegin) ? SEEK_SET : ((pos == SeekEnd) ? SEEK_END : SEEK_CUR);
+    off_t p = ::lseek(m_handle,(off_t)offset,whence);
+    if (p == (off_t)-1) {
+	copyError();
+	return -1;
+    }
+    return (int64_t)p;
 #endif
 }
 
@@ -614,6 +812,10 @@ int File::readData(void* buffer, int length)
 	clearError();
 	return nbytes;
     }
+    else if (::GetLastError() == ERROR_HANDLE_EOF) {
+	clearError();
+	return 0;
+    } 
     copyError();
     return -1;
 #else
@@ -650,11 +852,315 @@ bool File::createPipe(File& reader, File& writer)
     return false;
 }
 
-bool File::remove(const char* name)
+// Retrive the file's modification time (the file must be already opened)
+bool File::getFileTime(unsigned int& secEpoch)
 {
-    if (null(name))
+#ifdef _WINDOWS
+    FILETIME ftWrite;
+    if (::GetFileTime(handle(),NULL,NULL,&ftWrite)) {
+	clearError();
+	secEpoch = ftToEpoch(ftWrite);
+	return true;
+    }
+#else
+    struct stat st;
+    if (0 == ::fstat(handle(),&st)) {
+	clearError();
+	secEpoch = st.st_mtime;
+	return true;
+    }
+#endif
+    copyError();
+    return false;
+}
+
+// Build the MD5 hex digest of an opened file.
+bool File::md5(String& buffer)
+{
+    if (-1 == Stream::seek(0))
 	return false;
-    return !::unlink(name);
+    MD5 md5;
+    unsigned char buf[65536];
+    bool ok = false;
+    unsigned int retry = 3;
+    while (retry) {
+	int n = readData(buf,sizeof(buf));
+	if (n < 0) {
+	    if (canRetry())
+		retry--;
+	    else
+		retry = 0;
+	    continue;
+	}
+	if (n == 0) {
+	    ok = true;
+	    break;
+	}
+	DataBlock tmp(buf,n,false);
+	md5 << tmp;
+	tmp.clear(false);
+    }
+    if (ok)
+	buffer = md5.hexDigest();
+    else
+	buffer = "";
+    return ok;
+}
+
+
+// Set last error and return false
+static inline bool getLastError(int* error)
+{
+    if (error)
+	*error = Thread::lastError();
+    return false;
+}
+
+// Check if a file name is non null
+// Set error and return false if it is
+static inline bool fileNameOk(const char* name, int* error)
+{
+    if (!null(name))
+	return true;
+    if (error)
+#ifdef _WINDOWS
+	*error = ERROR_INVALID_PARAMETER;
+#else
+	*error = EINVAL;
+#endif
+    return false;
+}
+
+// Set a file's modification time
+bool File::setFileTime(const char* name, unsigned int secEpoch, int* error)
+{
+    if (!fileNameOk(name,error))
+	return false;
+#ifdef _WINDOWS
+    File f;
+    if (f.openPath(name,true)) {
+	FILETIME ftWrite;
+	epochToFt(secEpoch,ftWrite);
+	bool ok = (0 != ::SetFileTime(f.handle(),NULL,NULL,&ftWrite));
+	if (!ok && error)
+	    *error = ::GetLastError();
+	f.terminate();
+	return ok;
+    }
+#else
+    struct stat st;
+    if (0 == ::stat(name,&st)) {
+	struct utimbuf tb;
+	tb.actime = st.st_atime;
+	tb.modtime = secEpoch;
+	if (0 == ::utime(name,&tb))
+	    return true;
+    }
+#endif
+    return getLastError(error);
+}
+
+// Retrieve a file's modification time
+bool File::getFileTime(const char* name, unsigned int& secEpoch, int* error)
+{
+    if (!fileNameOk(name,error))
+	return false;
+#ifdef _WINDOWS
+    WIN32_FILE_ATTRIBUTE_DATA fa;
+    if (::GetFileAttributesExA(name,GetFileExInfoStandard,&fa)) {
+	secEpoch = ftToEpoch(fa.ftLastWriteTime);
+	return true;
+    }
+#else
+    struct stat st;
+    if (0 == ::stat(name,&st)) {
+	secEpoch = st.st_mtime;
+	return true;
+    }
+#endif
+    return getLastError(error);
+}
+
+// Check if a file exists
+bool File::exists(const char* name, int* error)
+{
+    if (!fileNameOk(name,error))
+	return false;
+#ifdef _WINDOWS
+    WIN32_FIND_DATA d;
+    HANDLE h = ::FindFirstFile(name,&d);
+    if (h != invalidHandle()) {
+	::FindClose(h);
+	return true;
+    }
+#else
+    if (0 == ::access(name,F_OK))
+	return true;
+#endif
+    return getLastError(error);
+}
+
+// Rename (move) a file (or directory) entry from the filesystem
+bool File::rename(const char* oldFile, const char* newFile, int* error)
+{
+    if (!(fileNameOk(oldFile,error) && fileNameOk(newFile,error)))
+	return false;
+#ifdef _WINDOWS
+    DWORD flags = MOVEFILE_COPY_ALLOWED |  // Allow moving file on another volume
+	MOVEFILE_REPLACE_EXISTING |        // Replace existing
+	MOVEFILE_WRITE_THROUGH;            // Don't return until copy/delete is performed
+    if (::MoveFileExA(oldFile,newFile,flags))
+	return true;
+#else
+    if (0 == ::rename(oldFile,newFile))
+	return true;
+#endif
+    return getLastError(error);
+}
+
+bool File::remove(const char* name, int* error)
+{
+    if (!fileNameOk(name,error))
+	return false;
+#ifdef _WINDOWS
+    if (::DeleteFileA(name))
+	return true;
+#else
+    if (0 == ::unlink(name))
+	return true;
+#endif
+    return getLastError(error);
+}
+
+// Build the MD5 hex digest of a file.
+bool File::md5(const char* name, String& buffer, int* error)
+{
+    File f;
+    bool ok = false;
+    if (f.openPath(name,false,true) && f.md5(buffer))
+	ok = true;
+    else if (error)
+        *error = f.error();
+    f.terminate();
+    return ok;
+}
+
+// Create a folder (directory)
+bool File::mkDir(const char* path, int* error)
+{
+    if (!fileNameOk(path,error))
+	return false;
+#ifdef _WINDOWS
+    if (::CreateDirectoryA(path,NULL))
+	return true;
+#else
+    if (0 == ::mkdir(path,(mode_t)-1))
+	return true;
+#endif
+    return getLastError(error);
+}
+
+// Remove an empty folder (directory)
+bool File::rmDir(const char* path, int* error)
+{
+    if (!fileNameOk(path,error))
+	return false;
+#ifdef _WINDOWS
+    if (::RemoveDirectoryA(path))
+	return true;
+#else
+    if (0 == ::rmdir(path))
+	return true;
+#endif
+    return getLastError(error);
+}
+
+// Skip special directories (. or ..)
+static inline bool skipSpecial(const char* s)
+{
+    return *s && *s == '.' && (!s[1] || (s[1] == '.' && !s[2]));
+}
+
+// Enumerate a folder (directory) content
+bool File::listDirectory(const char* path, ObjList* dirs, ObjList* files, int* error)
+{
+    if (!(dirs || files))
+	return true;
+    if (!fileNameOk(path,error))
+	return false;
+    bool ok = false;
+#ifdef _WINDOWS
+    String name(path);
+    if (!name.endsWith("\\"))
+	name << "\\";
+    name << "*";
+    // Init find
+    WIN32_FIND_DATAA d;
+    HANDLE hFind = ::FindFirstFileA(name,&d);
+    if (hFind == INVALID_HANDLE_VALUE) {
+	if (::GetLastError() == ERROR_NO_MORE_FILES)
+	    return true;
+	return getLastError(error);
+    }
+    // Enumerate content
+    ::SetLastError(0);
+    do {
+        if (d.dwFileAttributes & FILE_ATTRIBUTE_DEVICE ||
+	    skipSpecial(d.cFileName))
+	    continue;
+        if (d.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
+	    if (dirs)
+		dirs->append(new String(d.cFileName));
+	}
+	else if (files)
+	    files->append(new String(d.cFileName));
+    }
+    while (::FindNextFileA(hFind,&d));
+    int code = ::GetLastError();
+    ok = !code || code == ERROR_NO_MORE_FILES;
+    // Get error before closing the handle to avoid having a wrong one
+    if (!ok && error)
+	*error = code;
+    ::FindClose(hFind);
+#else
+    DIR* dir = ::opendir(path);
+    if (!dir) {
+	if (!errno)
+	    return true;
+	return getLastError(error);
+    }
+    struct dirent* entry;
+    while ((entry = ::readdir(dir)) != 0) {
+	if (skipSpecial(entry->d_name))
+	    continue;
+#ifdef _DIRENT_HAVE_D_TYPE
+	if (entry->d_type == DT_DIR) {
+	    if (dirs)
+		dirs->append(new String(entry->d_name));
+	}
+	else if (entry->d_type == DT_REG && files)
+	    files->append(new String(entry->d_name));
+#else
+	struct stat stat_buf;
+	String p;
+	p << path << "/" << entry->d_name;
+	if (::stat(p,&stat_buf))
+	    break;
+	if (S_ISDIR(stat_buf.st_mode)) {
+	    if (dirs)
+		dirs->append(new String(entry->d_name));
+	}
+	else if (S_ISREG(stat_buf.st_mode) && files)
+	    files->append(new String(entry->d_name));
+#endif // _DIRENT_HAVE_D_TYPE
+    }
+    ok = !errno;
+    // Get error before closing DIR to avoid having a wrong one
+    if (!ok && error)
+	*error = errno;
+    ::closedir(dir);
+#endif // _WINDOWS
+    return ok;
 }
 
 
@@ -791,12 +1297,15 @@ bool Socket::checkError(int retcode, bool strict)
 
 bool Socket::canRetry() const
 {
+    if (!valid())
+	return false;
     if (!m_error)
 	return true;
+    return (m_error == EAGAIN) || (m_error == EINTR) ||
 #ifdef _WINDOWS
-    return (m_error == WSAEWOULDBLOCK);
+	(m_error == WSAEWOULDBLOCK);
 #else
-    return (m_error == EAGAIN) || (m_error == EINTR) || (m_error == EWOULDBLOCK);
+	(m_error == EWOULDBLOCK);
 #endif
 }
 
@@ -1004,15 +1513,30 @@ int Socket::readData(void* buffer, int length)
 #endif
 }
 
+bool Socket::canSelect(SOCKET handle)
+{
+    if (handle == invalidHandle())
+	return false;
+#ifdef FD_SETSIZE
+#ifndef _WINDOWS
+    if (handle >= (SOCKET)FD_SETSIZE)
+	return false;
+#endif
+#endif
+    return true;
+}
+
 bool Socket::select(bool* readok, bool* writeok, bool* except, struct timeval* timeout)
 {
+    if (!valid())
+	return false;
 #ifdef FD_SETSIZE
 #ifndef _WINDOWS
     static bool localFail = true;
-    if (m_handle >= FD_SETSIZE) {
+    if (m_handle >= (SOCKET)FD_SETSIZE) {
 	if (localFail) {
 	    localFail = false;
-	    Debug(DebugFail,"Socket::select: handle %d larger than compiled in maximum %d",
+	    Debug(DebugGoOn,"Socket::select: handle %d larger than compiled in maximum %d",
 		m_handle,FD_SETSIZE);
 	}
 	return false;
@@ -1092,7 +1616,7 @@ bool Socket::setBlocking(bool block)
     return checkError(::ioctlsocket(m_handle,FIONBIO,(unsigned long *) &flags));
 #else
     flags = ::fcntl(m_handle,F_GETFL);
-    if (flags < 0) {
+    if ((signed long)flags < 0) {
 	copyError();
 	return false;
     }
