@@ -22,6 +22,8 @@
 
 #include "yateclass.h"
 
+#include <limits.h>
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
@@ -29,6 +31,27 @@
 
 namespace TelEngine {
 
+// String to regular integer conversion, takes into account overflows
+static int strtoi(const char* nptr, char** endptr, int base)
+{
+    errno = 0;
+    long int val = ::strtol(nptr,endptr,base);
+#if INT_MAX != LONG_MAX
+    if (val >= INT_MAX) {
+	errno = ERANGE;
+	val = INT_MAX;
+    }
+    else if (val <= INT_MIN) {
+	errno = ERANGE;
+	val = INT_MIN;
+    }
+#endif
+    // on overflow/underflow mark the entire string as unreadable
+    if ((errno == ERANGE) && endptr)
+	*endptr = (char*) nptr;
+    return (int) val;
+}
+
 String operator+(const String& s1, const String& s2)
 {
     String s(s1.c_str());
@@ -60,7 +83,7 @@ int lookup(const char* str, const TokenDict* tokens, int defvalue, int base)
 		return tokens->value;
     }
     char *eptr = 0;
-    long int val= ::strtol(str,&eptr,base);
+    int val = strtoi(str,&eptr,base);
     if (!eptr || *eptr)
 	return defvalue;
     return val;
@@ -95,7 +118,7 @@ using namespace TelEngine;
 
 static bool isWordBreak(char c, bool nullOk = false)
 {
-    return (c == ' ' || c == '\t' || c == '\n' || (nullOk && !c));
+    return (c == ' ' || c == '\t' || c == '\r' || c == '\n' || (nullOk && !c));
 }
 
 // Decode a single nibble, return -1 on error
@@ -117,6 +140,7 @@ static inline char hexEncode(char nib)
     return hex[nib & 0x0f];
 }
 
+
 StringMatchPrivate::StringMatchPrivate()
 {
     XDebug(DebugAll,"StringMatchPrivate::StringMatchPrivate() [%p]",this);
@@ -157,7 +181,8 @@ void StringMatchPrivate::fixup()
     count = c;
 }
 
-static String s_empty;
+
+static const String s_empty;
 
 const String& String::empty()
 {
@@ -178,7 +203,8 @@ String::String(const char* value, int len)
 }
 
 String::String(const String& value)
-    : m_string(0), m_length(0), m_hash(INIT_HASH), m_matches(0)
+    : GenObject(),
+      m_string(0), m_length(0), m_hash(INIT_HASH), m_matches(0)
 {
     XDebug(DebugAll,"String::String(%p) [%p]",&value,this);
     if (!value.null()) {
@@ -396,7 +422,7 @@ int String::toInteger(int defvalue, int base) const
     if (!m_string)
 	return defvalue;
     char *eptr = 0;
-    long int val= ::strtol(m_string,&eptr,base);
+    int val = strtoi(m_string,&eptr,base);
     if (!eptr || *eptr)
 	return defvalue;
     return val;
@@ -425,8 +451,8 @@ double String::toDouble(double defvalue) const
     return val;
 }
 
-static const char* str_false[] = { "false", "no", "off", "disable", 0 };
-static const char* str_true[] = { "true", "yes", "on", "enable", 0 };
+static const char* str_false[] = { "false", "no", "off", "disable", "f", 0 };
+static const char* str_true[] = { "true", "yes", "on", "enable", "t", 0 };
 
 bool String::toBoolean(bool defvalue) const
 {
@@ -495,6 +521,21 @@ String& String::trimBlanks()
     return *this;
 }
 
+String& String::trimSpaces()
+{
+    if (m_string) {
+	const char *s = m_string;
+	while (*s == ' ' || *s == '\t' || *s == '\v' || *s == '\f' || *s == '\r' || *s == '\n')
+	    s++;
+	const char *e = s;
+	for (const char *p = e; *p; p++)
+	    if (*p != ' ' && *p != '\t' && *p != '\v' && *p != '\f' && *p != '\r' && *p != '\n')
+		e = p+1;
+	assign(s,e-s);
+    }
+    return *this;
+}
+
 String& String::operator=(const char* value)
 {
     if (value && !*value)
@@ -602,7 +643,7 @@ String& String::operator>>(int& store)
 {
     if (m_string) {
 	char *end = 0;
-	long int l = ::strtol(m_string,&end,0);
+	int l = strtoi(m_string,&end,0);
 	if (end && (m_string != end)) {
 	    store = l;
 	    assign(end);
@@ -615,8 +656,15 @@ String& String::operator>>(unsigned int& store)
 {
     if (m_string) {
 	char *end = 0;
+	errno = 0;
 	unsigned long int l = ::strtoul(m_string,&end,0);
-	if (end && (m_string != end)) {
+#if UINT_MAX != ULONG_MAX
+	if (l > UINT_MAX) {
+	    l = UINT_MAX;
+	    errno = ERANGE;
+	}
+#endif
+	if (!errno && end && (m_string != end)) {
 	    store = l;
 	    assign(end);
 	}
@@ -939,14 +987,15 @@ String String::sqlEscape(const char* str, char extraEsc)
     return s;
 }
 
-String String::uriEscape(const char* str, char extraEsc)
+String String::uriEscape(const char* str, char extraEsc, const char* noEsc)
 {
     String s;
     if (TelEngine::null(str))
 	return s;
     char c;
     while ((c=*str++)) {
-	if ((unsigned char)c <= ' ' || c == '%' || c == '+' || c == '?' || c == '&' || c == extraEsc)
+	if ((unsigned char)c <= ' ' || c == '%' || c == extraEsc ||
+	    ((c == '+' || c == '?' || c == '&') && !(noEsc && ::strchr(noEsc,c))))
 	    s << '%' << hexEncode(c >> 4) << hexEncode(c);
 	else
 	    s += c;
@@ -1007,6 +1056,199 @@ unsigned int String::hash(const char* value)
     return h;
 }
 
+int String::lenUtf8(const char* value, unsigned int maxSeq, bool overlong)
+{
+    if (!value)
+	return 0;
+    if (maxSeq < 1)
+	maxSeq = 4; // RFC 3629 default limit
+
+    int count = 0;
+    unsigned int more = 0;
+    int32_t min = 0;
+    int32_t val = 0;
+
+    while (unsigned char c = (unsigned char) *value++) {
+	if (more) {
+	    // all continuation bytes are in range [128..191]
+	    if ((c & 0xc0) != 0x80)
+		return -1;
+	    val = (val << 6) | (c & 0x3f);
+	    if (!--more) {
+		if (overlong)
+		    continue;
+		// got full value, check for overlongs
+		if (val < min)
+		    return -1;
+	    }
+	    continue;
+	}
+	count++;
+	// from 1st byte we find out how many are supposed to follow
+	if (c < 128)      // 1 byte, 0...0x7F, ASCII characters, no check
+	    ;
+	else if (c < 192) // invalid as first UFT-8 byte
+	    return -1;
+	else if (c < 224) {
+	    // 2 bytes, 0x80...0x7FF
+	    min = 0x80;
+	    val = c & 0x1f;
+	    more = 1;
+	}
+	else if (c < 240) {
+	    // 3 bytes, 0x800...0xFFFF, Basic Multilingual Plane
+	    min = 0x800;
+	    val = c & 0x0f;
+	    more = 2;
+	}
+	else if (c < 248) {
+	    // 4 bytes, 0x10000...0x1FFFFF, RFC 3629 limit (10FFFF)
+	    min = 0x10000;
+	    val = c & 0x07;
+	    more = 3;
+	}
+	else if (c < 252) {
+	    // 5 bytes, 0x200000...0x3FFFFFF
+	    min = 0x200000;
+	    val = c & 0x03;
+	    more = 4;
+	}
+	else if (c < 254) {
+	    // 6 bytes, 0x4000000...0x7FFFFFFF
+	    min = 0x4000000;
+	    val = c & 0x01;
+	    more = 5;
+	}
+	else
+	    return -1;
+	// check if we accept a character with such sequence length
+	if (more >= maxSeq)
+	    return -1;
+    }
+    if (more)
+	return -1;
+    return count;
+}
+
+int String::fixUtf8(const char* replace, unsigned int maxSeq, bool overlong)
+{
+    if (null())
+	return 0;
+    if (maxSeq < 1)
+	maxSeq = 4; // RFC 3629 default limit
+    if (!replace)
+	replace = "\xEF\xBF\xBD";
+
+    int count = 0;
+    unsigned int more = 0;
+    int32_t min = 0;
+    int32_t val = 0;
+    unsigned int pos = 0;
+    bool bad = false;
+    String tmp;
+
+    for (unsigned int i = 0; i < m_length; i++) {
+	unsigned char c = (unsigned char) at(i);
+	if (more) {
+	    // remember to reject a character with a too long sequence
+	    if (more >= maxSeq)
+		bad = true;
+	    // all continuation bytes are in range [128..191]
+	    if ((c & 0xc0) != 0x80) {
+		// truncated sequence, must search for 1st byte again
+		more = 0;
+		count++;
+		tmp += replace;
+	    }
+	    else {
+		val = (val << 6) | (c & 0x3f);
+		if (!--more) {
+		    // got full value, check for overlongs
+		    if ((val < min) && !overlong)
+			bad = true;
+		    // finished multibyte, add it to temporary
+		    if (bad) {
+			count++;
+			tmp += replace;
+		    }
+		    else
+			tmp += substr(pos,(int)(i+1-pos));
+		}
+		continue;
+	    }
+	}
+	pos = i;
+	bad = false;
+	// from 1st byte we find out how many are supposed to follow
+	if (c < 128)      // 1 byte, 0...0x7F, ASCII characters, good
+	    ;
+	else if (c < 192) // invalid as first UFT-8 byte
+	    bad = true;
+	else if (c < 224) {
+	    // 2 bytes, 0x80...0x7FF
+	    min = 0x80;
+	    val = c & 0x1f;
+	    more = 1;
+	}
+	else if (c < 240) {
+	    // 3 bytes, 0x800...0xFFFF, Basic Multilingual Plane
+	    min = 0x800;
+	    val = c & 0x0f;
+	    more = 2;
+	}
+	else if (c < 248) {
+	    // 4 bytes, 0x10000...0x1FFFFF, RFC 3629 limit (10FFFF)
+	    min = 0x10000;
+	    val = c & 0x07;
+	    more = 3;
+	}
+	else if (c < 252) {
+	    // 5 bytes, 0x200000...0x3FFFFFF
+	    min = 0x200000;
+	    val = c & 0x03;
+	    more = 4;
+	}
+	else if (c < 254) {
+	    // 6 bytes, 0x4000000...0x7FFFFFFF
+	    min = 0x4000000;
+	    val = c & 0x01;
+	    more = 5;
+	}
+	else
+	    bad = true;
+	if (!more) {
+	    if (bad) {
+		count++;
+		tmp += replace;
+	    }
+	    else
+		tmp += (char)c;
+	}
+    }
+    if (more) {
+	// UTF-8 truncated at end of string
+	count++;
+	tmp += replace;
+    }
+
+    if (count)
+	operator=(tmp);
+    return count;
+}
+
+void* String::getObject(const String& name) const
+{
+    if (name == "String")
+	return const_cast<String*>(this);
+    return GenObject::getObject(name);
+}
+
+const String& String::toString() const
+{
+    return *this;
+}
+
+
 Regexp::Regexp()
     : m_regexp(0), m_flags(0)
 {
@@ -1106,12 +1348,73 @@ bool Regexp::isCaseInsensitive() const
     return (m_flags & REG_ICASE) != 0;
 }
 
+
 NamedString::NamedString(const char* name, const char* value)
     : String(value), m_name(name)
 {
     XDebug(DebugAll,"NamedString::NamedString(\"%s\",\"%s\") [%p]",name,value,this);
 }
 
+const String& NamedString::toString() const
+{
+    return m_name;
+}
+
+void* NamedString::getObject(const String& name) const
+{
+    if (name == "NamedString")
+	return (void*)this;
+    return String::getObject(name);
+}
+
+
+NamedPointer::NamedPointer(const char* name, GenObject* data, const char* value)
+    : NamedString(name,value),
+    m_data(0)
+{
+    userData(data);
+}
+
+NamedPointer::~NamedPointer()
+{
+    userData(0);
+}
+
+// Set obscure data carried by this object.
+void NamedPointer::userData(GenObject* data)
+{
+    TelEngine::destruct(m_data);
+    m_data = data;
+}
+
+// Retrive the pointer carried by this object and release ownership
+GenObject* NamedPointer::takeData()
+{
+    GenObject* tmp = m_data;
+    m_data = 0;
+    return tmp;
+}
+
+void* NamedPointer::getObject(const String& name) const
+{
+    if (name == "NamedPointer")
+	return (void*)this;
+    void* p = NamedString::getObject(name);
+    if (p)
+	return p;
+    if (m_data)
+	return m_data->getObject(name);
+    return 0;
+}
+
+// Called whenever the string value changed. Release the pointer
+void NamedPointer::changed()
+{
+    userData(0);
+    NamedString::changed();
+}
+
+
 void* GenObject::getObject(const String& name) const
 {
     return 0;
@@ -1122,21 +1425,4 @@ const String& GenObject::toString() const
     return String::empty();
 }
 
-void* String::getObject(const String& name) const
-{
-    if (name == "String")
-	return const_cast<String*>(this);
-    return GenObject::getObject(name);
-}
-
-const String& String::toString() const
-{
-    return *this;
-}
-
-const String& NamedString::toString() const
-{
-    return m_name;
-}
-
 /* vi: set ts=8 sw=4 sts=4 noet: */
