@@ -110,7 +110,7 @@ class AnalyzerCons : public DataConsumer, public Runnable
 public:
     AnalyzerCons(const String& type, const char* window = 0);
     virtual ~AnalyzerCons();
-    virtual void Consume(const DataBlock& data, unsigned long tStamp);
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags);
     virtual void statusParams(String& str);
     virtual void run();
 protected:
@@ -130,6 +130,7 @@ class AnalyzerChan : public Channel
 public:
     AnalyzerChan(const String& type, bool outgoing, const char* window = 0);
     virtual ~AnalyzerChan();
+    virtual void destroyed();
     virtual void statusParams(String& str);
     virtual bool callRouted(Message& msg);
     virtual bool msgRinging(Message& msg);
@@ -183,7 +184,7 @@ static TokenDict dict_windows[] = {
     { 0,   0 }
 };
 
-static Mutex s_mutex;
+static Mutex s_mutex(false,"Analyzer");
 
 static int s_res = 1;
 
@@ -223,7 +224,7 @@ AsyncFFT* AsyncFFT::create(unsigned int length, WinType window, Priority prio)
 }
 
 AsyncFFT::AsyncFFT(unsigned int length, WinType window, Priority prio)
-    : Thread("AsyncFFT",prio),
+    : Thread("Async FFT",prio),
       m_ready(false), m_start(false), m_stop(false), m_notify(0),
       m_length(0), m_window(0), m_real(0), m_imag(0), m_nBits(0), m_winName(0)
 {
@@ -303,7 +304,7 @@ void AsyncFFT::run()
     DDebug(&__plugin,DebugAll,"AsyncFFT::run() [%p]",this);
     while (!m_stop) {
 	while (!m_start) {
-	    Thread::msleep(5);
+	    Thread::idle();
 	    if (m_stop)
 		return;
 	}
@@ -448,13 +449,13 @@ AnalyzerCons::~AnalyzerCons()
     s_mutex.unlock();
 }
 
-void AnalyzerCons::Consume(const DataBlock& data, unsigned long tStamp)
+unsigned long AnalyzerCons::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 {
     if (!m_timeStart) {
 	// the first data block may be garbled or have bad timestamp - ignore
 	m_timeStart = Time::now();
 	m_tsStart = tStamp;
-	return;
+	return invalidStamp();
     }
     unsigned int samples = data.length() / 2;
     long delta = tStamp - timeStamp() - samples;
@@ -467,11 +468,11 @@ void AnalyzerCons::Consume(const DataBlock& data, unsigned long tStamp)
 	m_tsGapLength += delta;
     }
     if (!m_spectrum)
-	return;
+	return invalidStamp();
     m_data += data;
     unsigned int len = 2 * m_spectrum->samples();
     if (m_data.length() < len)
-	return;
+	return invalidStamp();
     // limit the length of the buffer
     int toCut = data.length() - (2 * len);
     if (toCut > 0) {
@@ -480,6 +481,7 @@ void AnalyzerCons::Consume(const DataBlock& data, unsigned long tStamp)
     }
     if (m_spectrum->prepare((const short*)m_data.data()))
 	m_data.cut(-(int)len);
+    return invalidStamp();
 }
 
 void AnalyzerCons::run()
@@ -553,6 +555,11 @@ AnalyzerChan::AnalyzerChan(const String& type, bool outgoing, const char* window
 AnalyzerChan::~AnalyzerChan()
 {
     DDebug(this,DebugAll,"AnalyzerChan::~AnalyzerChan() %s [%p]",id().c_str(),this);
+    Engine::enqueue(message("chan.hangup"));
+}
+
+void AnalyzerChan::destroyed()
+{
     RefPointer<AnalyzerCons> cons = YOBJECT(AnalyzerCons,getConsumer());
     char buf[32];
     printTime(buf,(unsigned int)(Time::now() - m_timeStart));
@@ -562,7 +569,7 @@ AnalyzerChan::~AnalyzerChan()
     if (cons)
 	cons->statusParams(str);
     Output("Finished '%s' status: %s",id().c_str(),str.c_str());
-    Engine::enqueue(message("chan.hangup"));
+    Channel::destroyed();
 }
 
 void AnalyzerChan::statusParams(String& str)
@@ -710,6 +717,7 @@ bool AnalyzerDriver::msgExecute(Message& msg, String& dest)
     if (ch) {
 	AnalyzerChan *ac = new AnalyzerChan(dest,true,msg.getValue("window"));
 	if (ch->connect(ac,msg.getValue("reason"))) {
+	    ac->callConnect(msg);
 	    msg.setParam("peerid",ac->id());
 	    ac->startChannel(msg);
 	    ac->deref();
