@@ -72,7 +72,7 @@ class ConfChan;
 static ObjList s_rooms;
 
 // Mutex that protects the source while accessed by the consumer
-static Mutex s_srcMutex;
+static Mutex s_srcMutex(false,"Conference");
 
 // Hold the number of the newest allocated dynamic room
 static int s_roomAlloc = 0;
@@ -96,9 +96,13 @@ public:
 	{ return m_users >= m_maxusers; }
     inline ConfChan* recorder() const
 	{ return m_record; }
+    inline const String& notify() const
+	{ return m_notify; }
     void mix(ConfConsumer* cons = 0);
     void addChannel(ConfChan* chan, bool player = false);
     void delChannel(ConfChan* chan);
+    void dropAll(const char* reason = 0);
+    void msgStatus(Message& msg);
     bool setRecording(const NamedList& params);
 private:
     ConfRoom(const String& name, const NamedList& params);
@@ -118,17 +122,25 @@ class ConfChan : public Channel
 {
     YCLASS(ConfChan,Channel)
 public:
-    ConfChan(const String& name, const NamedList& params);
+    ConfChan(const String& name, const NamedList& params, bool counted, bool utility);
     ConfChan(ConfRoom* room, bool voice = false);
     virtual ~ConfChan();
+    virtual bool msgTone(Message& msg, const char* tone);
+    virtual bool msgText(Message& msg, const char* text);
+    inline bool isCounted() const
+	{ return m_counted; }
     inline bool isUtility() const
 	{ return m_utility; }
     inline bool isRecorder() const
 	{ return m_room && (m_room->recorder() == this); }
+    void populateMsg(Message& msg) const;
 private:
+    void alterMsg(Message& msg, const char* event);
     RefPointer<ConfRoom> m_room;
+    bool m_counted;
     bool m_utility;
     bool m_billing;
+    bool m_keepTarget;
 };
 
 // The data consumer computes energy and noise levels (if required) and
@@ -142,10 +154,10 @@ public:
     ConfConsumer(ConfRoom* room, bool smart = false)
 	: m_room(room), m_src(0), m_muted(false), m_smart(smart),
 	  m_energy2(ENERGY_MIN), m_noise2(ENERGY_MIN)
-	{ }
+	{ DDebug(DebugAll,"ConfConsumer::ConfConsumer(%p,%s) [%p]",room,String::boolText(smart),this); }
     ~ConfConsumer()
-	{ }
-    virtual void Consume(const DataBlock& data, unsigned long tStamp);
+	{ DDebug(DebugAll,"ConfConsumer::~ConfConsumer() [%p]",this); }
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags);
     unsigned int energy() const;
     unsigned int noise() const;
     inline unsigned int energy2() const
@@ -159,7 +171,7 @@ public:
 private:
     void consumed(const int* mixed, unsigned int samples);
     void dataForward(const int* mixed, unsigned int samples);
-    ConfRoom* m_room;
+    RefPointer<ConfRoom> m_room;
     ConfSource* m_src;
     bool m_muted;
     bool m_smart;
@@ -197,17 +209,27 @@ class ConferenceDriver : public Driver
 public:
     ConferenceDriver();
     virtual ~ConferenceDriver();
-    bool checkRoom(String& room, bool existing);
+    bool checkRoom(String& room, bool existing, bool counted);
+    bool unload();
 protected:
     virtual void initialize();
+    virtual bool received(Message &msg, int id);
     virtual bool msgExecute(Message& msg, String& dest);
+    virtual bool commandComplete(Message& msg, const String& partLine, const String& partWord);
     virtual void statusParams(String& str);
 private:
-    bool m_init;
+    ConfHandler* m_handler;
 };
 
 INIT_PLUGIN(ConferenceDriver);
 
+UNLOAD_PLUGIN(unloadNow)
+{
+    if (unloadNow)
+	return __plugin.unload();
+    return true;
+}
+
 // Count the position of the most significant 1 bit - pretty close to logarithm
 static unsigned int binLog(unsigned int x)
 {
@@ -256,6 +278,7 @@ ConfRoom::ConfRoom(const String& name, const NamedList& params)
 	m->addParam("targetid",m_notify);
 	m->addParam("event","created");
 	m->addParam("room",m_name);
+	m->addParam("maxusers",String(m_maxusers));
 	m->addParam("caller",params.getValue("caller"));
 	m->addParam("called",params.getValue("called"));
 	m->addParam("billid",params.getValue("billid"));
@@ -276,6 +299,7 @@ void ConfRoom::destroyed()
 	m->addParam("targetid",m_notify);
 	m->addParam("event","destroyed");
 	m->addParam("room",m_name);
+	m->addParam("maxusers",String(m_maxusers));
 	Engine::enqueue(m);
     }
     DataSource::destroyed();
@@ -289,20 +313,21 @@ void ConfRoom::addChannel(ConfChan* chan, bool player)
     m_chans.append(chan);
     if (player)
 	m_playerId = chan->id();
-    if (!chan->isUtility()) {
+    if (chan->isCounted())
 	m_users++;
-	if (m_notify) {
-	    String tmp(m_users);
-	    Message* m = new Message("chan.notify");
-	    m->addParam("id",chan->id());
-	    m->addParam("targetid",m_notify);
-	    m->addParam("event","joined");
-	    m->addParam("room",m_name);
-	    m->addParam("users",tmp);
-	    if (m_playerId)
-		m->addParam("player",m_playerId);
-	    Engine::enqueue(m);
-	}
+    if (m_notify && !chan->isUtility()) {
+	String tmp(m_users);
+	Message* m = new Message("chan.notify");
+	m->addParam("id",chan->id());
+	m->addParam("targetid",m_notify);
+	chan->populateMsg(*m);
+	m->addParam("event","joined");
+	m->addParam("maxusers",String(m_maxusers));
+	m->addParam("users",tmp);
+	if (m_playerId)
+	    m->addParam("player",m_playerId);
+	m->setParam("lastpeerid",chan->lastPeerId());
+	Engine::enqueue(m);
     }
 }
 
@@ -311,43 +336,69 @@ void ConfRoom::delChannel(ConfChan* chan)
 {
     if (!chan)
 	return;
-    Lock lock(mutex());
+    Lock mylock(this);
     if (chan == m_record)
 	m_record = 0;
     if (m_playerId && (chan->id() == m_playerId))
 	m_playerId.clear();
-    if (m_chans.remove(chan,false) && !chan->isUtility()) {
+    if (m_chans.remove(chan,false) && chan->isCounted())
 	m_users--;
-	bool alone = (m_users == 1);
-	lock.drop();
-	if (m_notify) {
-	    String tmp(m_users);
-	    Message* m = new Message("chan.notify");
-	    m->addParam("id",chan->id());
-	    m->addParam("targetid",m_notify);
-	    m->addParam("event","left");
-	    m->addParam("room",m_name);
-	    m->addParam("users",tmp);
-	    // easy to check parameter indicating one user will be left alone
-	    if (m_lonely)
-		m->addParam("lonely",String::boolText(alone));
-	    if (m_playerId)
-		m->addParam("player",m_playerId);
-	    Engine::enqueue(m);
-	}
-
-	// cleanup if there are only 1 or 0 (if lonely==true) real users left
-	if (m_users <= (m_lonely ? 0 : 1)) {
-	    // all channels left are utility or the lonely user - drop them
-	    // make sure we continue to exist at least as long as the iterator
-	    if (!ref())
-		return;
-	    ListIterator iter(m_chans);
-	    while (ConfChan* ch = static_cast<ConfChan*>(iter.get()))
-		ch->disconnect("hangup");
-	    deref();
-	}
+    bool alone = (m_users == 1);
+    mylock.drop();
+    if (m_notify && !chan->isUtility()) {
+	String tmp(m_users);
+	Message* m = new Message("chan.notify");
+	m->addParam("id",chan->id());
+	m->addParam("targetid",m_notify);
+	chan->populateMsg(*m);
+	m->addParam("event","left");
+	m->addParam("maxusers",String(m_maxusers));
+	m->addParam("users",tmp);
+	// easy to check parameter indicating one user will be left alone
+	if (m_lonely)
+	    m->addParam("lonely",String::boolText(alone));
+	if (m_playerId)
+	    m->addParam("player",m_playerId);
+	m->setParam("lastpeerid",chan->lastPeerId());
+	Engine::enqueue(m);
     }
+
+    // cleanup if there are only 1 or 0 (if lonely==true) real users left
+    if (m_users <= (m_lonely ? 0 : 1))
+	// all channels left are utility or the lonely user - drop them
+	dropAll("hangup");
+}
+
+// Drop all channels attached to the room, the lock must not be hold
+void ConfRoom::dropAll(const char* reason)
+{
+    // make sure we continue to exist at least as long as the iterator
+    if (!ref())
+	return;
+    ListIterator iter(m_chans);
+    while (ConfChan* ch = static_cast<ConfChan*>(iter.get()))
+	ch->disconnect(reason);
+    deref();
+}
+
+// Retrive status information about this room
+void ConfRoom::msgStatus(Message& msg)
+{
+    Lock mylock(this);
+    msg.retValue().clear();
+    msg.retValue() << "name=" << __plugin.prefix() << m_name;
+    msg.retValue() << ",type=conference";
+    msg.retValue() << ";module=" << __plugin.name();
+    msg.retValue() << ",room=" << m_name;
+    msg.retValue() << ",maxusers=" << m_maxusers;
+    msg.retValue() << ",lonely=" << m_lonely;
+    msg.retValue() << ",users=" << m_users;
+    msg.retValue() << ",chans=" << m_chans.count();
+    if (m_notify)
+	msg.retValue() << ",notify=" << m_notify;
+    if (m_playerId)
+	msg.retValue() << ",player=" << m_playerId;
+    msg.retValue() << "\r\n";
 }
 
 // Create or stop outgoing call to room record utility channel
@@ -382,7 +433,11 @@ bool ConfRoom::setRecording(const NamedList& params)
 	m->setParam("called",params.getValue("called"));
 	m->setParam("billid",params.getValue("billid"));
 	m->setParam("username",params.getValue("username"));
+	m->setParam("maxlen",params.getValue("maxlen"));
+	m->setParam("notify",params.getValue("notify"));
 	m->addParam("room",m_name);
+	if (m_notify)
+	    m->setParam("targetid",m_notify);
 	Engine::enqueue(m);
 	if (warn) {
 	    // play record warning to the entire conference
@@ -396,9 +451,12 @@ bool ConfRoom::setRecording(const NamedList& params)
 	    Debug(&__plugin,DebugNote,"Recording '%s' without playing tone!",m_name.c_str());
 	if (m_notify) {
 	    m = ch->message("chan.notify",true,true);
+	    m->addParam("id",ch->id());
 	    m->addParam("targetid",m_notify);
+	    ch->populateMsg(*m);
 	    m->addParam("event","recording");
-	    m->addParam("room",m_name);
+	    m->addParam("maxusers",String(m_maxusers));
+	    m->addParam("users",String(m_users));
 	    m->addParam("record",*record);
 	    Engine::enqueue(m);
 	}
@@ -415,7 +473,7 @@ void ConfRoom::mix(ConfConsumer* cons)
 {
     unsigned int len = MAX_BUFFER;
     unsigned int mlen = 0;
-    Lock mylock(mutex());
+    Lock mylock(this);
     // find out the minimum and maximum amount of data in buffers
     ObjList* l = m_chans.skipNull();
     for (; l; l = l->skipNext()) {
@@ -451,9 +509,15 @@ void ConfRoom::mix(ConfConsumer* cons)
 	    // avoid mixing in noise
 	    if (co->hasSignal()) {
 		unsigned int n = co->m_buffer.length() / 2;
-		XDebug(ch,DebugAll,"Cons %p samp=%u |%s%s>",
-		    co,n,String('#',co->noise()).safe(),
-		    String('=',co->energy() - co->noise()).safe());
+#ifdef XDEBUG
+		int noise = co->noise();
+		int energy = co->energy() - noise;
+		if (energy < 0)
+		    energy = 0;
+		Debug(ch,DebugAll,"Cons %p samp=%u |%s%s>",
+		    co,n,String('#',noise).safe(),
+		    String('=',energy).safe());
+#endif
 		if (n > len)
 		    n = len;
 		const int16_t* p = (const int16_t*)co->m_buffer.data();
@@ -483,10 +547,10 @@ void ConfRoom::mix(ConfConsumer* cons)
 
 
 // Compute the energy level and noise threshold, store the data and call mixer
-void ConfConsumer::Consume(const DataBlock& data, unsigned long tStamp)
+unsigned long ConfConsumer::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 {
     if (m_muted || data.null() || !m_room)
-	return;
+	return 0;
     if (m_smart) {
 	// we need to compute the average energy and take decay into account
 	int64_t sum2 = m_energy2;
@@ -513,18 +577,19 @@ void ConfConsumer::Consume(const DataBlock& data, unsigned long tStamp)
 	    m_noise2 = ENERGY_MIN;
     }
     // make sure looping back conferences is not fatal
-    if (!m_room->mutex()->lock(150000)) {
+    if (!m_room->lock(150000)) {
 	Debug(&__plugin,DebugWarn,"Failed to lock room '%s' - data loopback? [%p]",
 	    m_room->toString().c_str(),this);
 	// mute the channel to avoid getting back here
 	m_muted = true;
-	return;
+	return 0;
     }
     if (m_buffer.length()+data.length() <= MAX_BUFFER)
 	m_buffer += data;
-    m_room->mutex()->unlock();
+    m_room->unlock();
     if (m_buffer.length() >= MIN_BUFFER)
 	m_room->mix(this);
+    return invalidStamp();
 }
 
 // Take out of the buffer the samples mixed in or skipped
@@ -609,20 +674,26 @@ ConfSource::~ConfSource()
 
 // Constructor of a new conference leg, creates or attaches to an existing
 //  conference room; noise and echo suppression are also set here
-ConfChan::ConfChan(const String& name, const NamedList& params)
-    : Channel(__plugin,0,true), m_utility(true), m_billing(false)
+ConfChan::ConfChan(const String& name, const NamedList& params, bool counted, bool utility)
+    : Channel(__plugin,0,true),
+      m_counted(counted), m_utility(utility), m_billing(false), m_keepTarget(true)
 {
     DDebug(this,DebugAll,"ConfChan::ConfChan(%s,%p) %s [%p]",
 	name.c_str(),&params,id().c_str(),this);
     // much of the defaults depend if this is an utility channel or not
-    m_utility = params.getBoolValue("utility",false);
     m_billing = params.getBoolValue("billing",false);
+    m_keepTarget = params.getBoolValue("keeptarget",false);
     bool smart = params.getBoolValue("smart",!m_utility);
     bool echo = params.getBoolValue("echo",m_utility);
     bool voice = params.getBoolValue("voice",true);
     m_room = ConfRoom::get(name,&params);
     if (m_room) {
 	m_address = name;
+	if (!m_utility) {
+	    int tout = params.getIntValue("timeout", driver() ? driver()->timeout() : 0);
+	    if (tout > 0)
+		timeout(Time::now() + tout*(u_int64_t)1000);
+	}
 	m_room->addChannel(this,params.getBoolValue("player",false));
 	RefPointer<ConfConsumer> cons;
 	if (voice) {
@@ -652,7 +723,8 @@ ConfChan::ConfChan(const String& name, const NamedList& params)
 
 // Constructor of an utility conference leg (incoming call)
 ConfChan::ConfChan(ConfRoom* room, bool voice)
-    : Channel(__plugin), m_utility(true), m_billing(false)
+    : Channel(__plugin),
+      m_counted(false), m_utility(true), m_billing(false)
 {
     DDebug(this,DebugAll,"ConfChan::ConfChan(%p,%s) %s [%p]",
 	room,String::boolText(voice),id().c_str(),this);
@@ -685,6 +757,53 @@ ConfChan::~ConfChan()
 	Engine::enqueue(message("chan.hangup"));
 }
 
+// Intercept DTMF messages, possibly turn them into room notifications
+bool ConfChan::msgTone(Message& msg, const char* tone)
+{
+    alterMsg(msg,"dtmf");
+    return false;
+}
+
+// Intercept text messages, possibly turn them into room notifications
+bool ConfChan::msgText(Message& msg, const char* text)
+{
+    alterMsg(msg,"text");
+    return false;
+}
+
+// Populate messages with common conference leg parameters
+void ConfChan::populateMsg(Message& msg) const
+{
+    msg.setParam("counted",String::boolText(m_counted));
+    msg.setParam("utility",String::boolText(m_utility));
+    msg.setParam("room",m_address);
+}
+
+// Alter messages, possibly turn them into room event notifications
+void ConfChan::alterMsg(Message& msg, const char* event)
+{
+    if (m_keepTarget) {
+	String* target = msg.getParam("targetid");
+	// if the message is already targeted to something else don't touch it
+	if (target && *target && (id() != *target))
+	    return;
+    }
+    populateMsg(msg);
+    // if we were the target or it was none send it to the room's notifier
+    if (m_room && m_room->notify()) {
+	msg = "chan.notify";
+	const char* peerid = msg.getValue("id");
+	if (peerid)
+	    msg.setParam("peerid",peerid);
+	msg.setParam("id",id());
+	msg.setParam("event",event);
+	msg.setParam("users",String(m_room->users()));
+	msg.setParam("full",String::boolText(m_room->full()));
+	msg.setParam("targetid",m_room->notify());
+	msg.setParam("lastpeerid",lastPeerId());
+    }
+}
+
 
 bool ConfHandler::received(Message& msg)
 {
@@ -710,7 +829,9 @@ bool ConfHandler::received(Message& msg)
 	return false;
     }
 
-    if (!__plugin.checkRoom(room,msg.getBoolValue("existing")))
+    bool utility = msg.getBoolValue("utility",false);
+    bool counted = msg.getBoolValue("counted",!utility);
+    if (!__plugin.checkRoom(room,msg.getBoolValue("existing"),counted))
 	return false;
 
     const char* reason = msg.getValue("reason","conference");
@@ -730,14 +851,14 @@ bool ConfHandler::received(Message& msg)
     }
 
     // create a conference leg or even a room for the caller
-    ConfChan *c = new ConfChan(room,msg);
+    ConfChan *c = new ConfChan(room,msg,counted,utility);
     if (chan->connect(c,reason,false)) {
 	msg.setParam("peerid",c->id());
 	c->deref();
 	msg.setParam("room",__plugin.prefix()+room);
 	if (peer) {
 	    // create a conference leg for the old peer too
-	    ConfChan *p = new ConfChan(room,msg);
+	    ConfChan *p = new ConfChan(room,msg,counted,utility);
 	    peer->connect(p,reason,false);
 	    p->deref();
 	}
@@ -748,6 +869,39 @@ bool ConfHandler::received(Message& msg)
 }
 
 
+// Message received override to drop entire rooms
+bool ConferenceDriver::received(Message &msg, int id)
+{
+    while (((id == Drop) || (id == Status)) && prefix()) {
+	String dest;
+	switch (id) {
+	    case Drop:
+		dest = msg.getValue("id");
+		break;
+	    case Status:
+		dest = msg.getValue("module");
+		break;
+	}
+	if (!dest.startSkip(prefix(),false))
+	    break;
+	ConfRoom* room = ConfRoom::get(dest);
+	if (!room)
+	    break;
+	switch (id) {
+	    case Drop:
+		room->dropAll(msg.getValue("reason"));
+		room->deref();
+		break;
+	    case Status:
+		room->msgStatus(msg);
+		room->deref();
+		break;
+	}
+	return true;
+    }
+    return Driver::received(msg,id);
+}
+
 /* "call.execute" message
     Input parameters - per room:
 	"callto" - must be of the form "conf/NAME" where NAME is the name of
@@ -764,6 +918,8 @@ bool ConfHandler::received(Message& msg)
     Input parameters - per conference leg:
 	"utility" - true creates a channel that is used for housekeeping
 	    tasks like recording or playing prompts to everybody
+	"counted" - set to false (default for utility) to not count the
+	    conference leg against maxusers or disconnect criteria
 	"voice" - set to false to have the conference leg just listen to the
 	    voice mix without being able to talk
 	"smart" - set to false to disable energy and noise level calculation
@@ -779,12 +935,15 @@ bool ConfHandler::received(Message& msg)
 // Handle call.execute by creating or attaching to an existing conference
 bool ConferenceDriver::msgExecute(Message& msg, String& dest)
 {
-    if (!checkRoom(dest,msg.getBoolValue("existing")))
+    bool utility = msg.getBoolValue("utility",false);
+    bool counted = msg.getBoolValue("counted",!utility);
+    if (!checkRoom(dest,msg.getBoolValue("existing"),counted))
 	return false;
     CallEndpoint* ch = static_cast<CallEndpoint*>(msg.userData());
     if (ch) {
-	ConfChan *c = new ConfChan(dest,msg);
+	ConfChan *c = new ConfChan(dest,msg,counted,utility);
 	if (ch->connect(c)) {
+	    c->callConnect(msg);
 	    msg.setParam("peerid",c->id());
 	    c->deref();
 	    msg.setParam("room",prefix()+dest);
@@ -801,13 +960,13 @@ bool ConferenceDriver::msgExecute(Message& msg, String& dest)
 }
 
 // Check if a room exists, allocates a new room name if not and asked so
-bool ConferenceDriver::checkRoom(String& room, bool existing)
+bool ConferenceDriver::checkRoom(String& room, bool existing, bool counted)
 {
     ConfRoom* conf = ConfRoom::get(room);
     if (existing && !conf)
 	return false;
     if (conf) {
-	bool ok = !conf->full();
+	bool ok = !(counted && conf->full());
 	conf->deref();
 	return ok;
     }
@@ -820,8 +979,21 @@ bool ConferenceDriver::checkRoom(String& room, bool existing)
     return true;
 }
 
+bool ConferenceDriver::unload()
+{
+    Lock lock(this,500000);
+    if (!lock.locked())
+	return false;
+    if (isBusy() || s_rooms.count())
+	return false;
+    uninstallRelays();
+    Engine::uninstall(m_handler);
+    m_handler = 0;
+    return true;
+}
+
 ConferenceDriver::ConferenceDriver()
-    : Driver("conf","misc"), m_init(false)
+    : Driver("conf","misc"), m_handler(0)
 {
     Output("Loaded module Conference");
 }
@@ -832,21 +1004,42 @@ ConferenceDriver::~ConferenceDriver()
     s_rooms.clear();
 }
 
+bool ConferenceDriver::commandComplete(Message& msg, const String& partLine, const String& partWord)
+{
+    bool ok = Driver::commandComplete(msg,partLine,partWord);
+    if (ok && String(msg.getValue("complete")) == "channels") {
+	String tmp = partWord;
+	if (tmp.startSkip(prefix(),false)) {
+	    lock();
+	    ObjList* l = s_rooms.skipNull();
+	    for (; l; l=l->skipNext()) {
+		ConfRoom* r = static_cast<ConfRoom*>(l->get());
+		if (tmp.null() || r->toString().startsWith(tmp))
+		    msg.retValue().append(prefix() + r->toString(),"\t");
+	    }
+	    unlock();
+	}
+    }
+    return ok;
+}
+
 void ConferenceDriver::statusParams(String& str)
 {
     Driver::statusParams(str);
     str.append("rooms=",",") << s_rooms.count();
 }
-    
 
 void ConferenceDriver::initialize()
 {
     Output("Initializing module Conference");
+    // install intercept relays with a priority slightly higher than default
+    installRelay(Tone,75);
+    installRelay(Text,75);
     setup();
-    if (m_init)
+    if (m_handler)
 	return;
-    m_init = true;
-    Engine::install(new ConfHandler(150));
+    m_handler = new ConfHandler(150);
+    Engine::install(m_handler);
 }
 
 }; // anonymous namespace
