@@ -52,12 +52,25 @@ namespace { // anonymous
 // Default message timeout in milliseconds
 #define MSG_TIMEOUT 10000
 
+// Safety wait time after we flushed watchers, relays or messages (in ms)
+#define WAIT_FLUSH 5
+
 static Configuration s_cfg;
 static ObjList s_chans;
 static ObjList s_modules;
-static Mutex s_mutex(true);
+static Mutex s_mutex(true,"ExtModule");
+static int s_waitFlush = WAIT_FLUSH;
 static int s_timeout = MSG_TIMEOUT;
 static bool s_timebomb = false;
+static bool s_pluginSafe = true;
+
+static const char* s_cmds[] = {
+    "info",
+    "start",
+    "stop",
+    "restart",
+    0
+};
 
 class ExtModReceiver;
 class ExtModChan;
@@ -80,7 +93,7 @@ class ExtModConsumer : public DataConsumer
 public:
     ExtModConsumer(Stream* str);
     ~ExtModConsumer();
-    virtual void Consume(const DataBlock& data, unsigned long timestamp);
+    virtual unsigned long Consume(const DataBlock& data, unsigned long timestamp, unsigned long flags);
 private:
     Stream* m_str;
     unsigned m_total;
@@ -114,12 +127,22 @@ public:
 	{ return m_disconn; }
     inline void setId(const String& id)
 	{ CallEndpoint::setId(id); }
+    inline const Message* waitMsg() const
+	{ return m_waitRet; }
+    inline void waitMsg(const Message* msg)
+	{ m_waitRet = msg; }
+    inline bool waiting() const
+	{ return m_waiting; }
+    inline void waiting(bool wait)
+	{ m_waiting = wait; }
 private:
     ExtModChan(const char* file, const char* args, int type);
     ExtModReceiver *m_recv;
+    const Message* m_waitRet;
     int m_type;
     bool m_running;
     bool m_disconn;
+    bool m_waiting;
 };
 
 // Great idea - thanks, Maciek!
@@ -127,10 +150,12 @@ class ExtMessage : public Message
 {
     YCLASS(ExtMessage,Message)
 public:
-    inline ExtMessage(ExtModReceiver* recv)
-	: Message(""), m_receiver(recv)
+    inline ExtMessage()
+	: Message(""),
+	  m_receiver(0), m_accepted(false)
 	{ }
-    void startup();
+    virtual ~ExtMessage();
+    void startup(ExtModReceiver* recv);
     virtual void dispatched(bool accepted);
     inline bool belongsTo(ExtModReceiver* recv) const
 	{ return m_receiver == recv; }
@@ -141,6 +166,7 @@ public:
 private:
     ExtModReceiver* m_receiver;
     String m_id;
+    bool m_accepted;
 };
 
 class MsgHolder : public GenObject
@@ -151,6 +177,8 @@ public:
     bool m_ret;
     String m_id;
     bool decode(const char *s);
+    inline const Message* msg() const
+	{ return &m_msg; }
 };
 
 // Yet Another of Maciek's ideas
@@ -195,8 +223,7 @@ public:
     void cleanup();
     bool flush();
     void die(bool clearChan = true);
-    inline void use()
-	{ m_use++; }
+    void use();
     bool unuse();
     inline const String& scriptFile() const
 	{ return m_script; }
@@ -204,6 +231,9 @@ public:
 	{ return m_args; }
     inline bool selfWatch() const
 	{ return m_selfWatch; }
+    inline void setRestart(bool restart)
+	{ m_restart = restart; }
+    void describe(String& rval) const;
 
 private:
     ExtModReceiver(const char* script, const char* args,
@@ -225,8 +255,10 @@ private:
     MsgWatcher* m_watcher;
     bool m_selfWatch;
     bool m_reenter;
+    bool m_setdata;
     int m_timeout;
     bool m_timebomb;
+    bool m_restart;
     String m_script, m_args;
     ObjList m_waiting;
     ObjList m_relays;
@@ -237,7 +269,7 @@ class ExtThread : public Thread
 {
 public:
     ExtThread(ExtModReceiver* receiver)
-	: Thread("ExtModule"), m_receiver(receiver)
+	: Thread("ExtMod Receiver"), m_receiver(receiver)
 	{ }
     virtual void run()
 	{ m_receiver->run(); }
@@ -259,6 +291,8 @@ class ExtModCommand : public MessageHandler
 public:
     ExtModCommand(const char *name) : MessageHandler(name) { }
     virtual bool received(Message &msg);
+private:
+    bool complete(const String& partLine, const String& partWord, String& rval) const;
 };
 
 class ExtListener : public Thread
@@ -335,7 +369,7 @@ ExtModSource::ExtModSource(Stream* str, ExtModChan* chan)
     Debug(DebugAll,"ExtModSource::ExtModSource(%p) [%p]",str,this);
     if (m_str) {
 	chan->setRunning(true);
-	start("ExtModSource");
+	start("ExtMod Source");
     }
 }
 
@@ -353,9 +387,9 @@ ExtModSource::~ExtModSource()
 void ExtModSource::run()
 {
     char data[320];
-    int r = 0;
+    int r = 1;
     u_int64_t tpos = Time::now();
-    do {
+    while ((r > 0) && looping()) {
 	if (!m_str) {
 	    Thread::yield();
 	    continue;
@@ -381,11 +415,12 @@ void ExtModSource::run()
 	buf.clear(false);
 	m_total += r;
 	tpos += (r*(u_int64_t)1000000/m_brate);
-    } while (r > 0);
+    }
     Debug(DebugAll,"ExtModSource [%p] end of data total=%u",this,m_total);
     m_chan->setRunning(false);
 }
 
+
 ExtModConsumer::ExtModConsumer(Stream* str)
     : m_str(str), m_total(0)
 {
@@ -402,14 +437,17 @@ ExtModConsumer::~ExtModConsumer()
     }
 }
 
-void ExtModConsumer::Consume(const DataBlock& data, unsigned long timestamp)
+unsigned long ExtModConsumer::Consume(const DataBlock& data, unsigned long timestamp, unsigned long flags)
 {
     if ((m_str) && !data.null()) {
 	m_str->writeData(data);
 	m_total += data.length();
+	return invalidStamp();
     }
+    return 0;
 }
 
+
 ExtModChan* ExtModChan::build(const char* file, const char* args, int type)
 {
     ExtModChan* chan = new ExtModChan(file,args,type);
@@ -421,7 +459,8 @@ ExtModChan* ExtModChan::build(const char* file, const char* args, int type)
 }
 
 ExtModChan::ExtModChan(const char* file, const char* args, int type)
-    : CallEndpoint("ExtModule"), m_recv(0), m_type(type), m_disconn(false)
+    : CallEndpoint("ExtModule"),
+      m_recv(0), m_waitRet(0), m_type(type), m_disconn(false), m_waiting(false)
 {
     Debug(DebugAll,"ExtModChan::ExtModChan(%d) [%p]",type,this);
     File* reader = 0;
@@ -501,6 +540,7 @@ void ExtModChan::disconnected(bool final, const char *reason)
     }
 }
 
+
 MsgHolder::MsgHolder(Message &msg)
     : m_msg(msg), m_ret(false)
 {
@@ -516,19 +556,27 @@ bool MsgHolder::decode(const char *s)
 }
 
 
-void ExtMessage::startup()
+ExtMessage::~ExtMessage()
 {
-    if (m_receiver && m_id)
+    if (m_receiver) {
+	m_receiver->returnMsg(this,m_id,m_accepted);
+	m_receiver->unuse();
+    }
+}
+
+void ExtMessage::startup(ExtModReceiver* recv)
+{
+    if (recv && m_id) {
+	m_receiver = recv;
 	m_receiver->use();
+    }
     Engine::enqueue(this);
 }
 
 void ExtMessage::dispatched(bool accepted)
 {
-    if (m_receiver && m_id) {
-	m_receiver->returnMsg(this,m_id,accepted);
-	m_receiver->unuse();
-    }
+    m_accepted = accepted;
+    Message::dispatched(accepted);
 }
   
 
@@ -589,21 +637,13 @@ ExtModReceiver* ExtModReceiver::build(const char* script, const char* args,
 				      File* ain, File* aout, ExtModChan* chan)
 {
     ExtModReceiver* recv = new ExtModReceiver(script,args,ain,aout,chan);
-    if (!recv->start()) {
-	recv->destruct();
-	return 0;
-    }
-    return recv;
+    return recv->start() ? recv : 0;
 }
 
 ExtModReceiver* ExtModReceiver::build(const char* name, Stream* io, ExtModChan* chan, int role)
 {
     ExtModReceiver* recv = new ExtModReceiver(name,io,chan,role);
-    if (!recv->start()) {
-	recv->destruct();
-	return 0;
-    }
-    return recv;
+    return recv->start() ? recv : 0;
 }
 
 ExtModReceiver* ExtModReceiver::find(const String& script)
@@ -618,20 +658,30 @@ ExtModReceiver* ExtModReceiver::find(const String& script)
     return 0;
 }
 
+void ExtModReceiver::use()
+{
+    lock();
+    if (m_use > 0)
+	++m_use;
+    unlock();
+}
+
 bool ExtModReceiver::unuse()
 {
+    lock();
     int u = --m_use;
+    unlock();
     if (!u)
 	destruct();
     return (u <= 0);
 }
 
 ExtModReceiver::ExtModReceiver(const char* script, const char* args, File* ain, File* aout, ExtModChan* chan)
-    : Mutex(true),
-      m_role(RoleUnknown), m_dead(false), m_use(0), m_pid(-1),
+    : Mutex(true,"ExtModReceiver"),
+      m_role(RoleUnknown), m_dead(false), m_use(1), m_pid(-1),
       m_in(0), m_out(0), m_ain(ain), m_aout(aout),
-      m_chan(chan), m_watcher(0), m_selfWatch(false), m_reenter(false),
-      m_timeout(s_timeout), m_timebomb(s_timebomb),
+      m_chan(chan), m_watcher(0), m_selfWatch(false), m_reenter(false), m_setdata(true),
+      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false),
       m_script(script), m_args(args)
 {
     Debug(DebugAll,"ExtModReceiver::ExtModReceiver(\"%s\",\"%s\") [%p]",script,args,this);
@@ -644,11 +694,11 @@ ExtModReceiver::ExtModReceiver(const char* script, const char* args, File* ain,
 }
 
 ExtModReceiver::ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, int role)
-    : Mutex(true),
-      m_role(role), m_dead(false), m_use(0), m_pid(-1),
+    : Mutex(true,"ExtModReceiver"),
+      m_role(role), m_dead(false), m_use(1), m_pid(-1),
       m_in(io), m_out(io), m_ain(0), m_aout(0),
-      m_chan(chan), m_watcher(0), m_selfWatch(false), m_reenter(false),
-      m_timeout(s_timeout), m_timebomb(s_timebomb),
+      m_chan(chan), m_watcher(0), m_selfWatch(false), m_reenter(false), m_setdata(true),
+      m_timeout(s_timeout), m_timebomb(s_timebomb), m_restart(false),
       m_script(name)
 {
     Debug(DebugAll,"ExtModReceiver::ExtModReceiver(\"%s\",%p,%p) [%p]",name,io,chan,this);
@@ -662,16 +712,16 @@ ExtModReceiver::ExtModReceiver(const char* name, Stream* io, ExtModChan* chan, i
 
 ExtModReceiver::~ExtModReceiver()
 {   
-    Debug(DebugAll,"ExtModReceiver::~ExtModReceiver() [%p] pid=%d",this,m_pid);
+    Debug(DebugAll,"ExtModReceiver::~ExtModReceiver() pid=%d [%p]",m_pid,this);
     Lock lock(this);
     // One destruction is plenty enough
-    m_use = -100;
+    m_use = -1;
     s_mutex.lock();
     s_modules.remove(this,false);
     s_mutex.unlock();
     die();
     if (m_pid > 1)
-	Debug(DebugWarn,"ExtModReceiver::~ExtModReceiver() [%p] pid=%d",this,m_pid);
+	Debug(DebugWarn,"ExtModReceiver::~ExtModReceiver() pid=%d [%p]",m_pid,this);
     closeAudio();
 }
 
@@ -679,10 +729,12 @@ void ExtModReceiver::closeIn()
 {
     if (!m_in)
 	return;
+    lock();
     Stream* tmp = m_in;
     m_in = 0;
     if (m_out == tmp)
 	m_out = 0;
+    unlock();
     if (tmp)
 	delete tmp;
 }
@@ -691,10 +743,12 @@ void ExtModReceiver::closeOut()
 {
     if (!m_out)
 	return;
+    lock();
     Stream* tmp = m_out;
     m_out = 0;
     if (m_in == tmp)
 	m_in = 0;
+    unlock();
     if (tmp)
 	delete tmp;
 }
@@ -715,8 +769,11 @@ bool ExtModReceiver::start()
 {
     if (m_pid < 0) {
 	ExtThread *ext = new ExtThread(this);
-	if (!ext->startup())
+	if (!ext->startup()) {
+	    // self destruct here since there is no thread to do it later
+	    unuse();
 	    return false;
+	}
 	while (m_pid < 0)
 	    Thread::yield();
     }
@@ -726,22 +783,35 @@ bool ExtModReceiver::start()
 
 bool ExtModReceiver::flush()
 {
+    lock();
+    bool needWait = (0 != m_watcher);
     TelEngine::destruct(m_watcher);
     // Make sure we release all pending messages and not accept new ones
-    if (!Engine::exiting())
+    needWait = needWait || (m_relays.count() != 0);
+    if (s_pluginSafe)
 	m_relays.clear();
     else {
 	ObjList *p = &m_relays;
 	for (; p; p=p->next())
 	    p->setDelete(false);
     }
+    bool flushed = false;
     if (m_waiting.get()) {
-	DDebug(DebugAll,"ExtModReceiver releasing pending messages [%p]",this);
+	Debug(DebugInfo,"ExtModReceiver releasing %u pending messages [%p]",
+	    m_waiting.count(),this);
 	m_waiting.clear();
-	Thread::yield();
-	return true;
+	needWait = flushed = true;
     }
-    return false;
+    unlock();
+    if (needWait && s_pluginSafe) {
+	int ms = s_waitFlush;
+	// During shutdown longer delays are not acceptable
+	if ((ms > WAIT_FLUSH) && Engine::exiting())
+	    ms = WAIT_FLUSH;
+	DDebug(DebugAll,"ExtModReceiver sleeping %d ms [%p]",ms,this);
+	Thread::msleep(ms);
+    }
+    return flushed;
 }
 
 void ExtModReceiver::die(bool clearChan)
@@ -753,21 +823,24 @@ void ExtModReceiver::die(bool clearChan)
     Debug(DebugAll,"ExtModReceiver::die() pid=%d dead=%s [%p]",
 	m_pid,m_dead ? "yes" : "no",this);
 #endif
+    if (m_dead)
+	return;
+    Lock mylock(this);
     if (m_dead)
 	return;
     m_dead = true;
     use();
-    flush();
 
-    ExtModChan *chan = m_chan;
+    RefPointer<ExtModChan> chan = m_chan;
     m_chan = 0;
     if (chan)
 	chan->setRecv(0);
+    mylock.drop();
 
     // Give the external script a chance to die gracefully
     closeOut();
     if (m_pid > 1) {
-	Debug(DebugAll,"ExtModReceiver::die() waiting for pid=%d to die",m_pid);
+	Debug(DebugAll,"ExtModReceiver::die() waiting for pid=%d to die [%p]",m_pid,this);
 	for (int i=0; i<100; i++) {
 	    Thread::yield();
 	    if (m_pid <= 0)
@@ -775,24 +848,32 @@ void ExtModReceiver::die(bool clearChan)
 	}
     }
     if (m_pid > 1)
-	Debug(DebugInfo,"ExtModReceiver::die() pid=%d did not exit?",m_pid);
+	Debug(DebugInfo,"ExtModReceiver::die() pid=%d did not exit? [%p]",m_pid,this);
 
-    // Now terminate the process and close its stdout pipe
+    // Close the stdout pipe before terminating the process
     closeIn();
+    // Release relays and messages since no confirmation can be received anymore
+    flush();
 #ifndef _WINDOWS
     if (m_pid > 1)
 	::kill(m_pid,SIGTERM);
 #endif
     if (chan && clearChan)
 	chan->disconnect(m_reason);
+    if (m_restart && !Engine::exiting()) {
+	Debug(DebugMild,"Restarting external '%s' '%s'",m_script.safe(),m_args.safe());
+	ExtModReceiver::build(m_script,m_args);
+    }
     unuse();
 }
 
 bool ExtModReceiver::received(Message &msg, int id)
 {
+    if (m_dead)
+	return false;
     lock();
     // check if we are no longer running
-    bool ok = (m_pid > 0) && m_in && m_out;
+    bool ok = (m_pid > 0) && m_in && m_out && !m_dead;
     if (ok && !m_reenter) {
 	// check if the message was generated by ourselves - avoid reentrance
 	ExtMessage* m = YOBJECT(ExtMessage,&msg);
@@ -810,10 +891,10 @@ bool ExtModReceiver::received(Message &msg, int id)
     MsgHolder h(msg);
     if (outputLine(msg.encode(h.m_id))) {
 	m_waiting.append(&h)->setDelete(false);
-	DDebug(DebugAll,"ExtMod [%p] queued message '%s' [%p]",this,msg.c_str(),&msg);
+	DDebug(DebugAll,"ExtMod queued message %p '%s' [%p]",&msg,msg.c_str(),this);
     }
     else {
-	Debug(DebugWarn,"ExtMod [%p] could not queue message '%s'",this,msg.c_str());
+	Debug(DebugWarn,"ExtMod could not queue message %p '%s' [%p]",&msg,msg.c_str(),this);
 	ok = false;
 	fail = true;
     }
@@ -826,15 +907,16 @@ bool ExtModReceiver::received(Message &msg, int id)
 	lock();
 	ok = (m_waiting.find(&h) != 0);
 	if (ok && tout && (Time::now() > tout)) {
-	    Debug(DebugWarn,"Message '%s' did not return in %d msec [%p]",
-		msg.c_str(),m_timeout,this);
+	    Debug(DebugWarn,"Message %p '%s' did not return in %d msec [%p]",
+		&msg,msg.c_str(),m_timeout,this);
 	    m_waiting.remove(&h,false);
 	    ok = false;
 	    fail = true;
 	}
 	unlock();
     }
-    DDebug(DebugAll,"ExtMod [%p] message '%s' [%p] returning %s",this,msg.c_str(),&msg, h.m_ret ? "true" : "false");
+    DDebug(DebugAll,"ExtMod message %p '%s' returning %s [%p]",
+	&msg,msg.c_str(),String::boolText(h.m_ret),this);
     if (fail && m_timebomb)
 	die();
     unuse();
@@ -852,7 +934,9 @@ bool ExtModReceiver::create(const char *script, const char *args)
     HANDLE yate2ext[2];
     int x;
     if (script[0] != '/') {
-	tmp = s_cfg.getValue("general","scripts_dir",SCR_PATH);
+	tmp = Engine::sharedPath();
+	tmp << Engine::pathSeparator() << "scripts";
+	tmp = s_cfg.getValue("general","scripts_dir",tmp);
 	if (!tmp.endsWith(Engine::pathSeparator()))
 	    tmp += Engine::pathSeparator();
 	tmp += script;
@@ -961,31 +1045,36 @@ void ExtModReceiver::run()
 	m_pid = 0;
 	return;
     }
-    use();
-    if (m_in)
-	m_in->setBlocking(false);
+    if (m_in && !m_in->setBlocking(false))
+	Debug("ExtModule",DebugWarn,"Failed to set nonblocking mode, expect trouble [%p]",this);
     char buffer[MAX_INCOMING_LINE];
     int posinbuf = 0;
+    bool invalid = true;
     DDebug(DebugAll,"ExtModReceiver::run() entering loop [%p]",this);
     for (;;) {
 	use();
+	lock();
 	int readsize = m_in ? m_in->readData(buffer+posinbuf,sizeof(buffer)-posinbuf-1) : 0;
+	unlock();
 	if (unuse())
 	    return;
 	if (!readsize) {
-	    lock();
 	    if (m_in)
 		Debug("ExtModule",DebugInfo,"Read EOF on %p [%p]",m_in,this);
 	    closeIn();
 	    flush();
-	    unlock();
+	    if (invalid)
+		Debug("ExtModule",DebugWarn,"Never got anything valid from '%s' '%s'",
+		m_script.c_str(),m_args.safe());
 	    if (m_chan && m_chan->running())
 		Thread::sleep(1);
 	    break;
 	}
 	else if (readsize < 0) {
+	    Lock mylock(this);
 	    if (m_in && m_in->canRetry()) {
-		Thread::msleep(5);
+		mylock.drop();
+		Thread::idle();
 		continue;
 	    }
 	    Debug("ExtModule",DebugWarn,"Read error %d on %p [%p]",errno,m_in,this);
@@ -1004,6 +1093,7 @@ void ExtModReceiver::run()
 	    if ((eoline > buffer) && (eoline[-1] == '\r'))
 		eoline[-1] = 0;
 	    if (buffer[0]) {
+		invalid = invalid && (buffer[0] != '%' || buffer[1] != '%');
 		use();
 		bool goOut = processLine(buffer);
 		if (unuse() || goOut)
@@ -1018,14 +1108,50 @@ void ExtModReceiver::run()
 
 bool ExtModReceiver::outputLine(const char* line)
 {
+    lock();
     DDebug("ExtModReceiver",DebugAll,"%soutputLine '%s'",
-	(m_out ? "" : "failing "), line);
-    if (!m_out)
+	((m_out && !m_dead) ? "" : "failing "), line);
+    if (m_dead || !m_out) {
+	unlock();
 	return false;
-    m_out->writeData(line);
+    }
+    // since m_out can be non-blocking (the socket) we have to loop
+    int len = ::strlen(line);
+    while (m_out && (len > 0) && !m_dead) {
+	int w = m_out->writeData(line,len);
+	if (w < 0) {
+	    if (m_dead || !m_out || !m_out->canRetry()) {
+		unlock();
+		return false;
+	    }
+	}
+	else {
+	    line += w;
+	    len -= w;
+	}
+	if (len > 0) {
+	    unlock();
+	    Thread::yield();
+	    lock();
+	}
+    }
     char nl = '\n';
-    m_out->writeData(&nl,sizeof(nl));
-    return true;
+    for (;;) {
+	if (m_dead || !m_out) {
+	    unlock();
+	    return false;
+	}
+	int w = m_out->writeData(&nl,1);
+	if ((w < 0) && m_out->canRetry())
+	    w = 0;
+	unlock();
+	if (w > 0)
+	    return true;
+	if (w < 0)
+	    return false;
+	Thread::yield();
+	lock();
+    }
 }
 
 void ExtModReceiver::reportError(const char* line)
@@ -1037,14 +1163,14 @@ void ExtModReceiver::reportError(const char* line)
 void ExtModReceiver::returnMsg(const Message* msg, const char* id, bool accepted)
 {
     String ret(msg->encode(accepted,id));
-    lock();
     outputLine(ret);
-    unlock();
 }
 
 bool ExtModReceiver::addWatched(const String& name)
 {
     Lock mylock(this);
+    if (m_dead)
+	return false;
     if (!m_watcher) {
 	m_watcher = new MsgWatcher(this);
 	Engine::self()->setHook(m_watcher);
@@ -1055,11 +1181,15 @@ bool ExtModReceiver::addWatched(const String& name)
 bool ExtModReceiver::delWatched(const String& name)
 {
     Lock mylock(this);
+    if (m_dead)
+	return false;
     return m_watcher && m_watcher->delWatched(name);
 }
 
 bool ExtModReceiver::processLine(const char* line)
 {
+    if (m_dead)
+	return false;
     DDebug("ExtModReceiver",DebugAll,"processLine '%s'", line);
     String id(line);
     if (m_role == RoleUnknown) {
@@ -1103,18 +1233,23 @@ bool ExtModReceiver::processLine(const char* line)
 	for (; p; p=p->next()) {
 	    MsgHolder *msg = static_cast<MsgHolder *>(p->get());
 	    if (msg && msg->decode(line)) {
-		DDebug("ExtModReceiver",DebugInfo,"Matched message");
+		DDebug("ExtModReceiver",DebugInfo,"Matched message %p [%p]",msg->msg(),this);
+		if (m_chan && (m_chan->waitMsg() == msg->msg())) {
+		    DDebug("ExtModReceiver",DebugNote,"Entering wait mode on channel %p [%p]",m_chan,this);
+		    m_chan->waitMsg(0);
+		    m_chan->waiting(true);
+		}
 		p->remove(false);
 		return false;
 	    }
 	}
-	Debug("ExtModReceiver",DebugWarn,"Unmatched message: %s",line);
+	Debug("ExtModReceiver",(m_dead ? DebugInfo : DebugWarn),
+	    "Unmatched%s message: %s [%p]",(m_dead ? " dead" : ""),line,this);
 	return false;
     }
     else if (id.startSkip("%%>install:",false)) {
 	int prio = 100;
 	id >> prio >> ":";
-	bool ok = true;
 	String fname;
 	String fvalue;
 	Regexp r("^\\([^:]*\\):\\([^:]*\\):\\?\\(.*\\)");
@@ -1125,7 +1260,8 @@ bool ExtModReceiver::processLine(const char* line)
 	    id = id.matchString(1);
 	}
 	// sanity checks
-	ok = ok && id && !m_relays.find(id);
+	lock();
+	bool ok = id && !m_dead && !m_relays.find(id);
 	if (ok) {
 	    MessageRelay *r = new MessageRelay(id,this,0,prio);
 	    if (fname)
@@ -1133,6 +1269,7 @@ bool ExtModReceiver::processLine(const char* line)
 	    m_relays.append(r);
 	    Engine::install(r);
 	}
+	unlock();
 	if (debugAt(DebugAll)) {
 	    String tmp;
 	    if (fname)
@@ -1149,6 +1286,7 @@ bool ExtModReceiver::processLine(const char* line)
     else if (id.startSkip("%%>uninstall:",false)) {
 	int prio = 0;
 	bool ok = false;
+	lock();
 	ObjList *p = &m_relays;
 	for (; p; p=p->next()) {
 	    MessageRelay *r = static_cast<MessageRelay *>(p->get());
@@ -1159,6 +1297,7 @@ bool ExtModReceiver::processLine(const char* line)
 		break;
 	    }
 	}
+	unlock();
 	Debug("ExtModReceiver",DebugAll,"Uninstall '%s' %s", id.c_str(),ok ? "ok" : "failed");
 	String out("%%<uninstall:");
 	out << prio << ":" << id << ":" << ok;
@@ -1192,6 +1331,9 @@ bool ExtModReceiver::processLine(const char* line)
 	    String val(id.substr(col+1));
 	    id = id.substr(0,col);
 	    bool ok = false;
+	    Lock mylock(this);
+	    if (m_dead)
+		return false;
 	    if (m_chan && (id == "id")) {
 		if (val.null())
 		    val = m_chan->id();
@@ -1218,16 +1360,36 @@ bool ExtModReceiver::processLine(const char* line)
 		val = m_timebomb;
 		ok = true;
 	    }
+	    else if (id == "restart") {
+		m_restart = (RoleGlobal == m_role) && val.toBoolean(m_restart);
+		val = m_restart;
+		ok = true;
+	    }
 	    else if (id == "reenter") {
 		m_reenter = val.toBoolean(m_reenter);
 		val = m_reenter;
 		ok = true;
 	    }
+	    else if (id == "setdata") {
+		m_setdata = val.toBoolean(m_setdata);
+		val = m_setdata;
+		ok = true;
+	    }
 	    else if (id == "selfwatch") {
 		m_selfWatch = val.toBoolean(m_selfWatch);
 		val = m_selfWatch;
 		ok = true;
 	    }
+	    else if (id.startsWith("engine.")) {
+		// keep the index in substr in sync with length of "engine."
+		const NamedString* param = Engine::runParams().getParam(id.substr(7));
+		ok = val.null() && param;
+		val = param;
+	    }
+	    else if (id == "runid") {
+		ok = val.null();
+		val = Engine::runId();
+	    }
 	    DDebug("ExtModReceiver",DebugAll,"Set '%s'='%s' %s",
 		id.c_str(),val.c_str(),ok ? "ok" : "failed");
 	    String out("%%<setlocal:");
@@ -1241,36 +1403,53 @@ bool ExtModReceiver::processLine(const char* line)
 	return true;
     }
     else {
-	ExtMessage* m = new ExtMessage(this);
+	ExtMessage* m = new ExtMessage;
 	if (m->decode(line) == -2) {
 	    DDebug("ExtModReceiver",DebugAll,"Created message %p '%s' [%p]",m,m->c_str(),this);
 	    lock();
+	    bool note = true;
+	    while (!m_dead && m_chan && m_chan->waiting()) {
+		if (note) {
+		    note = false;
+		    Debug("ExtModReceiver",DebugNote,"Waiting before enqueueing new message %p '%s' [%p]",
+			m,m->c_str(),this);
+		}
+		unlock();
+		Thread::yield();
+		if (m_dead) {
+		    m->destruct();
+		    return false;
+		}
+		lock();
+	    }
 	    ExtModChan* chan = 0;
-	    if ((m_role == RoleChannel) && !m_chan && (*m == "call.execute")) {
+	    if ((m_role == RoleChannel) && !m_chan && m_setdata && (*m == "call.execute")) {
 		// we delayed channel creation as there was nothing to ref() it
 		chan = new ExtModChan(this);
 		m_chan = chan;
 		m->setParam("id",chan->id());
 	    }
-	    m->userData(m_chan);
+	    if (m_setdata)
+		m->userData(m_chan);
 	    // now the newly created channel is referenced by the message
 	    if (chan)
 		chan->deref();
 	    id = m->id();
-	    if (id) {
+	    if (id && !chan) {
 		// Copy the user data pointer from waiting message with same id
 		ObjList *p = &m_waiting;
 		for (; p; p=p->next()) {
 		    MsgHolder *h = static_cast<MsgHolder *>(p->get());
 		    if (h && (h->m_id == id)) {
 			RefObject* ud = h->m_msg.userData();
-			Debug("ExtModReceiver",DebugAll,"Copying data pointer %p from %p",ud,&(h->m_msg));
+			Debug("ExtModReceiver",DebugAll,"Copying data pointer %p from %p '%s' [%p]",
+			    ud,h->msg(),h->msg()->c_str(),this);
 			m->userData(ud);
 			break;
 		    }
 		}
 	    }
-	    m->startup();
+	    m->startup(this);
 	    unlock();
 	    return false;
 	}
@@ -1280,6 +1459,35 @@ bool ExtModReceiver::processLine(const char* line)
     return false;
 }
 
+void ExtModReceiver::describe(String& rval) const
+{
+    rval << "\t";
+    switch (m_role) {
+	case RoleUnknown:
+	    rval << "Unknown";
+	    break;
+	case RoleGlobal:
+	    rval << "Global";
+	    break;
+	case RoleChannel:
+	    rval << "Channel";
+	    break;
+	default:
+	    rval << "Invalid";
+	    break;
+    }
+    if (m_dead)
+	rval << ", dead, use=" << m_use;
+    if (m_chan)
+	rval << ", has channel";
+    if (m_restart)
+	rval << ", autorestart";
+    if (m_pid > 0)
+	rval << ", pid=" << m_pid;
+    rval << "\r\n";
+}
+
+
 bool ExtModHandler::received(Message& msg)
 {
     String dest(msg.getValue("callto"));
@@ -1317,36 +1525,48 @@ bool ExtModHandler::received(Message& msg)
 	Debug(DebugGoOn,"Failed to create ExtMod for '%s'",dest.matchString(2).c_str());
 	return false;
     }
+    // new messages must be blocked until connect() returns (if applicable)
+    if (ch)
+	em->waitMsg(&msg);
     if (!(em->receiver() && em->receiver()->received(msg,1))) {
+	em->waitMsg(0);
 	int level = DebugWarn;
 	if (msg.getValue("error") || msg.getValue("reason"))
 	    level = DebugNote;
 	Debug(level,"ExtMod '%s' did not handle call message",dest.matchString(2).c_str());
+	em->waiting(false);
 	em->deref();
 	return false;
     }
-    if (ch)
+    if (ch) {
+	em->waitMsg(0);
 	ch->connect(em);
+	em->waiting(false);
+    }
     em->deref();
     return true;
 }
 
+
 bool ExtModCommand::received(Message& msg)
 {
     String line(msg.getValue("line"));
     if (!line.startsWith("external",true))
-	return false;
+	return complete(msg.getValue("partline"),msg.getValue("partword"),msg.retValue());;
     line >> "external";
     line.trimBlanks();
-    if (line.null()) {
+    if (line.null() || line == "info") {
 	msg.retValue() = "";
 	int n = 0;
 	Lock lock(s_mutex);
 	ObjList *l = &s_modules;
 	for (; l; l=l->next()) {
 	    ExtModReceiver *r = static_cast<ExtModReceiver *>(l->get());
-	    if (r)
+	    if (r) {
 		msg.retValue() << ++n << ". " << r->scriptFile() << " " << r->commandArg() << "\r\n";
+		if (line)
+		    r->describe(msg.retValue());
+	    }
 	}
 	return true;
     }
@@ -1364,6 +1584,7 @@ bool ExtModCommand::received(Message& msg)
 		return true;
 	    }
 	    else {
+		r->setRestart(false);
 		r->die();
 		msg.retValue() = "External command stopped\r\n";
 	    }
@@ -1379,9 +1600,24 @@ bool ExtModCommand::received(Message& msg)
     return true;
 }
 
+bool ExtModCommand::complete(const String& partLine, const String& partWord, String& rval) const
+{
+    if (partLine.null() && partWord.null())
+	return false;
+    if (partLine.null())
+	Module::itemComplete(rval,"external",partWord);
+    else if (partLine == "external") {
+	for (const char** list = s_cmds; *list; list++)
+	    Module::itemComplete(rval,*list,partWord);
+	return true;
+    }
+    return false;
+}
+
 
 ExtListener::ExtListener(const char* name)
-    : Thread("ExtListener"), m_name(name), m_role(ExtModReceiver::RoleUnknown)
+    : Thread("ExtMod Listener"),
+      m_name(name), m_role(ExtModReceiver::RoleUnknown)
 {
 }
 
@@ -1437,7 +1673,7 @@ void ExtListener::run()
 {
     SocketAddr addr;
     for (;;) {
-	Thread::msleep(5,true);
+	Thread::idle(true);
 	Socket* skt = m_socket.accept(addr);
 	if (!skt) {
 	    if (m_socket.canRetry())
@@ -1486,6 +1722,7 @@ ExtModulePlugin::~ExtModulePlugin()
 {
     Output("Unloading module ExtModule");
     s_mutex.lock();
+    s_pluginSafe = false;
     s_modules.clear();
     // the receivers destroyed above should also clear chans but better be sure
     s_chans.clear();
@@ -1505,6 +1742,12 @@ void ExtModulePlugin::initialize()
     s_cfg.load();
     s_timeout = s_cfg.getIntValue("general","timeout",MSG_TIMEOUT);
     s_timebomb = s_cfg.getBoolValue("general","timebomb",false);
+    int wf = s_cfg.getIntValue("general","waitflush",WAIT_FLUSH);
+    if (wf < 1)
+	wf = 1;
+    else if (wf > 100)
+	wf = 100;
+    s_waitFlush = wf;
     if (!m_handler) {
 	m_handler = new ExtModHandler("call.execute",s_cfg.getIntValue("general","priority",100));
 	Engine::install(m_handler);
