@@ -30,6 +30,8 @@
 // For SpanDSP we have to ask for various C99 stuff
 #define __STDC_LIMIT_MACROS
 
+#define SPANDSP_EXPOSE_INTERNAL_STRUCTURES
+
 #include <math.h>
 #include <stdint.h>
 #include <stdlib.h>
@@ -45,6 +47,15 @@
 #include <fcntl.h>
 #include <errno.h>
 
+#ifdef SPANDSP_PRE006
+#define fax_get_t30_state(x) (&(x)->t30_state)
+#define t38_get_t38_state(x) (&(x)->t38)
+#define t38_get_t30_state(x) (&(x)->t30_state)
+#else
+#define t38_get_t38_state(x) (&(x)->t38_fe.t38)
+#define t38_get_t30_state(x) (&(x)->t30)
+#endif
+
 using namespace TelEngine;
 
 namespace { // anonymous
@@ -59,7 +70,7 @@ class FaxThread : public Thread
 {
 public:
     inline FaxThread(FaxWrapper* wrapper)
-	: Thread("Fax"), m_wrap(wrapper)
+	: Thread("Fax Wrapper"), m_wrap(wrapper)
 	{ }
     virtual void run();
 private:
@@ -80,7 +91,7 @@ class FaxConsumer : public DataConsumer
 public:
     FaxConsumer(FaxWrapper* wrapper, const char* format = "slin");
     ~FaxConsumer();
-    virtual void Consume(const DataBlock& data, unsigned long tStamp);
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags);
 private:
     RefPointer<FaxWrapper> m_wrap;
 };
@@ -108,6 +119,7 @@ protected:
     FaxWrapper();
     void init(t30_state_t* t30, const char* ident, const char* file, bool sender);
     String m_name;
+    String m_error;
     t30_state_t* m_t30;
     FaxSource* m_source;
     FaxConsumer* m_consumer;
@@ -155,13 +167,17 @@ private:
 // A channel (terminal) that sends or receives a local TIFF file
 class FaxChan : public Channel
 {
+    friend class FaxWrapper;
+    YCLASS(FaxChan,Channel)
 public:
     FaxChan(bool outgoing, const char *file, bool sender, Message& msg);
     virtual ~FaxChan();
     virtual bool msgAnswered(Message& msg);
     void answer(const char* targetid);
-    inline const String& ident() const
-	{ return m_ident; }
+    inline const String& localId() const
+	{ return m_localId; }
+    inline const String& remoteId() const
+	{ return m_remoteId; }
     inline bool isSender() const
 	{ return m_sender; }
     inline bool isCaller() const
@@ -169,10 +185,13 @@ public:
 private:
     bool startup(FaxWrapper* wrap, const char* type = "audio");
     bool startup(bool digital = false);
-    String m_ident;
+    void updateInfo(t30_state_t* t30);
+    String m_localId;
+    String m_remoteId;
     bool m_sender;
     bool m_caller;
     bool m_ecm;
+    int m_pages;
 };
 
 class FaxHandler : public MessageHandler
@@ -234,24 +253,27 @@ FaxConsumer::~FaxConsumer()
     m_wrap = 0;
 }
 
-void FaxConsumer::Consume(const DataBlock& data, unsigned long tStamp)
+unsigned long FaxConsumer::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 {
     if (data.null() || !m_wrap)
-	return;
+	return 0;
     m_wrap->rxData(data,tStamp);
+    return invalidStamp();
 }
 
 
-static void phase_b_handler(t30_state_t* s, void* user_data, int result)
+static int phase_b_handler(t30_state_t* s, void* user_data, int result)
 {
     if (user_data)
 	static_cast<FaxWrapper*>(user_data)->phaseB(result);
+    return T30_ERR_OK;
 }
 
-static void phase_d_handler(t30_state_t* s, void* user_data, int result)
+static int phase_d_handler(t30_state_t* s, void* user_data, int result)
 {
     if (user_data)
 	static_cast<FaxWrapper*>(user_data)->phaseD(result);
+    return T30_ERR_OK;
 }
 
 static void phase_e_handler(t30_state_t* s, void* user_data, int result)
@@ -262,7 +284,7 @@ static void phase_e_handler(t30_state_t* s, void* user_data, int result)
 
 
 FaxWrapper::FaxWrapper()
-    : Mutex(true),
+    : Mutex(true,"FaxWrapper"),
       m_t30(0), m_source(0), m_consumer(0), m_chan(0), m_eof(false)
 {
     debugChain(&plugin);
@@ -299,9 +321,7 @@ void FaxWrapper::debugName(const char* name)
 // Initialize terminal T.30 state
 void FaxWrapper::init(t30_state_t* t30, const char* ident, const char* file, bool sender)
 {
-    if (!ident)
-	ident = "anonymous";
-    t30_set_local_ident(t30,ident);
+    t30_set_tx_ident(t30,c_safe(ident));
     t30_set_phase_e_handler(t30,phase_e_handler,this);
     t30_set_phase_d_handler(t30,phase_d_handler,this);
     t30_set_phase_b_handler(t30,phase_b_handler,this);
@@ -341,7 +361,7 @@ bool FaxWrapper::startup(CallEndpoint* chan)
 void FaxWrapper::cleanup()
 {
     if (m_chan && (m_source || m_consumer))
-	m_chan->disconnect();
+	m_chan->disconnect(m_error);
 }
 
 // Called on intermediate states
@@ -356,33 +376,23 @@ void FaxWrapper::phaseD(int result)
 {
     Debug(this,DebugInfo,"Phase D code 0x%X: %s [%p]",
 	result,t30_completion_code_to_str(result),this);
-
-    t30_stats_t t;
-    char ident[21];
-
-    t30_get_transfer_statistics(t30(), &t);
-    Debug(this,DebugAll,"bit rate %d", t.bit_rate);
-    Debug(this,DebugAll,"pages transferred %d", t.pages_transferred);
-    Debug(this,DebugAll,"image size %d x %d", t.width, t.length);
-    Debug(this,DebugAll,"image resolution %d x %d", t.x_resolution, t.y_resolution);
-    Debug(this,DebugAll,"bad rows %d", t.bad_rows);
-    Debug(this,DebugAll,"longest bad row run %d", t.longest_bad_row_run);
-    Debug(this,DebugAll,"compression type %d", t.encoding);
-    Debug(this,DebugAll,"image size %d", t.image_size);
-
-    t30_get_local_ident(t30(), ident);
-    Debug(this,DebugAll,"local ident '%s'", ident);
-
-    t30_get_far_ident(t30(), ident);
-    Debug(this,DebugAll,"remote ident '%s'", ident);
+    FaxChan* chan = YOBJECT(FaxChan,m_chan);
+    if (chan)
+	chan->updateInfo(t30());
 }
 
 // Called to report end of transfer
 void FaxWrapper::phaseE(int result)
 {
+    const char* err = t30_completion_code_to_str(result);
     Debug(this,DebugInfo,"Phase E code 0x%X: %s [%p]",
-	result,t30_completion_code_to_str(result),this);
+	result,err,this);
+    if (T30_ERR_OK != result)
+	m_error = err;
     m_eof = true;
+    FaxChan* chan = YOBJECT(FaxChan,m_chan);
+    if (chan)
+	chan->updateInfo(t30());
 }
 
 
@@ -390,12 +400,12 @@ void FaxWrapper::phaseE(int result)
 FaxTerminal::FaxTerminal(const char *file, const char *ident, bool sender, bool iscaller)
     : m_lastr(0)
 {
-    Debug(this,DebugAll,"FaxTerminal::FaxTerminal(%s %s \"%s\") [%p]",
+    Debug(this,DebugAll,"FaxTerminal::FaxTerminal(%s %s '%s','%s') [%p]",
 	(iscaller ? "caller" : "called"),
 	(sender ? "transmit" : "receive"),
-	file,this);
+	file,ident,this);
     fax_init(&m_fax,iscaller);
-    init(&m_fax.t30_state,ident,file,sender);
+    init(fax_get_t30_state(&m_fax),ident,file,sender);
     fax_set_transmit_on_idle(&m_fax,1);
 }
 
@@ -416,8 +426,8 @@ void FaxTerminal::run()
 
 	tpos += ((u_int64_t)1000000*r/16000);
 	int64_t dly = tpos - Time::now();
-	if (dly > 10000)
-	    dly = 10000;
+	if (dly > 30000)
+	    dly = 30000;
 	if (dly > 0)
 	    Thread::usleep(dly,true);
     }
@@ -467,13 +477,13 @@ void FaxTerminal::rxData(const DataBlock& data, unsigned long tStamp)
 // Constructor for the digital fax terminal
 T38Terminal::T38Terminal(const char *file, const char *ident, bool sender, bool iscaller)
 {
-    Debug(this,DebugAll,"T38Terminal::T38Terminal(%s %s \"%s\") [%p]",
+    Debug(this,DebugAll,"T38Terminal::T38Terminal(%s %s '%s','%s') [%p]",
 	(iscaller ? "caller" : "called"),
 	(sender ? "transmit" : "receive"),
-	file,this);
+	file,ident,this);
     t38_terminal_init(&m_t38,iscaller,txHandler,this);
-    t38_set_t38_version(&m_t38.t38,1);
-    init(&m_t38.t30_state,ident,file,sender);
+    t38_set_t38_version(t38_get_t38_state(&m_t38),1);
+    init(t38_get_t30_state(&m_t38),ident,file,sender);
 }
 
 T38Terminal::~T38Terminal()
@@ -501,7 +511,7 @@ int T38Terminal::txHandler(t38_core_state_t* t38s, void* userData,
 void T38Terminal::rxData(const DataBlock& data, unsigned long tStamp)
 {
     Debug(this,DebugStub,"Please implement T38Terminal::rxData()");
-    t38_core_rx_ifp_packet(&m_t38.t38,tStamp,(uint8_t*)data.data(),data.length());
+    t38_core_rx_ifp_packet(t38_get_t38_state(&m_t38),(uint8_t*)data.data(),data.length(),tStamp);
 }
 
 int T38Terminal::txData(const void* buf, int len, int seq, int count)
@@ -520,20 +530,20 @@ void FaxThread::run()
 
 // Constructor for a generic fax terminal channel
 FaxChan::FaxChan(bool outgoing, const char *file, bool sender, Message& msg)
-    : Channel(plugin,0,outgoing), m_sender(sender)
+    : Channel(plugin,0,outgoing), m_sender(sender), m_pages(0)
 {
     Debug(this,DebugAll,"FaxChan::FaxChan(%s \"%s\") [%p]",
 	(sender ? "transmit" : "receive"),
 	file,this);
-    const char* ident = msg.getValue("faxident",msg.getValue("caller"));
-    if (!ident)
-	ident = "anonymous";
-    m_ident = ident;
+    m_localId = msg.getValue("faxident",msg.getValue(outgoing ? "called" : "caller"));
     // outgoing means from Yate to file so the fax should answer by default
     m_caller = msg.getBoolValue("faxcaller",!outgoing);
-    m_ecm = msg.getBoolValue("faxecm");
+    m_ecm = msg.getBoolValue("faxecm",true);
     m_address = file;
-    Engine::enqueue(message("chan.startup",msg));
+    Message* s = message("chan.startup",msg);
+    if (outgoing)
+	s->copyParams(msg,"called,caller,callername,billid");
+    Engine::enqueue(s);
 }
 
 // Destructor - clears all (audio, image) endpoints early
@@ -541,7 +551,16 @@ FaxChan::~FaxChan()
 {
     Debug(DebugAll,"FaxChan::~FaxChan() [%p]",this);
     clearEndpoint();
-    Engine::enqueue(message("chan.hangup"));
+    Message* msg = message("chan.hangup");
+    if (m_localId)
+	msg->addParam("faxident_local",m_localId);
+    if (m_remoteId)
+	msg->addParam("faxident_remote",m_remoteId);
+    if (m_pages)
+	msg->addParam("faxpages",String(m_pages));
+    msg->addParam("faxecm",String::boolText(m_ecm));
+    msg->addParam("faxcaller",String::boolText(m_caller));
+    Engine::enqueue(msg);
 }
 
 // Build data channels, attaches a wrapper and starts it up
@@ -564,9 +583,9 @@ bool FaxChan::startup(FaxWrapper* wrap, const char* type)
 bool FaxChan::startup(bool digital)
 {
     if (digital)
-	return startup(new T38Terminal(address(),m_ident,m_sender,m_caller),"image");
+	return startup(new T38Terminal(address(),m_localId,m_sender,m_caller),"image");
     else
-	return startup(new FaxTerminal(address(),m_ident,m_sender,m_caller));
+	return startup(new FaxTerminal(address(),m_localId,m_sender,m_caller));
 }
 
 // Handler for an originator fax start request
@@ -589,6 +608,35 @@ void FaxChan::answer(const char* targetid)
     Engine::enqueue(message("call.answered"));
 }
 
+void FaxChan::updateInfo(t30_state_t* t30)
+{
+    const char* ident = t30_get_rx_ident(t30);
+    if (!null(ident))
+	m_remoteId = ident;
+
+    t30_stats_t t;
+    t30_get_transfer_statistics(t30, &t);
+    if (!t.error_correcting_mode)
+	m_ecm = false;
+#ifdef SPANDSP_TXRXSTATS
+    m_pages = t.pages_tx + t.pages_rx;
+#else
+    m_pages = t.pages_transferred;
+#endif
+
+    Debug(this,DebugAll,"bit rate %d", t.bit_rate);
+    Debug(this,DebugAll,"error correction %d", t.error_correcting_mode);
+    Debug(this,DebugAll,"pages transferred %d", m_pages);
+    Debug(this,DebugAll,"image size %d x %d", t.width, t.length);
+    Debug(this,DebugAll,"image resolution %d x %d", t.x_resolution, t.y_resolution);
+    Debug(this,DebugAll,"bad rows %d", t.bad_rows);
+    Debug(this,DebugAll,"longest bad row run %d", t.longest_bad_row_run);
+    Debug(this,DebugAll,"compression type %d", t.encoding);
+    Debug(this,DebugAll,"image size %d", t.image_size);
+
+    Debug(this,DebugAll,"local ident '%s'", t30_get_tx_ident(t30));
+    Debug(this,DebugAll,"remote ident '%s'", ident);
+}
 
 bool FaxDriver::msgExecute(Message& msg, String& dest)
 {
