@@ -30,6 +30,7 @@
 
 #include <ptlib.h>
 #include <h323.h>
+#include <h323neg.h>
 #include <h323pdu.h>
 #include <h323caps.h>
 #include <ptclib/delaychan.h>
@@ -94,18 +95,18 @@ static int s_maxCleaning = 0;
 
 static Configuration s_cfg;
 
-static Mutex s_mutex;
+static Mutex s_mutex(false,"H323Chan");
 static int s_connCount = 0;
 static int s_chanCount = 0;
 
-static TokenDict dict_str2code[] = {
+static const TokenDict dict_str2code[] = {
     { "alpha" , PProcess::AlphaCode },
     { "beta" , PProcess::BetaCode },
     { "release" , PProcess::ReleaseCode },
     { 0 , 0 },
 };
 
-const char* h323_formats[] = {
+static const char* h323_formats[] = {
     "G.711-ALaw-64k", "alaw",
     "G.711-uLaw-64k", "mulaw",
     "GSM-06.10", "gsm",
@@ -117,11 +118,11 @@ const char* h323_formats[] = {
     "G.723", "g723",
     "G.726", "g726",
     "G.728", "g728",
+    "G.729B", "g729b",
     "G.729", "g729",
     "PCM-16", "slin",
 #if 0
     "G.729A", "g729a",
-    "G.729B", "g729b",
     "G.729A/B", "g729ab",
     "G.723.1", "g723.1",
     "G.723.1(5.3k)", "g723.1-5k3",
@@ -142,21 +143,22 @@ const char* h323_formats[] = {
     0
 };
 
-static TokenDict dict_h323_dir[] = {
+static const TokenDict dict_h323_dir[] = {
     { "receive", H323Channel::IsReceiver },
     { "send", H323Channel::IsTransmitter },
     { "bidir", H323Channel::IsBidirectional },
     { 0 , 0 },
 };
 
-static TokenDict dict_silence[] = {
+static const TokenDict dict_silence[] = {
     { "none", H323AudioCodec::NoSilenceDetection },
     { "fixed", H323AudioCodec::FixedSilenceDetection },
     { "adaptive", H323AudioCodec::AdaptiveSilenceDetection },
     { 0 , 0 },
 };
 
-static TokenDict dict_errors[] = {
+// OpenH323 cause code mappings
+static const TokenDict dict_errors[] = {
     { "noroute", H323Connection::EndedByUnreachable },
     { "noroute", H323Connection::EndedByNoUser },
     { "noconn", H323Connection::EndedByNoEndPoint },
@@ -174,6 +176,18 @@ static TokenDict dict_errors[] = {
     { 0 , 0 },
 };
 
+// Q.931/Q.850 cause code mappings
+static const TokenDict q931_errors[] = {
+    { "incomplete", Q931::InvalidNumberFormat },
+    { "congestion", Q931::NoCircuitChannelAvailable },
+    { "congestion", Q931::TemporaryFailure },
+    { "congestion", Q931::Congestion },
+    { "offline", Q931::SubscriberAbsent },
+    { "nocall", Q931::InvalidCallReference },
+    { "nocall", Q931::IdentifiedChannelNonExistent },
+    { 0 , 0 },
+};
+
 static const char* CallEndReasonText(int reason)
 {
 #define MAKE_END_REASON(r) case H323Connection::r: return #r
@@ -247,8 +261,9 @@ public:
     H323Driver();
     virtual ~H323Driver();
     virtual void initialize();
-    virtual bool msgExecute(Message& msg, String& dest);
+    virtual bool hasLine(const String& line) const;
     virtual bool msgRoute(Message& msg);
+    virtual bool msgExecute(Message& msg, String& dest);
     virtual void msgTimer(Message& msg);
     virtual bool received(Message &msg, int id);
     virtual void statusParams(String& str);
@@ -312,24 +327,23 @@ private:
     bool m_exit;
 };
 
-class YateH323AudioConsumer : public DataConsumer, public PIndirectChannel
+class YateH323AudioConsumer : public DataConsumer, public Mutex, public PIndirectChannel
 {
     PCLASSINFO(YateH323AudioConsumer, PIndirectChannel)
     YCLASS(YateH323AudioConsumer, DataConsumer)
 public:
     YateH323AudioConsumer()
-	: m_exit(false)
+	: Mutex(false,"YateH323AudioConsumer"), m_exit(false)
 	{ Debug(&hplugin,DebugAll,"YateH323AudioConsumer::YateH323AudioConsumer() [%p]",this); }
     ~YateH323AudioConsumer();
     virtual BOOL Close(); 
     virtual BOOL IsOpen() const;
     virtual BOOL Read(void *buf, PINDEX len);
-    virtual void Consume(const DataBlock &data, unsigned long tStamp);
+    virtual unsigned long Consume(const DataBlock &data, unsigned long tStamp, unsigned long flags);
 private:
     PAdaptiveDelay readDelay;
     DataBlock m_buffer;
     bool m_exit;
-    Mutex m_mutex;
 };
 
 class YateH323_ExternalRTPChannel;
@@ -397,6 +411,8 @@ public:
     void setCallerID(const char* number, const char* name);
     void rtpExecuted(Message& msg);
     void rtpForward(Message& msg, bool init = false);
+    void updateFormats(const Message& msg);
+    bool adjustCapabilities();
     void answerCall(AnswerCallResponse response, bool autoEarly = false);
     static BOOL decodeCapability(const H323Capability& capability, const char** dataFormat, int* payload = 0, String* capabName = 0);
     inline bool hasRemoteAddress() const
@@ -474,6 +490,7 @@ public:
 	{ m_targetid = targetid; }
 private:
     YateH323Connection* m_conn;
+    H323Connection::CallEndReason m_reason;
     bool m_hungup;
     bool m_inband;
 };
@@ -611,6 +628,7 @@ DEFINE_YATE_CAPAB(YateG729AB,BaseG729Capab,H245_AudioCapability::e_g729AnnexAwAn
 
 // end of fake capabilities code
 
+#ifndef DISABLE_CAPS_DUMP
 #ifdef USE_CAPABILITY_FACTORY
 static void ListRegisteredCaps(int level)
 {
@@ -651,7 +669,7 @@ bool FakeH323CapabilityRegistration::IsRegistered(const PString& name)
     return false;
 }
 #endif
-
+#endif // DISABLE_CAPS_DUMP
 
 YateGatekeeperServer::YateGatekeeperServer(YateH323EndPoint& ep)
     : H323GatekeeperServer(ep), m_endpoint(ep)
@@ -727,6 +745,7 @@ H323Connection* YateH323EndPoint::CreateConnection(unsigned callReference,
 
 bool YateH323EndPoint::Init(const NamedList* params)
 {
+#ifndef DISABLE_CAPS_DUMP
     if (null()) {
 	int dump = s_cfg.getIntValue("general","dumpcodecs");
 	if (dump > 0)
@@ -736,6 +755,7 @@ bool YateH323EndPoint::Init(const NamedList* params)
 	    FakeH323CapabilityRegistration::ListRegistered(dump);
 #endif
     }
+#endif
 
     String csect("codecs");
     if (!null()) {
@@ -787,6 +807,8 @@ bool YateH323EndPoint::Init(const NamedList* params)
 	 : H323AudioCodec::NoSilenceDetection));
 
     PIPSocket::Address addr = INADDR_ANY;
+    if (params && params->getValue("addr"))
+	addr = params->getValue("addr");
     int port = 1720;
     if (params)
 	port = params-> getIntValue("port",port);
@@ -1004,6 +1026,7 @@ YateH323Connection::YateH323Connection(YateH323EndPoint& endpoint,
 
     setCallerID(msg->getValue("caller"),msg->getValue("callername"));
     rtpForward(*msg,s_passtrough);
+    updateFormats(*msg);
     m_needMedia = msg->getBoolValue("needmedia",m_needMedia);
 
     CallEndpoint* ch = YOBJECT(CallEndpoint,msg->userData());
@@ -1169,6 +1192,68 @@ void YateH323Connection::rtpForward(Message& msg, bool init)
     }
 }
 
+// Update the formats when RTP is proxied
+void YateH323Connection::updateFormats(const Message& msg)
+{
+    // when doing RTP forwarding formats are altered in rtpForward()
+    if (m_passtrough)
+	return;
+    // only audio is currently supported
+    const char* formats = msg.getValue("formats");
+    if (!formats)
+	return;
+    if (m_formats != formats) {
+	Debug(this,DebugNote,"Formats changed to '%s'",formats);
+	m_formats = formats;
+	// send changed capability set only if another was already sent
+	if (adjustCapabilities() && capabilityExchangeProcedure->HasSentCapabilities())
+	    SendCapabilitySet(FALSE);
+    }
+}
+
+// Adjust local capabilities to not exceed the format list
+bool YateH323Connection::adjustCapabilities()
+{
+    if (m_formats.null())
+	return false;
+    // remote has a list of supported codecs - remove unsupported capabilities
+    bool nocodecs = true;
+    bool changed = false;
+    for (int i = 0; i < localCapabilities.GetSize(); i++) {
+	const char* format = 0;
+	String fname;
+	decodeCapability(localCapabilities[i],&format,0,&fname);
+	if (format) {
+	    if (m_formats.find(format) < 0) {
+		Debug(this,DebugAll,"Removing capability '%s' (%s) not in remote '%s'",
+		    fname.c_str(),format,m_formats.c_str());
+		changed = true;
+		// also remove any matching fast start channels
+		for (PINDEX idx = 0; idx < fastStartChannels.GetSize(); idx++) {
+		    if (fastStartChannels[idx].GetCapability() == localCapabilities[i]) {
+			Debug(this,DebugInfo,"Removing fast start channel %s '%s' (%s)",
+			    lookup(fastStartChannels[idx].GetDirection(),dict_h323_dir,"?"),
+			    fname.c_str(),format);
+			fastStartChannels.RemoveAt(idx--);
+		    }
+		}
+		localCapabilities.Remove(fname.c_str());
+		i--;
+	    }
+	    else
+		nocodecs = false;
+	}
+    }
+    if (nocodecs) {
+	Debug(DebugWarn,"No codecs remaining for H323 connection [%p]",this);
+	if (m_needMedia) {
+	    changed = false;
+	    ClearCall(EndedByCapabilityExchange);
+	}
+    }
+    return changed;
+}
+
 void YateH323Connection::answerCall(AnswerCallResponse response, bool autoEarly)
 {
     bool media = false;
@@ -1239,7 +1324,9 @@ void YateH323Connection::OnCleared()
 {
     int reason = GetCallEndReason();
     const char* rtext = CallEndReasonText(reason);
-    const char* error = lookup(reason,dict_errors);
+    const char* error = lookup(GetQ931Cause(),q931_errors);
+    if (!error)
+	error = lookup(reason,dict_errors);
     Debug(this,DebugInfo,"YateH323Connection::OnCleared() error: '%s' reason: %s (%d) [%p]",
 	error,rtext,reason,this);
     TelEngine::Lock lock(m_mutex);
@@ -1303,7 +1390,8 @@ void YateH323Connection::OnUserInputTone(char tone, unsigned duration, unsigned
     buf[1] = 0;
     m->addParam("text",buf);
     m->addParam("duration",String(duration));
-    Engine::enqueue(m);
+    m->addParam("detected","h323");
+    m_chan->dtmfEnqueue(m);
 }
 
 void YateH323Connection::OnUserInputString(const PString &value)
@@ -1368,7 +1456,8 @@ H323Channel* YateH323Connection::CreateRealTimeLogicalChannel(const H323Capabili
 	}
 	PIPSocket::Address externalIpAddress;
 	GetControlChannel().GetLocalAddress().GetIpAddress(externalIpAddress);
-	Debug(this,DebugInfo,"address '%s'",(const char *)externalIpAddress.AsString());
+	Debug(this,DebugAll,"Logical control channel address '%s'",
+	    (const char *)externalIpAddress.AsString());
 	WORD externalPort = 0;
 	if (!m_passtrough) {
 	    TelEngine::Lock lock(m_mutex);
@@ -1385,7 +1474,7 @@ H323Channel* YateH323Connection::CreateRealTimeLogicalChannel(const H323Capabili
 		}
 	    }
 	    else {
-		Debug(this,DebugInfo,"Not creating logical channel for a dead channel [%p]",this);
+		Debug(this,DebugNote,"Not creating logical channel for a dead channel [%p]",this);
 		return 0;
 	    }
 	}
@@ -1418,30 +1507,7 @@ void YateH323Connection::OnSetLocalCapabilities()
     Debug(this,DebugAll,"YateH323Connection::OnSetLocalCapabilities()%s%s [%p]",
 	m_externalRtp ? " external" : "",m_passtrough ? " passtrough" : "",this);
     H323Connection::OnSetLocalCapabilities();
-    if (m_formats.null())
-	return;
-    // remote has a list of supported codecs - remove unsupported capabilities
-    bool nocodecs = true;
-    for (int i = 0; i < localCapabilities.GetSize(); i++) {
-	const char* format = 0;
-	String fname;
-	decodeCapability(localCapabilities[i],&format,0,&fname);
-	if (format) {
-	    if (m_formats.find(format) < 0) {
-		Debug(this,DebugAll,"Removing capability '%s' (%s) not in remote '%s'",
-		    fname.c_str(),format,m_formats.c_str());
-		localCapabilities.Remove(fname.c_str());
-		i--;
-	    }
-	    else
-		nocodecs = false;
-	}
-    }
-    if (nocodecs) {
-	Debug(DebugWarn,"No codecs remaining for H323 connection [%p]",this);
-	if (m_needMedia)
-	    ClearCall(EndedByCapabilityExchange);
-    }
+    adjustCapabilities();
 }
 
 BOOL YateH323Connection::OnStartLogicalChannel(H323Channel & channel) 
@@ -1502,7 +1568,8 @@ BOOL YateH323Connection::decodeCapability(const H323Capability& capability, cons
 void YateH323Connection::setRemoteAddress(const char* remoteIP, WORD remotePort)
 {
     if (!m_remotePort) {
-	Debug(this,DebugInfo,"Copying remote RTP address [%p]",this);
+	Debug(this,DebugInfo,"Got remote RTP address %s:%u [%p]",
+	    remoteIP,remotePort,this);
 	m_remotePort = remotePort;
 	m_remoteAddr = remoteIP;
     }
@@ -1513,6 +1580,14 @@ BOOL YateH323Connection::startExternalRTP(const char* remoteIP, WORD remotePort,
     const char* sdir = lookup(dir,dict_h323_dir);
     Debug(this,DebugAll,"YateH323Connection::startExternalRTP(\"%s\",%u,%s,%p) [%p]",
 	remoteIP,remotePort,sdir,chan,this);
+    int payload = 128;
+    const char *format = 0;
+    decodeCapability(chan->GetCapability(),&format,&payload);
+    if (format && m_formats && (m_formats.find(format) < 0)) {
+	Debug(this,DebugNote,"Refusing RTP '%s' payload %d, not in '%s'",
+	    format,payload,m_formats.c_str());
+	return FALSE;
+    }
     if (m_passtrough && m_rtpPort) {
 	setRemoteAddress(remoteIP,remotePort);
 
@@ -1532,9 +1607,6 @@ BOOL YateH323Connection::startExternalRTP(const char* remoteIP, WORD remotePort,
 	m.addParam("direction",sdir);
     m.addParam("remoteip",remoteIP);
     m.addParam("remoteport",String(remotePort));
-    int payload = 128;
-    const char *format = 0;
-    decodeCapability(chan->GetCapability(),&format,&payload);
     if (format)
 	m.addParam("format",format);
     if ((payload >= 0) && (payload < 127))
@@ -1620,7 +1692,7 @@ BOOL YateH323_ExternalRTPChannel::Start()
     PIPSocket::Address remoteIpAddress;
     WORD remotePort;
     GetRemoteAddress(remoteIpAddress,remotePort);
-    Debug(&hplugin,DebugInfo,"external rtp ip address %s:%u",(const char *)remoteIpAddress.AsString(),remotePort);
+    Debug(&hplugin,DebugInfo,"External RTP address %s:%u",(const char *)remoteIpAddress.AsString(),remotePort);
 
     return isRunning = m_conn->startExternalRTP((const char *)remoteIpAddress.AsString(), remotePort, GetDirection(), this);
 }
@@ -1637,7 +1709,7 @@ BOOL YateH323_ExternalRTPChannel::OnReceivedPDU(
     PIPSocket::Address remoteIpAddress;
     WORD remotePort;
     GetRemoteAddress(remoteIpAddress,remotePort);
-    Debug(&hplugin,DebugInfo,"Remote RTP IP address %s:%u",(const char *)remoteIpAddress.AsString(),remotePort);
+    Debug(&hplugin,DebugAll,"Remote RTP address %s:%u",(const char *)remoteIpAddress.AsString(),remotePort);
     m_conn->setRemoteAddress((const char *)remoteIpAddress.AsString(), remotePort);
     return TRUE;
 }
@@ -1665,9 +1737,9 @@ YateH323AudioSource::~YateH323AudioSource()
     DDebug(&hplugin,DebugAll,"YateH323AudioSource::~YateH323AudioSource() [%p]",this);
     m_exit = true;
     // Delay actual destruction until the mutex is released
-    m_mutex.lock();
+    lock();
     m_data.clear(false);
-    m_mutex.unlock();
+    unlock();
 }
 
 YateH323AudioConsumer::~YateH323AudioConsumer()
@@ -1675,7 +1747,7 @@ YateH323AudioConsumer::~YateH323AudioConsumer()
     DDebug(&hplugin,DebugAll,"YateH323AudioConsumer::~YateH323AudioConsumer() [%p]",this);
     m_exit = true;
     // Delay actual destruction until the mutex is released
-    m_mutex.check();
+    check();
 }
 
 BOOL YateH323AudioConsumer::Close()
@@ -1690,24 +1762,27 @@ BOOL YateH323AudioConsumer::IsOpen() const
     return !m_exit;
 }
 
-void YateH323AudioConsumer::Consume(const DataBlock &data, unsigned long tStamp)
+unsigned long YateH323AudioConsumer::Consume(const DataBlock &data, unsigned long tStamp, unsigned long flags)
 {
     if (m_exit)
-	return;
-    Lock lock(m_mutex);
-    if ((m_buffer.length() + data.length()) <= (480*5))
+	return 0;
+    Lock lock(this);
+    if ((m_buffer.length() + data.length()) <= (480*5)) {
 	m_buffer += data;
+	return invalidStamp();
+    }
 #ifdef DEBUG
     else
 	Debug(&hplugin,DebugAll,"Consumer skipped %u bytes, buffer is full [%p]",data.length(),this);
 #endif
+    return 0;
 }
 
 BOOL YateH323AudioConsumer::Read(void *buf, PINDEX len)
 {
     readDelay.Delay(len/16);
     while (!m_exit) {
-	Lock lock(m_mutex);
+	Lock lock(this);
 	if (!getConnSource()) {
 	    ::memset(buf,0,len);
 	    break;
@@ -1941,7 +2016,9 @@ void YateH323Connection::setCallerID(const char* number, const char* name)
 }
 
 YateH323Chan::YateH323Chan(YateH323Connection* conn,Message* msg,const char* addr)
-    : Channel(hplugin,0,(msg != 0)), m_conn(conn), m_hungup(false), m_inband(s_inband)
+    : Channel(hplugin,0,(msg != 0)),
+      m_conn(conn), m_reason(H323Connection::EndedByLocalUser),
+      m_hungup(false), m_inband(s_inband)
 {
     s_mutex.lock();
     s_chanCount++;
@@ -2015,18 +2092,20 @@ void YateH323Chan::hangup(bool dropChan, bool clearCall)
     m_conn = 0;
     if (clearCall && tmp) {
 	const char* err = 0;
-	const char* txt = "Normal cleanup";
-	int reason = tmp->GetCallEndReason();
-	if (reason != H323Connection::NumCallEndReasons) {
+	H323Connection::CallEndReason reason = tmp->GetCallEndReason();
+	if (reason == H323Connection::NumCallEndReasons)
+	    reason = m_reason;
+	else
+	    err = lookup(tmp->GetQ931Cause(),q931_errors);
+	if (!err)
 	    err = lookup(reason,dict_errors);
-	    txt = CallEndReasonText(reason);
-	}
+	const char* txt = CallEndReasonText(reason);
 	if (err)
 	    m->setParam("error",err);
 	if (txt)
 	    m->setParam("reason",txt);
 	tmp->cleanups(false,dropChan);
-	tmp->ClearCall();
+	tmp->ClearCall(reason);
     }
     Engine::enqueue(m);
 }
@@ -2034,13 +2113,13 @@ void YateH323Chan::hangup(bool dropChan, bool clearCall)
 void YateH323Chan::disconnected(bool final, const char *reason)
 {
     Debugger debug("YateH323Chan::disconnected()"," '%s' [%p]",reason,this);
-    if (!final) {
-	Channel::disconnected(final,reason);
+    Channel::disconnected(final,reason);
+    m_reason = (H323Connection::CallEndReason)lookup(reason,dict_errors,H323Connection::EndedByLocalUser);
+    if (!final)
 	return;
-    }
     stopDataLinks();
     if (m_conn)
-	m_conn->ClearCall((H323Connection::CallEndReason)lookup(reason,dict_errors,H323Connection::EndedByLocalUser));
+	m_conn->ClearCall(m_reason);
 }
 
 // Set the signalling address
@@ -2109,6 +2188,7 @@ bool YateH323Chan::callRouted(Message& msg)
 	    m_conn->Unlock();
 	    return false;
 	}
+	m_conn->updateFormats(msg);
 	return true;
     }
     return false;
@@ -2119,6 +2199,7 @@ void YateH323Chan::callAccept(Message& msg)
     Channel::callAccept(msg);
     if (m_conn) {
 	m_conn->rtpExecuted(msg);
+	m_conn->updateFormats(msg);
 	m_conn->answerCall(H323Connection::AnswerCallDeferred);
     }
 }
@@ -2138,6 +2219,7 @@ bool YateH323Chan::msgProgress(Message& msg)
 	return false;
     if (msg.getParam("rtp_forward"))
 	m_conn->rtpForward(msg);
+    m_conn->updateFormats(msg);
     m_conn->answerCall(H323Connection::AnswerCallDeferred,msg.getBoolValue("earlymedia",true));
     return true;
 }
@@ -2149,6 +2231,7 @@ bool YateH323Chan::msgRinging(Message& msg)
 	return false;
     if (msg.getParam("rtp_forward"))
 	m_conn->rtpForward(msg);
+    m_conn->updateFormats(msg);
     m_conn->answerCall(H323Connection::AnswerCallPending,msg.getBoolValue("earlymedia",true));
     return true;
 }
@@ -2158,6 +2241,7 @@ bool YateH323Chan::msgAnswered(Message& msg)
     if (!m_conn)
 	return false;
     m_conn->rtpForward(msg);
+    m_conn->updateFormats(msg);
     m_conn->answerCall(H323Connection::AnswerCallNow);
     return true;
 }
@@ -2252,22 +2336,18 @@ void H323Driver::statusParams(String& str)
     Driver::statusParams(str);
     str.append("cleaning=",",") << cleaningCount();
 }
+
+bool H323Driver::hasLine(const String& line) const
+{
+    return line && hplugin.findEndpoint(line);
+}
     
 bool H323Driver::msgRoute(Message& msg)
 {
-    String called = msg.getValue("called");
-    if (called.null() || (called.find('@') >= 0))
+    String* called = msg.getParam("called");
+    if (!called || (called->find('@') >= 0))
 	return false;
-    String line = msg.getValue("line");
-    if (line.null())
-	line = msg.getValue("account");
-    if (line && hplugin.findEndpoint(line)) {
-	// asked to route to a line we have locally
-	msg.setParam("line",line);
-	msg.retValue() = prefix() + called;
-	return true;
-    }
-    return false;
+    return Driver::msgRoute(msg);
 }
 
 bool H323Driver::msgExecute(Message& msg, String& dest)
