@@ -93,6 +93,7 @@ static const CommandInfo s_cmdInfo[] =
     { "help", "[command]", 0, "Provide help on all or given command" },
     { "status", "[overview] [modulename]", s_oview, "Shows status of all or selected modules or channels" },
     { "uptime", 0, 0, "Show information on how long Yate has run" },
+    { "echo", "[on|off]", s_bools, "Show or turn remote echo on or off" },
     { "machine", "[on|off]", s_bools, "Show or turn machine output mode on or off" },
     { "output", "[on|off]", s_bools, "Show or turn local output on or off" },
     { "color", "[on|off]", s_bools, "Show status or turn local colorization on or off" },
@@ -101,8 +102,9 @@ static const CommandInfo s_cmdInfo[] =
 #ifdef HAVE_COREDUMPER
     { "coredump", "[filename]", 0, "Dumps memory image of running Yate to a file" },
 #endif
-    { "drop", "{chan|*|all}", s_dall, "Drops one or all active calls" },
+    { "drop", "{chan|*|all} [reason]", s_dall, "Drops one or all active calls" },
     { "call", "chan target", 0, "Execute an outgoing call" },
+    { "control", "chan [operation] [param=val] [param=...]", 0, "Apply arbitrary control operations to a channel or entity" },
     { "reload", 0, 0, "Reloads module configuration files" },
     { "restart", "[now]", s_rnow, "Restarts the engine if executing supervised" },
     { "stop", "[exitcode]", 0, "Stops the engine with optionally provided exit code" },
@@ -126,7 +128,7 @@ static bool s_output = false;
 
 /* I need this here because i'm gonna use it in both classes */
 Socket s_sock;
-Mutex s_mutex(true);
+Mutex s_mutex(true,"RManager");
 
 //we gonna create here the list with all the new connections.
 static ObjList connectionlist;
@@ -209,7 +211,7 @@ void RManagerThread::run()
 {
     for (;;)
     {
-	Thread::msleep(10,true);
+	Thread::idle(true);
 	SocketAddr sa;
 	Socket* as = s_sock.accept(sa);
 	if (!as) {
@@ -283,15 +285,15 @@ void Connection::run()
 
     Output("Remote connection from %s",m_address.c_str());
     m_auth = !s_cfg.getValue("general","password");
-    const char *hdr = s_cfg.getValue("general","header","YATE (http://YATE.null.ro) ready.");
-    if (hdr) {
-	writeStr(hdr);
-	writeStr("\r\n");
-	hdr = 0;
-    }
+    String hdr = s_cfg.getValue("general","header","YATE ${version}-${release} (http://YATE.null.ro) ready.");
+    Engine::runParams().replaceParams(hdr);
     if (s_cfg.getBoolValue("general","telnet",true)) {
 	// WILL SUPPRESS GO AHEAD, WILL ECHO - and enough BS and blanks to hide them
-	writeStr("\377\373\003\377\373\001\b\b\b\b\b\b      \b\b\b\b\b\b");
+	writeStr("\377\373\003\377\373\001\r      \b\b\b\b\b\b");
+    }
+    if (hdr) {
+	writeStr("\r" + hdr + "\r\n");
+	hdr.clear();
     }
     unsigned char buffer[128];
     for (;;) {
@@ -438,11 +440,12 @@ bool Connection::processChar(unsigned char c)
 	case 0x1B: // ESC
 	    m_escmode = c;
 	    return false;
-	case '\r':
 	case '\n':
 	    m_escmode = 0;
 	    if (m_buffer.null())
 		return false;
+	case '\r':
+	    m_escmode = 0;
 	    if (m_echoing)
 		writeStr("\r\n");
 	    if (processLine(m_buffer))
@@ -468,6 +471,19 @@ bool Connection::processChar(unsigned char c)
 	    m_escmode = 0;
 	    m_echoing = !m_echoing;
 	    return false;
+	case 0x0F: // ^O
+	    m_escmode = 0;
+	    // cycle [no output] -> [output] -> [debug (only if auth)]
+	    if (m_debug)
+		m_output = m_debug = false;
+	    else if (m_output) {
+		m_output = false;
+		if ((m_debug = m_auth))
+		    Debugger::enableOutput(true);
+	    }
+	    else
+		m_output = true;
+	    return false;
 	case 0x0C: // ^L
 	    if (!m_echoing)
 		break;
@@ -546,12 +562,11 @@ bool Connection::processChar(unsigned char c)
 		m_escmode = c;
 		return false;
 	}
-	unsigned char last = m_escmode;
-	m_escmode = 0;
 	DDebug("RManager",DebugInfo,"ANSI '%s%c' last '%s%c'",
 	    (c >= ' ') ? "" : "^", (c >= ' ') ? c : c+0x40,
-	    (last >= ' ') ? "" : "^", (last >= ' ') ? last : last+0x40
+	    (m_escmode >= ' ') ? "" : "^", (m_escmode >= ' ') ? m_escmode : m_escmode+0x40
 	);
+	m_escmode = 0;
 	switch (c) {
 	    case 'A': // Up arrow
 	    case 'B': // Down arrow
@@ -711,8 +726,12 @@ bool Connection::processLine(const char *line)
     DDebug("RManager",DebugInfo,"processLine = '%s'",line);
     String str(line);
     str.trimBlanks();
+    if (str.null())
+	return false;
 
     m_lastcmd = str;
+    line = 0;
+    m_buffer.clear();
 
     if (str.startSkip("status"))
     {
@@ -731,6 +750,14 @@ bool Connection::processLine(const char *line)
 	writeStr(str);
 	return false;
     }
+    else if (str.startSkip("echo"))
+    {
+	str >> m_echoing;
+	str = "Remote echo: ";
+	str += (m_echoing ? "on\r\n" : "off\r\n");
+	writeStr(str);
+	return false;
+    }
     else if (str.startSkip("machine"))
     {
 	str >> m_machine;
@@ -839,6 +866,12 @@ bool Connection::processLine(const char *line)
     }
     if (str.startSkip("drop"))
     {
+	String reason;
+	int pos = str.find(' ');
+	if (pos > 0) {
+	    reason = str.substr(pos+1);
+	    str = str.substr(0,pos);
+	}
 	if (str.null()) {
 	    writeStr(m_machine ? "%%=drop:fail=noarg\r\n" : "You must specify what connection to drop!\r\n");
 	    return false;
@@ -851,6 +884,8 @@ bool Connection::processLine(const char *line)
 	}
 	else
 	    m.addParam("id",str); 
+	if (reason)
+	    m.addParam("reason",reason);
 	if (Engine::dispatch(m))
 	    str = (m_machine ? "%%=drop:success:" : "Dropped ") + str + "\r\n";
 	else if (all)
@@ -926,6 +961,37 @@ bool Connection::processLine(const char *line)
 	}
 	writeStr(str);
     }
+    else if (str.startSkip("control"))
+    {
+	int pos = str.find(' ');
+	String id = str.substr(0,pos).trimBlanks();
+	String ctrl = str.substr(pos+1).trimBlanks();
+	if ((pos <= 0) || id.null() || ctrl.null()) {
+	    writeStr(m_machine ? "%%=control:fail=noarg\r\n" : "You must specify channel and operation!\r\n");
+	    return false;
+	}
+	Message m("chan.control");
+	m.addParam("targetid",id);
+	m.addParam("component",id);
+	Regexp r("^\\(.* \\)\\?\\([^= ]\\+\\)=\\([^=]*\\)$");
+	while (ctrl) {
+	    if (!ctrl.matches(r)) {
+		m.setParam("operation",ctrl);
+		break;
+	    }
+	    m.setParam(ctrl.matchString(2),ctrl.matchString(3).trimBlanks());
+	    ctrl = ctrl.matchString(1).trimBlanks();
+	}
+	if (Engine::dispatch(m)) {
+	    if (m_machine)
+		str = "%%=control:success:" + id + ":" + m.retValue() + "\r\n";
+	    else
+		str = "Control '" + id + "' " + m.retValue().safe("OK") + "\r\n";
+	}
+	else
+	    str = (m_machine ? "%%=control:fail:" : "Could not control ") + str + "\r\n";
+	writeStr(str);
+    }
 #ifdef HAVE_COREDUMPER
     else if (str.startSkip("coredump"))
     {
@@ -1012,6 +1078,8 @@ void Connection::writeDebug(const char *str, int level)
     if (null(str))
 	return;
     if (m_debug || (m_output && (level < 0))) {
+	if (m_echoing && m_buffer)
+	    clearLine();
 	const char* col = m_colorize ? debugColor(level) : 0;
 	if (col)
 	    writeStr(col,::strlen(col));
@@ -1026,6 +1094,8 @@ void Connection::writeDebug(const char *str, int level)
 	    col = debugColor(-2);
 	if (col)
 	    writeStr(col,::strlen(col));
+	if (m_echoing && m_buffer)
+	    writeStr(m_buffer);
     }
 }
 
