@@ -61,7 +61,8 @@ public:
 	: m_f1(f1), m_f2(f2), m_mod(modulated), m_data(0)
 	{ }
     inline ToneData(const ToneData& original)
-	: m_f1(original.f1()), m_f2(original.f2()),
+	: GenObject(),
+	  m_f1(original.f1()), m_f2(original.f2()),
 	  m_mod(original.modulated()), m_data(0)
 	{ }
     virtual ~ToneData();
@@ -96,12 +97,15 @@ public:
 	{ return m_name; }
     bool startup();
     static ToneSource* getTone(String& tone);
-    static const ToneDesc* getBlock(String& tone);
+    static const ToneDesc* getBlock(String& tone, bool oneShot = false);
     static Tone* buildCadence(const String& desc);
     static Tone* buildDtmf(const String& dtmf, int len = DTMF_LEN, int gap = DTMF_GAP);
 protected:
     ToneSource(const ToneDesc* tone = 0);
-    virtual void zeroRefs();
+    virtual bool noChan() const
+	{ return false; }
+    virtual void cleanup();
+    static const ToneDesc* getBlock(String& tone, const ToneDesc* table);
     String m_name;
     const Tone* m_tone;
     int m_repeat;
@@ -115,12 +119,14 @@ private:
 class TempSource : public ToneSource
 {
 public:
-    TempSource(String& desc);
+    TempSource(String& desc, DataBlock* rawdata);
     virtual ~TempSource();
 protected:
-    virtual void cleanup();
+    virtual bool noChan() const
+	{ return true; }
 private:
     Tone* m_single;
+    DataBlock* m_rawdata;                // Raw linear data to be sent
 };
 
 class ToneChan : public Channel
@@ -205,6 +211,12 @@ static const Tone t_outoforder[] = {
     { 1600, tone421hz }, { 1600, 0 },
     { 0, 0 } };
 
+static const Tone t_callwait[] = {
+    { 160, 0 },
+    { 800, tone421hz }, { 800, 0 }, { 800, tone421hz },
+    { 160, 0 },
+    { 0, 0 } };
+
 static const Tone t_info[] = {
     { 2640, tone941hz }, { 240, 0 },
     { 2640, tone1454hz }, { 240, 0 },
@@ -265,6 +277,14 @@ static const ToneDesc s_desc[] = {
     { t_mwatt, "milliwatt", "mw" },
     { t_silence, "silence", 0 },
     { t_noise, "noise", "cn" },
+    { t_probes[0], "probe/0", "probe" },
+    { t_probes[1], "probe/1", 0 },
+    { t_probes[2], "probe/2", 0 },
+    { 0, 0, 0 }
+};
+
+static const ToneDesc s_descOne[] = {
+    { t_callwait, "callwaiting", "cw" },
     { t_dtmf[0], "dtmf/0", "0" },
     { t_dtmf[1], "dtmf/1", "1" },
     { t_dtmf[2], "dtmf/2", "2" },
@@ -281,9 +301,6 @@ static const ToneDesc s_desc[] = {
     { t_dtmf[13], "dtmf/b", "b" },
     { t_dtmf[14], "dtmf/c", "c" },
     { t_dtmf[15], "dtmf/d", "d" },
-    { t_probes[0], "probe/0", "probe" },
-    { t_probes[1], "probe/1", 0 },
-    { t_probes[2], "probe/2", 0 },
     { 0, 0, 0 }
 };
 
@@ -435,7 +452,6 @@ ToneSource::ToneSource(const ToneDesc* tone)
     }
     Debug(&__plugin,DebugAll,"ToneSource::ToneSource(%p) '%s' [%p]",
 	tone,m_name.c_str(),this);
-    asyncDelete(true);
 }
 
 void ToneSource::destroyed()
@@ -447,33 +463,43 @@ void ToneSource::destroyed()
 	Debug(&__plugin,DebugInfo,"ToneSource rate=%u b/s",byteRate(m_time,m_total));
 }
 
-void ToneSource::zeroRefs()
+bool ToneSource::startup()
 {
-    Debug(&__plugin,DebugAll,"ToneSource::zeroRefs() '%s' [%p]",m_name.c_str(),this);
+    DDebug(&__plugin,DebugAll,"ToneSource::startup(\"%s\") tone=%p",m_name.c_str(),m_tone);
+    return m_tone && start("Tone Source");
+}
+
+void ToneSource::cleanup()
+{
+    Debug(&__plugin,DebugAll,"ToneSource::cleanup() '%s' [%p]",m_name.c_str(),this);
     __plugin.lock();
     tones.remove(this,false);
     __plugin.unlock();
-    ThreadedSource::zeroRefs();
+    ThreadedSource::cleanup();
 }
 
-bool ToneSource::startup()
+const ToneDesc* ToneSource::getBlock(String& tone, const ToneDesc* table)
 {
-    return m_tone && start("ToneSource");
+    for (; table->tone; table++) {
+	if (tone == table->name)
+	    return table;
+	if (table->alias && (tone == table->alias)) {
+	    tone = table->name;
+	    return table;
+	}
+    }
+    return 0;
 }
 
-const ToneDesc* ToneSource::getBlock(String& tone)
+const ToneDesc* ToneSource::getBlock(String& tone, bool oneShot)
 {
     if (tone.trimBlanks().toLower().null())
 	return 0;
-    const ToneDesc* d = s_desc;
-    for (; d->tone; d++) {
-	if (tone == d->name)
-	    return d;
-	if (d->alias && (tone == d->alias)) {
-	    tone = d->name;
-	    return d;
-	}
-    }
+    const ToneDesc* d = getBlock(tone,s_desc);
+    if (d)
+	return d;
+    if (oneShot)
+	return getBlock(tone,s_descOne);
     return 0;
 }
 
@@ -531,9 +557,13 @@ ToneSource* ToneSource::getTone(String& tone)
     ObjList* l = &tones;
     for (; l; l = l->next()) {
 	ToneSource* t = static_cast<ToneSource*>(l->get());
-	if (t && (t->name() == tone) && t->ref())
+	if (t && (t->name() == tone) && t->running() && (t->refcount() > 1)) {
+	    t->ref();
 	    return t;
+	}
     }
+    if (!td)
+	return 0;
     ToneSource* t = new ToneSource(td);
     tones.append(t);
     t->startup();
@@ -551,7 +581,7 @@ void ToneSource::run()
     int nsam = tone->nsamples;
     if (nsam < 0)
 	nsam = -nsam;
-    while (alive() && m_tone) {
+    while (m_tone && looping(noChan())) {
 	Thread::check();
 	short *d = (short *) m_data.data();
 	for (unsigned int i = m_data.length()/2; i--; samp++,dpos++) {
@@ -587,6 +617,8 @@ void ToneSource::run()
 	    XDebug(&__plugin,DebugAll,"ToneSource sleeping for " FMT64 " usec",dly);
 	    Thread::usleep((unsigned long)dly);
 	}
+	if (!looping(noChan()))
+	    break;
 	Forward(m_data,m_total/2);
 	m_total += m_data.length();
 	tpos += (m_data.length()*(u_int64_t)1000000/m_brate);
@@ -597,16 +629,32 @@ void ToneSource::run()
 }
 
 
-TempSource::TempSource(String& desc)
-    : m_single(0)
+TempSource::TempSource(String& desc, DataBlock* rawdata)
+    : m_single(0), m_rawdata(rawdata)
 {
     Debug(&__plugin,DebugAll,"TempSource::TempSource(\"%s\") [%p]",desc.c_str(),this);
     if (desc.null())
 	return;
     if (desc.startSkip("*",false))
 	m_repeat = 0;
+    // Build a source used to send raw linear data
+    if (desc == "rawdata") {
+	if (!(m_rawdata && m_rawdata->length() >= sizeof(short))) {
+	    Debug(&__plugin,DebugNote,
+		"TempSource::TempSource(\"%s\") invalid data size=%u [%p]",
+		desc.c_str(),m_rawdata?m_rawdata->length():0,this);
+	    return;
+	}
+	m_name = "rawdata";
+	m_tone = m_single = (Tone*)::malloc(2*sizeof(Tone));
+	m_single[0].nsamples = m_rawdata->length() / sizeof(short);
+	m_single[0].data = (short*)m_rawdata->data();
+	m_single[1].nsamples = 0;
+	m_single[1].data = 0;
+	return;
+    }
     // try first the named tones
-    const ToneDesc* tde = getBlock(desc);
+    const ToneDesc* tde = getBlock(desc,true);
     if (tde) {
 	m_tone = tde->tone;
 	return;
@@ -640,12 +688,7 @@ TempSource::~TempSource()
 	::free(m_single);
 	m_single = 0;
     }
-}
-
-void TempSource::cleanup()
-{
-    ToneSource::cleanup();
-    deref();
+    TelEngine::destruct(m_rawdata);
 }
 
 
@@ -671,6 +714,20 @@ ToneChan::~ToneChan()
     Debug(this,DebugAll,"ToneChan::~ToneChan() %s [%p]",id().c_str(),this);
 }
 
+// Get a data block from a binary parameter of msg
+DataBlock* getRawData(Message& msg)
+{
+    NamedString* data = msg.getParam("rawdata");
+    if (!data)
+	return 0;
+    NamedPointer* p = static_cast<NamedPointer*>(data->getObject("NamedPointer"));
+    if (!p)
+	return 0;
+    GenObject* gen = p->userData();
+    if (!(gen && gen->getObject("DataBlock")))
+	return 0;
+    return static_cast<DataBlock*>(p->takeData());
+}
 
 bool AttachHandler::received(Message& msg)
 {
@@ -686,11 +743,14 @@ bool AttachHandler::received(Message& msg)
     if (src.null() && ovr.null() && repl.null())
 	return false;
 
-    DataEndpoint* de = static_cast<DataEndpoint*>(msg.userObject("DataEndpoint"));
+    RefPointer<DataEndpoint> de = static_cast<DataEndpoint*>(msg.userObject("DataEndpoint"));
     if (!de) {
 	CallEndpoint* ch = static_cast<CallEndpoint*>(msg.userObject("CallEndpoint"));
-	if (ch)
+	if (ch) {
+	    DataEndpoint::commonMutex().lock();
 	    de = ch->setEndpoint();
+	    DataEndpoint::commonMutex().unlock();
+	}
     }
 
     if (!de) {
@@ -710,39 +770,45 @@ bool AttachHandler::received(Message& msg)
 	    msg.clearParam("source");
 	}
 	else {
-	    Debug(DebugWarn,"No source tone '%s' could be attached to %p",src.c_str(),de);
+	    Debug(DebugWarn,"No source tone '%s' could be attached to %p",src.c_str(),(void*)de);
 	    ret = false;
 	}
     }
     if (ovr) {
-	DataConsumer* c = de->getConsumer();
+	DataEndpoint::commonMutex().lock();
+	RefPointer<DataConsumer> c = de->getConsumer();
+	DataEndpoint::commonMutex().unlock();
 	if (c) {
-	    TempSource* t = new TempSource(ovr);
+	    TempSource* t = new TempSource(ovr,getRawData(msg));
 	    if (DataTranslator::attachChain(t,c,true) && t->startup())
 		msg.clearParam("override");
 	    else {
 		Debug(DebugWarn,"Override source tone '%s' failed to start [%p]",ovr.c_str(),t);
 		ret = false;
 	    }
+	    t->deref();
 	}
 	else {
-	    Debug(DebugWarn,"Requested override '%s' to missing consumer of %p",ovr.c_str(),de);
+	    Debug(DebugWarn,"Requested override '%s' to missing consumer of %p",ovr.c_str(),(void*)de);
 	    ret = false;
 	}
     }
     if (repl) {
-	DataConsumer* c = de->getConsumer();
+	DataEndpoint::commonMutex().lock();
+	RefPointer<DataConsumer> c = de->getConsumer();
+	DataEndpoint::commonMutex().unlock();
 	if (c) {
-	    TempSource* t = new TempSource(repl);
+	    TempSource* t = new TempSource(repl,getRawData(msg));
 	    if (DataTranslator::attachChain(t,c,false) && t->startup())
 		msg.clearParam("replace");
 	    else {
 		Debug(DebugWarn,"Replacement source tone '%s' failed to start [%p]",repl.c_str(),t);
 		ret = false;
 	    }
+	    t->deref();
 	}
 	else {
-	    Debug(DebugWarn,"Requested replacement '%s' to missing consumer of %p",repl.c_str(),de);
+	    Debug(DebugWarn,"Requested replacement '%s' to missing consumer of %p",repl.c_str(),(void*)de);
 	    ret = false;
 	}
     }
@@ -756,6 +822,7 @@ bool ToneGenDriver::msgExecute(Message& msg, String& dest)
     if (ch) {
 	ToneChan *tc = new ToneChan(dest);
 	if (ch->connect(tc,msg.getValue("reason"))) {
+	    tc->callConnect(msg);
 	    msg.setParam("peerid",tc->id());
 	    tc->deref();
 	}
@@ -767,6 +834,7 @@ bool ToneGenDriver::msgExecute(Message& msg, String& dest)
     else {
 	Message m("call.route");
 	m.addParam("module",name());
+	m.addParam("cdrtrack",String::boolText(false));
 	m.copyParam(msg,"called");
 	m.copyParam(msg,"caller");
 	m.copyParam(msg,"callername");
