@@ -24,11 +24,7 @@
 
 #include <yatephone.h>
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <string.h>
-#include <fcntl.h>
-#include <errno.h>
 
 using namespace TelEngine;
 namespace { // anonymous
@@ -36,13 +32,13 @@ namespace { // anonymous
 class WaveSource : public ThreadedSource
 {
 public:
-    static WaveSource* create(const String& file, CallEndpoint* chan, bool autoclose = true, bool autorepeat = false);
+    static WaveSource* create(const String& file, CallEndpoint* chan,
+	bool autoclose = true, bool autorepeat = false, const NamedString* param = 0);
     ~WaveSource();
     virtual void run();
     virtual void cleanup();
-    virtual bool zeroRefsTest();
+    virtual void attached(bool added);
     void setNotify(const String& id);
-    bool derefReady();
 private:
     WaveSource(const char* file, CallEndpoint* chan, bool autoclose);
     void init(const String& file, bool autorepeat);
@@ -50,10 +46,10 @@ private:
     void detectWavFormat();
     void detectIlbcFormat();
     bool computeDataRate();
-    bool notify(WaveSource* source, const char* reason = 0);
+    void notify(WaveSource* source, const char* reason = 0);
     CallEndpoint* m_chan;
+    Stream* m_stream;
     DataBlock m_data;
-    int m_fd;
     bool m_swap;
     unsigned m_brate;
     long m_repeatPos;
@@ -61,10 +57,7 @@ private:
     u_int64_t m_time;
     String m_id;
     bool m_autoclose;
-    bool m_autoclean;
     bool m_nodata;
-    bool m_insert;
-    volatile bool m_derefOk;
 };
 
 class WaveConsumer : public DataConsumer
@@ -75,17 +68,18 @@ public:
 	Au,
 	Ilbc,
     };
-    WaveConsumer(const String& file, CallEndpoint* chan, unsigned maxlen, const char* format = 0);
+    WaveConsumer(const String& file, CallEndpoint* chan, unsigned maxlen, const char* format = 0, const NamedString* param = 0);
     ~WaveConsumer();
     virtual bool setFormat(const DataFormat& format);
-    virtual void Consume(const DataBlock& data, unsigned long tStamp);
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags);
+    virtual void attached(bool added);
     inline void setNotify(const String& id)
 	{ m_id = id; }
 private:
     void writeIlbcHeader() const;
     void writeAuHeader();
     CallEndpoint* m_chan;
-    int m_fd;
+    Stream* m_stream;
     bool m_swap;
     bool m_locked;
     Header m_header;
@@ -98,14 +92,14 @@ private:
 class WaveChan : public Channel
 {
 public:
-    WaveChan(const String& file, bool record, unsigned maxlen, bool autorepeat, const char* format = 0);
+    WaveChan(const String& file, bool record, unsigned maxlen, bool autorepeat, const char* format = 0, const NamedString* param = 0);
     ~WaveChan();
 };
 
 class Disconnector : public Thread
 {
 public:
-    Disconnector(CallEndpoint* chan, const String& id, WaveSource* source, bool disc, const char* reason = 0);
+    Disconnector(CallEndpoint* chan, const String& id, WaveSource* source, WaveConsumer* consumer, bool disc, const char* reason = 0);
     virtual ~Disconnector();
     virtual void run();
     bool init();
@@ -113,6 +107,7 @@ private:
     RefPointer<CallEndpoint> m_chan;
     Message* m_msg;
     WaveSource* m_source;
+    WaveConsumer* m_consumer;
     bool m_disc;
 };
 
@@ -136,20 +131,18 @@ public:
     WaveFileDriver();
     virtual void initialize();
     virtual bool msgExecute(Message& msg, String& dest);
+protected:
+    void statusParams(String& str);
 private:
     AttachHandler* m_handler;
 };
 
-bool s_asyncDelete = true;
+Mutex s_mutex(false,"WaveFile");
+int s_reading = 0;
+int s_writing = 0;
 bool s_dataPadding = true;
 bool s_pubReadable = false;
 
-#if defined (S_IRGRP) && defined(S_IROTH)
-#define CREATE_MODE (S_IRUSR|S_IWUSR|(s_pubReadable?(S_IRGRP|S_IROTH):0))
-#else
-#define CREATE_MODE (S_IRUSR|S_IWUSR)
-#endif
-
 INIT_PLUGIN(WaveFileDriver);
 
 
@@ -165,28 +158,49 @@ typedef struct {
 #define ILBC_HEADER_LEN 9
 
 
-WaveSource* WaveSource::create(const String& file, CallEndpoint* chan, bool autoclose, bool autorepeat)
+WaveSource* WaveSource::create(const String& file, CallEndpoint* chan, bool autoclose, bool autorepeat, const NamedString* param)
 {
     WaveSource* tmp = new WaveSource(file,chan,autoclose);
+    NamedPointer* ptr = YOBJECT(NamedPointer,param);
+    if (ptr) {
+	Stream* stream = YOBJECT(Stream,ptr);
+	if (stream) {
+	    DDebug(&__plugin,DebugInfo,"WaveSource using Stream %p [%p]",stream,tmp);
+	    tmp->m_stream = stream;
+	    ptr->takeData();
+	}
+	else {
+	    DataBlock* data = YOBJECT(DataBlock,ptr);
+	    if (data) {
+		DDebug(&__plugin,DebugInfo,"WaveSource using DataBlock %p len=%u [%p]",
+		    data,data->length(),tmp);
+		tmp->m_stream = new MemoryStream(*data);
+	    }
+	}
+    }
     tmp->init(file,autorepeat);
     return tmp;
 }
 
 void WaveSource::init(const String& file, bool autorepeat)
 {
-    if (file == "-") {
-	m_nodata = true;
-	m_brate = 8000;
-	start("WaveSource");
-	return;
-    }
-    m_fd = ::open(file.safe(),O_RDONLY|O_NOCTTY|O_BINARY);
-    if (m_fd < 0) {
-	Debug(DebugWarn,"Opening '%s': error %d: %s",
-	    file.c_str(), errno, ::strerror(errno));
-	m_format.clear();
-	notify(this,"error");
-	return;
+    if (!m_stream) {
+	if (file == "-") {
+	    m_nodata = true;
+	    m_brate = 8000;
+	    start("Wave Source");
+	    return;
+	}
+	m_stream = new File;
+	if (!static_cast<File*>(m_stream)->openPath(file,false,true,false,false,true)) {
+	    Debug(DebugWarn,"Opening '%s': error %d: %s",
+		file.c_str(), m_stream->error(), ::strerror(m_stream->error()));
+	    delete m_stream;
+	    m_stream = 0;
+	    m_format.clear();
+	    notify(this,"error");
+	    return;
+	}
     }
     if (file.endsWith(".gsm"))
 	m_format = "gsm";
@@ -204,6 +218,8 @@ void WaveSource::init(const String& file, bool autorepeat)
 	m_format = "ilbc20";
     else if (file.endsWith(".ilbc30"))
 	m_format = "ilbc30";
+    else if (file.endsWith(".g729"))
+	m_format = "g729";
     else if (file.endsWith(".au"))
 	detectAuFormat();
     else if (file.endsWith(".wav"))
@@ -214,9 +230,8 @@ void WaveSource::init(const String& file, bool autorepeat)
 	Debug(DebugMild,"Unknown format for playback file '%s', assuming signed linear",file.c_str());
     if (computeDataRate()) {
 	if (autorepeat)
-	    m_repeatPos = ::lseek(m_fd,0,SEEK_CUR);
-	asyncDelete(s_asyncDelete);
-	start("WaveSource");
+	    m_repeatPos = m_stream->seek(Stream::SeekCurrent);
+	start("Wave Source");
     }
     else {
 	Debug(DebugWarn,"Unable to compute data rate for file '%s'",file.c_str());
@@ -225,13 +240,14 @@ void WaveSource::init(const String& file, bool autorepeat)
 }
 
 WaveSource::WaveSource(const char* file, CallEndpoint* chan, bool autoclose)
-    : m_chan(chan), m_fd(-1), m_swap(false), m_brate(0), m_repeatPos(-1),
-      m_total(0), m_time(0), m_autoclose(autoclose), m_autoclean(false),
-      m_nodata(false), m_insert(false), m_derefOk(true)
+    : m_chan(chan), m_stream(0), m_swap(false), m_brate(0), m_repeatPos(-1),
+      m_total(0), m_time(0), m_autoclose(autoclose),
+      m_nodata(false)
 {
     Debug(&__plugin,DebugAll,"WaveSource::WaveSource(\"%s\",%p) [%p]",file,chan,this);
-    if (m_chan)
-	m_insert = true;
+    s_mutex.lock();
+    s_reading++;
+    s_mutex.unlock();
 }
 
 WaveSource::~WaveSource()
@@ -245,22 +261,23 @@ WaveSource::~WaveSource()
 	    Debug(&__plugin,DebugInfo,"WaveSource rate=" FMT64U " b/s",m_time);
 	}
     }
-    if (m_fd >= 0) {
-	::close(m_fd);
-	m_fd = -1;
-    }
+    delete m_stream;
+    m_stream = 0;
+    s_mutex.lock();
+    s_reading--;
+    s_mutex.unlock();
 }
 
 void WaveSource::detectAuFormat()
 {
     AuHeader header;
-    if ((::read(m_fd,&header,sizeof(header)) != sizeof(header)) ||
+    if ((m_stream->readData(&header,sizeof(header)) != sizeof(header)) ||
 	(ntohl(header.sign) != 0x2E736E64)) {
 	Debug(DebugMild,"Invalid .au file header, assuming raw signed linear");
-	::lseek(m_fd,0,SEEK_SET);
+	m_stream->seek(0);
 	return;
     }
-    ::lseek(m_fd,ntohl(header.offs),SEEK_SET);
+    m_stream->seek(ntohl(header.offs));
     int samp = ntohl(header.freq);
     int chan = ntohl(header.chan);
     m_brate = samp;
@@ -294,7 +311,7 @@ void WaveSource::detectWavFormat()
 void WaveSource::detectIlbcFormat()
 {
     char header[ILBC_HEADER_LEN+1];
-    if (::read(m_fd,&header,ILBC_HEADER_LEN) == ILBC_HEADER_LEN) {
+    if (m_stream->readData(&header,ILBC_HEADER_LEN) == ILBC_HEADER_LEN) {
 	header[ILBC_HEADER_LEN] = '\0';
 	if (::strcmp(header,"#!iLBC20\n") == 0) {
 	    m_format = "ilbc20";
@@ -323,35 +340,44 @@ void WaveSource::run()
 {
     unsigned long ts = 0;
     int r = 0;
+    // internally reference if used for override or replace purpose
+    bool noChan = (0 == m_chan);
     // wait until at least one consumer is attached
     while (!r) {
-	m_mutex.lock();
+	lock();
 	r = m_consumers.count();
-	m_mutex.unlock();
-	Thread::yield(true);
-	if (!alive())
+	unlock();
+	Thread::yield();
+	if (!looping(noChan)) {
+	    notify(0,"replaced");
 	    return;
+	}
     }
     unsigned int blen = (m_brate*20)/1000;
     DDebug(&__plugin,DebugAll,"Consumer found, starting to play data with rate %d [%p]",m_brate,this);
     m_data.assign(0,blen);
-    // start counting time from now
-    u_int64_t tpos = Time::now();
+    u_int64_t tpos = 0;
     m_time = tpos;
-    do {
-	r = (m_fd >= 0) ? ::read(m_fd,m_data.data(),m_data.length()) : m_data.length();
+    while ((r > 0) && looping(noChan)) {
+	r = m_stream ? m_stream->readData(m_data.data(),m_data.length()) : m_data.length();
 	if (r < 0) {
-	    if (errno == EINTR) {
-		r = 1;
-		continue;
+	    if (m_stream->canRetry()) {
+		if (looping(noChan)) {
+		    r = 1;
+		    continue;
+		}
+		r = 0;
 	    }
 	    break;
 	}
+	// start counting time after the first successful read
+	if (!tpos)
+	    m_time = tpos = Time::now();
 	if (!r) {
 	    if (m_repeatPos >= 0) {
 		DDebug(&__plugin,DebugAll,"Autorepeating from offset %ld [%p]",
 		    m_repeatPos,this);
-		::lseek(m_fd,m_repeatPos,SEEK_SET);
+		m_stream->seek(m_repeatPos);
 		m_data.assign(0,blen);
 		r = 1;
 		continue;
@@ -381,87 +407,64 @@ void WaveSource::run()
 	    XDebug(&__plugin,DebugAll,"WaveSource sleeping for " FMT64 " usec",dly);
 	    Thread::usleep((unsigned long)dly);
 	}
-	if (!alive()) {
-	    notify(0,"replaced");
-	    return;
-	}
+	if (!looping(noChan))
+	    break;
 	Forward(m_data,ts);
 	ts += m_data.length()*8000/m_brate;
 	m_total += r;
 	tpos += (r*(u_int64_t)1000000/m_brate);
-    } while (r > 0);
-    Debug(&__plugin,DebugAll,"WaveSource '%s' end of data (%u played) chan=%p [%p]",m_id.c_str(),m_total,m_chan,this);
-    if (!ref()) {
-	notify(0,"replaced");
-	return;
     }
-    // prevent disconnector thread from succeeding before notify returns
-    m_derefOk = false;
-    // at cleanup time deref the data source if we start no disconnector thread
-    m_autoclean = !notify(this,"eof");
-    if (!deref())
-	m_derefOk = m_autoclean;
+    if (r)
+	notify(0,"replaced");
+    else {
+	Debug(&__plugin,DebugAll,"WaveSource '%s' end of data (%u played) chan=%p [%p]",
+	    m_id.c_str(),m_total,m_chan,this);
+	notify(this,"eof");
+    }
 }
 
 void WaveSource::cleanup()
 {
-    Lock lock(DataEndpoint::commonMutex());
-    Debug(&__plugin,DebugAll,"WaveSource cleanup, total=%u, alive=%s, autoclean=%s chan=%p [%p]",
-	m_total,String::boolText(alive()),String::boolText(m_autoclean),m_chan,this);
-    clearThread();
-    if (m_autoclean) {
-	asyncDelete(false);
-	if (m_insert) {
-	    if (m_chan && (m_chan->getSource() == this))
-		m_chan->setSource();
-	}
-	else
-	    deref();
-	return;
+    RefPointer<CallEndpoint> chan;
+    if (m_chan) {
+	DataEndpoint::commonMutex().lock();
+	chan = m_chan;
+	m_chan = 0;
+	DataEndpoint::commonMutex().unlock();
     }
-    if (m_derefOk)
-	ThreadedSource::cleanup();
-    else
-	m_derefOk = true;
+    Debug(&__plugin,DebugAll,"WaveSource cleanup, total=%u, chan=%p [%p]",
+	m_total,(void*)chan,this);
+    if (chan)
+	chan->clearData(this);
+    ThreadedSource::cleanup();
 }
 
-bool WaveSource::zeroRefsTest()
+void WaveSource::attached(bool added)
 {
-    DDebug(&__plugin,DebugAll,"WaveSource::zeroRefsTest() chan=%p%s%s%s [%p]",
-	m_chan,
-	(thread() ? " thread" : ""),
-	(m_autoclose ? " close" : ""),
-	(m_autoclean ? " clean" : ""),
-	this);
-    // since this is a zombie it has no owner anymore and needs no removal
-    m_chan = 0;
-    m_autoclose = false;
-    m_autoclean = false;
-    return ThreadedSource::zeroRefsTest();
+    if (!added && m_chan && !m_chan->alive()) {
+	DDebug(&__plugin,DebugInfo,"WaveSource clearing dead chan %p [%p]",m_chan,this);
+	m_chan = 0;
+    }
 }
 
 void WaveSource::setNotify(const String& id)
 {
     m_id = id;
-    if ((m_fd < 0) && !m_nodata)
+    if (!(m_stream || m_nodata))
 	notify(this);
 }
 
-bool WaveSource::derefReady()
+void WaveSource::notify(WaveSource* source, const char* reason)
 {
-    for (int i = 0; i < 10; i++) {
-	if (m_derefOk)
-	    return true;
-	Thread::yield();
+    RefPointer<CallEndpoint> chan;
+    if (m_chan) {
+	DataEndpoint::commonMutex().lock();
+	if (source)
+	    chan = m_chan;
+	m_chan = 0;
+	DataEndpoint::commonMutex().unlock();
     }
-    Debug(&__plugin,DebugWarn,"Source not deref ready, waiting more... [%p]",this);
-    Thread::msleep(10);
-    return m_derefOk;
-}
-
-bool WaveSource::notify(WaveSource* source, const char* reason)
-{
-    if (!m_chan) {
+    if (!chan) {
 	if (m_id) {
 	    DDebug(&__plugin,DebugAll,"WaveSource enqueueing notify message [%p]",this);
 	    Message* m = new Message("chan.notify");
@@ -470,28 +473,39 @@ bool WaveSource::notify(WaveSource* source, const char* reason)
 		m->addParam("reason",reason);
 	    Engine::enqueue(m);
 	}
-	return false;
     }
-    if (m_id || m_autoclose) {
+    else if (m_id || m_autoclose) {
 	DDebug(&__plugin,DebugInfo,"Preparing '%s' disconnector for '%s' chan %p '%s' source=%p [%p]",
-	    reason,m_id.c_str(),m_chan,(m_chan ? m_chan->id().c_str() : ""),source,this);
-	Disconnector *disc = new Disconnector(m_chan,m_id,source,m_autoclose,reason);
-	return disc->init();
+	    reason,m_id.c_str(),(void*)chan,chan->id().c_str(),source,this);
+	Disconnector *disc = new Disconnector(chan,m_id,source,0,m_autoclose,reason);
+	if (!disc->init() && m_autoclose)
+	    chan->clearData(source);
     }
-    return false;
+    stop();
 }
 
 
-WaveConsumer::WaveConsumer(const String& file, CallEndpoint* chan, unsigned maxlen, const char* format)
-    : m_chan(chan), m_fd(-1), m_swap(false), m_locked(false), m_header(None),
+WaveConsumer::WaveConsumer(const String& file, CallEndpoint* chan, unsigned maxlen, const char* format, const NamedString* param)
+    : m_chan(chan), m_stream(0), m_swap(false), m_locked(false), m_header(None),
       m_total(0), m_maxlen(maxlen), m_time(0)
 {
-    Debug(&__plugin,DebugAll,"WaveConsumer::WaveConsumer(\"%s\",%p,%u,\"%s\") [%p]",
-	file.c_str(),chan,maxlen,format,this);
+    Debug(&__plugin,DebugAll,"WaveConsumer::WaveConsumer(\"%s\",%p,%u,\"%s\",%p) [%p]",
+	file.c_str(),chan,maxlen,format,param,this);
+    s_mutex.lock();
+    s_writing++;
+    s_mutex.unlock();
     if (format) {
 	m_locked = true;
 	m_format = format;
     }
+    NamedPointer* ptr = YOBJECT(NamedPointer,param);
+    if (ptr) {
+	m_stream = YOBJECT(Stream,ptr);
+	if (m_stream) {
+	    DDebug(&__plugin,DebugInfo,"WaveConsumer using Stream %p [%p]",m_stream,this);
+	    ptr->takeData();
+	}
+    }
     if (file == "-")
 	return;
     else if (file.endsWith(".gsm"))
@@ -510,16 +524,23 @@ WaveConsumer::WaveConsumer(const String& file, CallEndpoint* chan, unsigned maxl
 	m_format = "ilbc20";
     else if (file.endsWith(".ilbc30"))
 	m_format = "ilbc30";
+    else if (file.endsWith(".g729"))
+	m_format = "g729";
     else if (file.endsWith(".lbc"))
 	m_header=Ilbc;
     else if (file.endsWith(".au"))
 	m_header=Au;
     else if (!file.endsWith(".slin"))
 	Debug(DebugMild,"Unknown format for recorded file '%s', assuming signed linear",file.c_str());
-    m_fd = ::open(file.safe(),O_WRONLY|O_CREAT|O_TRUNC|O_NOCTTY|O_BINARY,CREATE_MODE);
-    if (m_fd < 0)
+    if (m_stream)
+	return;
+    m_stream = new File;
+    if (!static_cast<File*>(m_stream)->openPath(file,true,false,true,false,true,s_pubReadable)) {
 	Debug(DebugWarn,"Creating '%s': error %d: %s",
-	    file.c_str(), errno, ::strerror(errno));
+	    file.c_str(), m_stream->error(), ::strerror(m_stream->error()));
+	delete m_stream;
+	m_stream = 0;
+    }
 }
 
 WaveConsumer::~WaveConsumer()
@@ -532,18 +553,19 @@ WaveConsumer::~WaveConsumer()
 	    Debug(&__plugin,DebugInfo,"WaveConsumer rate=" FMT64U " b/s",m_time);
 	}
     }
-    if (m_fd >= 0) {
-	::close(m_fd);
-	m_fd = -1;
-    }
+    delete m_stream;
+    m_stream = 0;
+    s_mutex.lock();
+    s_writing--;
+    s_mutex.unlock();
 }
 
 void WaveConsumer::writeIlbcHeader() const
 {
     if (m_format == "ilbc20")
-	::write(m_fd,"#!iLBC20\n",ILBC_HEADER_LEN);
+	m_stream->writeData("#!iLBC20\n",ILBC_HEADER_LEN);
     else if (m_format == "ilbc30")
-	::write(m_fd,"#!iLBC30\n",ILBC_HEADER_LEN);
+	m_stream->writeData("#!iLBC30\n",ILBC_HEADER_LEN);
     else
 	Debug(DebugMild,"Invalid iLBC format '%s', not writing header",m_format.c_str());
 }
@@ -564,22 +586,22 @@ void WaveConsumer::writeAuHeader()
     }
     if (fmt == "slin") {
 	m_swap = true;
-	header.form = ntohl(3);
+	header.form = htonl(3);
     }
     else if (fmt == "mulaw")
-	header.form = ntohl(1);
+	header.form = htonl(1);
     else if (fmt == "alaw")
-	header.form = ntohl(27);
+	header.form = htonl(27);
     else {
 	Debug(DebugMild,"Invalid au format '%s', not writing header",m_format.c_str());
 	return;
     }
     header.sign = htonl(0x2E736E64);
     header.offs = htonl(sizeof(header));
-    header.freq = ntohl(rate);
-    header.chan = ntohl(chans);
+    header.freq = htonl(rate);
+    header.chan = htonl(chans);
     header.len = 0;
-    ::write(m_fd,&header,sizeof(header));
+    m_stream->writeData(&header,sizeof(header));
 }
 
 bool WaveConsumer::setFormat(const DataFormat& format)
@@ -608,12 +630,12 @@ bool WaveConsumer::setFormat(const DataFormat& format)
     return false;
 }
 
-void WaveConsumer::Consume(const DataBlock& data, unsigned long tStamp)
+unsigned long WaveConsumer::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 {
     if (!data.null()) {
 	if (!m_time)
 	    m_time = Time::now();
-	if (m_fd >= 0) {
+	if (m_stream) {
 	    switch (m_header) {
 		case Ilbc:
 		    writeIlbcHeader();
@@ -631,34 +653,48 @@ void WaveConsumer::Consume(const DataBlock& data, unsigned long tStamp)
 		const uint16_t* s = (const uint16_t*)data.data();
 		uint16_t* d = (uint16_t*)swapped.data();
 		for (unsigned int i = 0; i < n; i+= 2)
-		    *d++ = ntohs(*s++);
-		::write(m_fd,swapped.data(),swapped.length());
+		    *d++ = htons(*s++);
+		m_stream->writeData(swapped);
 	    }
 	    else
-		::write(m_fd,data.data(),data.length());
+		m_stream->writeData(data);
 	}
 	m_total += data.length();
 	if (m_maxlen && (m_total >= m_maxlen)) {
 	    m_maxlen = 0;
-	    if (m_fd >= 0) {
-		::close(m_fd);
-		m_fd = -1;
-	    }
+	    delete m_stream;
+	    m_stream = 0;
+	    RefPointer<CallEndpoint> chan;
 	    if (m_chan) {
-		DDebug(&__plugin,DebugInfo,"Preparing 'maxlen' disconnector for '%s' chan %p '%s' in consumer [%p]",
-		    m_id.c_str(),m_chan,(m_chan ? m_chan->id().c_str() : ""),this);
-		Disconnector *disc = new Disconnector(m_chan,m_id,0,false,"maxlen");
+		DataEndpoint::commonMutex().lock();
+		chan = m_chan;
 		m_chan = 0;
+		DataEndpoint::commonMutex().unlock();
+	    }
+	    if (chan) {
+		DDebug(&__plugin,DebugInfo,"Preparing 'maxlen' disconnector for '%s' chan %p '%s' in consumer [%p]",
+		    m_id.c_str(),(void*)chan,chan->id().c_str(),this);
+		Disconnector *disc = new Disconnector(chan,m_id,0,this,false,"maxlen");
 		disc->init();
 	    }
 	}
+	return invalidStamp();
+    }
+    return 0;
+}
+
+void WaveConsumer::attached(bool added)
+{
+    if (!added && m_chan && !m_chan->alive()) {
+	DDebug(&__plugin,DebugInfo,"WaveConsumer clearing dead chan %p [%p]",m_chan,this);
+	m_chan = 0;
     }
 }
 
 
-Disconnector::Disconnector(CallEndpoint* chan, const String& id, WaveSource* source, bool disc, const char* reason)
+Disconnector::Disconnector(CallEndpoint* chan, const String& id, WaveSource* source, WaveConsumer* consumer, bool disc, const char* reason)
     : Thread("WaveDisconnector"),
-      m_chan(chan), m_msg(0), m_source(0), m_disc(disc)
+      m_chan(chan), m_msg(0), m_source(0), m_consumer(consumer), m_disc(disc)
 {
     if (id) {
 	Message* m = new Message("chan.notify");
@@ -670,15 +706,6 @@ Disconnector::Disconnector(CallEndpoint* chan, const String& id, WaveSource* sou
 	m->userData(m_chan);
 	m_msg = m;
     }
-    if (source) {
-	if (source->ref())
-	    m_source = source;
-	else {
-	    Debug(&__plugin,DebugGoOn,"Disconnecting dead source %p, reason: '%s'",
-		source,reason);
-	    m_chan = 0;
-	}
-    }
 }
 
 Disconnector::~Disconnector()
@@ -691,12 +718,12 @@ Disconnector::~Disconnector()
 
 bool Disconnector::init()
 {
-    if (error()) {
-	Debug(&__plugin,DebugGoOn,"Error creating disconnector thread %p",this);
+    if (error() || !startup()) {
+	Debug(&__plugin,DebugGoOn,"Error starting disconnector thread %p",this);
 	delete this;
 	return false;
     }
-    return startup();
+    return true;
 }
 
 void Disconnector::run()
@@ -706,36 +733,31 @@ void Disconnector::run()
     if (!m_chan)
 	return;
     if (m_source) {
-	if (m_chan->getSource() == m_source)
-	    m_chan->setSource();
-	else
-	    Debug(&__plugin,DebugMild,"Source %p in channel %p was replaced with %p",
+	if (!m_chan->clearData(m_source))
+	    Debug(&__plugin,DebugNote,"Source %p in channel %p was replaced with %p",
 		m_source,(void*)m_chan,m_chan->getSource());
-	if (!m_source->derefReady())
-	    Debug(&__plugin,DebugGoOn,"Source %p is not deref ready, crash may occur",m_source);
-	m_source->deref();
 	if (m_disc)
 	    m_chan->disconnect("eof");
     }
     else {
 	if (m_msg)
-	    m_chan->setConsumer();
+	    m_chan->clearData(m_consumer);
 	else
 	    m_chan->disconnect();
     }
 }
 
 
-WaveChan::WaveChan(const String& file, bool record, unsigned maxlen, bool autorepeat, const char* format)
+WaveChan::WaveChan(const String& file, bool record, unsigned maxlen, bool autorepeat, const char* format, const NamedString* param)
     : Channel(__plugin)
 {
     Debug(this,DebugAll,"WaveChan::WaveChan(%s) [%p]",(record ? "record" : "play"),this);
     if (record) {
-	setConsumer(new WaveConsumer(file,this,maxlen,format));
+	setConsumer(new WaveConsumer(file,this,maxlen,format,param));
 	getConsumer()->deref();
     }
     else {
-	setSource(WaveSource::create(file,this,true,autorepeat));
+	setSource(WaveSource::create(file,this,true,autorepeat,param));
 	getSource()->deref();
     }
 }
@@ -835,8 +857,6 @@ bool AttachHandler::received(Message &msg)
     // if single attach was requested we can return true if everything is ok
     bool ret = msg.getBoolValue("single");
 
-    String ml(msg.getValue("maxlen"));
-    unsigned maxlen = ml.toInteger(0);
     CallEndpoint *ch = static_cast<CallEndpoint*>(msg.userData());
     if (!ch) {
 	if (!src.null())
@@ -848,57 +868,66 @@ bool AttachHandler::received(Message &msg)
 	return false;
     }
 
+    const char* notify = msg.getValue("notify");
+    unsigned int maxlen = msg.getIntValue("maxlen");
+
     if (!src.null()) {
-	WaveSource* s = WaveSource::create(src,ch,false,msg.getBoolValue("autorepeat"));
+	WaveSource* s = WaveSource::create(src,ch,false,msg.getBoolValue("autorepeat"),msg.getParam("source"));
 	ch->setSource(s);
-	s->setNotify(msg.getValue("notify"));
+	s->setNotify(msg.getValue("notify_source",notify));
 	s->deref();
 	msg.clearParam("source");
     }
 
     if (!cons.null()) {
-	WaveConsumer* c = new WaveConsumer(cons,ch,maxlen,msg.getValue("format"));
-	c->setNotify(msg.getValue("notify"));
+	WaveConsumer* c = new WaveConsumer(cons,ch,maxlen,msg.getValue("format"),msg.getParam("consumer"));
+	c->setNotify(msg.getValue("notify_consumer",notify));
 	ch->setConsumer(c);
 	c->deref();
 	msg.clearParam("consumer");
     }
 
     while (!ovr.null()) {
-	DataConsumer* c = ch->getConsumer();
+	DataEndpoint::commonMutex().lock();
+	RefPointer<DataConsumer> c = ch->getConsumer();
+	DataEndpoint::commonMutex().unlock();
 	if (!c) {
 	    Debug(DebugWarn,"Wave override '%s' attach request with no consumer!",ovr.c_str());
 	    ret = false;
 	    break;
 	}
-	WaveSource* s = WaveSource::create(ovr,0,false,msg.getBoolValue("autorepeat"));
-	s->setNotify(msg.getValue("notify"));
+	WaveSource* s = WaveSource::create(ovr,0,false,msg.getBoolValue("autorepeat"),msg.getParam("override"));
+	s->setNotify(msg.getValue("notify_override",notify));
 	if (DataTranslator::attachChain(s,c,true))
 	    msg.clearParam("override");
 	else {
-	    Debug(DebugWarn,"Failed to override attach wave '%s' to consumer %p",ovr.c_str(),c);
-	    s->deref();
+	    Debug(DebugWarn,"Failed to override attach wave '%s' to consumer %p",
+		ovr.c_str(),(void*)c);
 	    ret = false;
 	}
+	s->deref();
 	break;
     }
 
     while (!repl.null()) {
-	DataConsumer* c = ch->getConsumer();
+	DataEndpoint::commonMutex().lock();
+	RefPointer<DataConsumer> c = ch->getConsumer();
+	DataEndpoint::commonMutex().unlock();
 	if (!c) {
 	    Debug(DebugWarn,"Wave replacement '%s' attach request with no consumer!",repl.c_str());
 	    ret = false;
 	    break;
 	}
-	WaveSource* s = WaveSource::create(repl,0,false,msg.getBoolValue("autorepeat"));
-	s->setNotify(msg.getValue("notify"));
+	WaveSource* s = WaveSource::create(repl,0,false,msg.getBoolValue("autorepeat"),msg.getParam("replace"));
+	s->setNotify(msg.getValue("notify_replace",notify));
 	if (DataTranslator::attachChain(s,c,false))
 	    msg.clearParam("replace");
 	else {
-	    Debug(DebugWarn,"Failed to replacement attach wave '%s' to consumer %p",repl.c_str(),c);
-	    s->deref();
+	    Debug(DebugWarn,"Failed to replacement attach wave '%s' to consumer %p",
+		repl.c_str(),(void*)c);
 	    ret = false;
 	}
+	s->deref();
 	break;
     }
 
@@ -924,11 +953,11 @@ bool RecordHandler::received(Message &msg)
 	    else {
 		Debug(DebugWarn,"Could not attach call recorder with method '%s', use 'record'",
 		    c1.matchString(1).c_str());
-		c1 = "";
+		c1.clear();
 	    }
 	}
 	else
-	    c1 = "";
+	    c1.clear();
     }
 
     String c2(msg.getValue("peer"));
@@ -944,18 +973,18 @@ bool RecordHandler::received(Message &msg)
 	    else {
 		Debug(DebugWarn,"Could not attach peer recorder with method '%s', use 'record'",
 		    c2.matchString(1).c_str());
-		c2 = "";
+		c2.clear();
 	    }
 	}
 	else
-	    c2 = "";
+	    c2.clear();
     }
 
     if (c1.null() && c2.null())
 	return false;
 
-    String ml(msg.getValue("maxlen"));
-    unsigned maxlen = ml.toInteger(0);
+    const char* notify = msg.getValue("notify");
+    unsigned int maxlen = msg.getIntValue("maxlen");
 
     CallEndpoint *ch = static_cast<CallEndpoint*>(msg.userObject("CallEndpoint"));
     DataEndpoint *de = static_cast<DataEndpoint*>(msg.userObject("DataEndpoint"));
@@ -971,15 +1000,15 @@ bool RecordHandler::received(Message &msg)
     }
 
     if (!c1.null()) {
-	WaveConsumer* c = new WaveConsumer(c1,ch,maxlen,msg.getValue("format"));
-	c->setNotify(msg.getValue("notify"));
+	WaveConsumer* c = new WaveConsumer(c1,ch,maxlen,msg.getValue("format"),msg.getParam("call"));
+	c->setNotify(msg.getValue("notify_call",notify));
 	de->setCallRecord(c);
 	c->deref();
     }
 
     if (!c2.null()) {
-	WaveConsumer* c = new WaveConsumer(c2,ch,maxlen,msg.getValue("format"));
-	c->setNotify(msg.getValue("notify"));
+	WaveConsumer* c = new WaveConsumer(c2,ch,maxlen,msg.getValue("format"),msg.getParam("peer"));
+	c->setNotify(msg.getValue("notify_peer",notify));
 	de->setPeerRecord(c);
 	c->deref();
     }
@@ -1004,14 +1033,14 @@ bool WaveFileDriver::msgExecute(Message& msg, String& dest)
 	return false;
     }
 
-    String ml(msg.getValue("maxlen"));
-    unsigned maxlen = ml.toInteger(0);
+    unsigned int maxlen = msg.getIntValue("maxlen");
     CallEndpoint* ch = static_cast<CallEndpoint*>(msg.userData());
     if (ch) {
 	Debug(this,DebugInfo,"%s wave file '%s'", (meth ? "Record to" : "Play from"),
 	    dest.matchString(2).c_str());
-	WaveChan *c = new WaveChan(dest.matchString(2),meth,maxlen,msg.getBoolValue("autorepeat"),msg.getValue("format"));
+	WaveChan *c = new WaveChan(dest.matchString(2),meth,maxlen,msg.getBoolValue("autorepeat"),msg.getValue("format"),msg.getParam("callto"));
 	if (ch->connect(c,msg.getValue("reason"))) {
+	    c->callConnect(msg);
 	    msg.setParam("peerid",c->id());
 	    c->deref();
 	    return true;
@@ -1023,6 +1052,7 @@ bool WaveFileDriver::msgExecute(Message& msg, String& dest)
     }
     Message m("call.route");
     m.addParam("module",name());
+    m.addParam("cdrtrack",String::boolText(false));
     String callto(msg.getValue("direct"));
     if (callto.null()) {
 	const char *targ = msg.getValue("target");
@@ -1057,6 +1087,13 @@ bool WaveFileDriver::msgExecute(Message& msg, String& dest)
     return false;
 }
 
+void WaveFileDriver::statusParams(String& str)
+{
+    str.append("play=",",") << s_reading;
+    str << ",record=" << s_writing;
+    Driver::statusParams(str);
+}
+
 WaveFileDriver::WaveFileDriver()
     : Driver("wave","misc"), m_handler(0)
 {
@@ -1067,7 +1104,6 @@ void WaveFileDriver::initialize()
 {
     Output("Initializing module WaveFile");
     setup();
-    s_asyncDelete = Engine::config().getBoolValue("hacks","asyncdelete",true);
     s_dataPadding = Engine::config().getBoolValue("hacks","datapadding",true);
     s_pubReadable = Engine::config().getBoolValue("hacks","wavepubread",false);
     if (!m_handler) {
