@@ -126,7 +126,7 @@ private:
 class YIAXLineContainer : public Mutex
 {
 public:
-    inline YIAXLineContainer() : Mutex(true) {}
+    inline YIAXLineContainer() : Mutex(true,"YIAXLineContainer") {}
     inline ~YIAXLineContainer() {}
 
     /*
@@ -345,6 +345,7 @@ public:
     YIAXDriver();
     virtual ~YIAXDriver();
     virtual void initialize();
+    virtual bool hasLine(const String& line) const;
     virtual bool msgExecute(Message& msg, String& dest);
     virtual bool msgRoute(Message& msg);
     virtual bool received(Message& msg, int id);
@@ -400,7 +401,7 @@ class YIAXConsumer : public DataConsumer
 public:
     YIAXConsumer(YIAXConnection* conn, u_int32_t format, const char* formatText);
     ~YIAXConsumer();
-    virtual void Consume(const DataBlock &data, unsigned long tStamp);
+    virtual unsigned long Consume(const DataBlock &data, unsigned long tStamp, unsigned long flags);
 private:
     YIAXConnection* m_connection;
     unsigned m_total;
@@ -434,6 +435,7 @@ public:
     virtual void callRejected(const char* error, const char* reason = 0, const Message* msg = 0);
     virtual bool callPrerouted(Message& msg, bool handled);
     virtual bool callRouted(Message& msg);
+    virtual bool msgProgress(Message& msg);
     virtual bool msgRinging(Message& msg);
     virtual bool msgAnswered(Message& msg);
     virtual bool msgTone(Message& msg, const char* tone);
@@ -522,9 +524,9 @@ private:
  * Local data
  */
 static Configuration s_cfg; 		// Configuration file
-static YIAXDriver iplugin;		// Init the driver
 static YIAXLineContainer s_lines;	// Lines
 static Thread::Priority s_priority = Thread::Normal;  // Threads priority
+static YIAXDriver iplugin;		// Init the driver
 
 
 /*
@@ -965,11 +967,11 @@ void YIAXEngine::start(u_int16_t listenThreadCount, u_int16_t eventThreadCount,
     if (!trunkThreadCount)
 	Debug(this,DebugWarn,"YIAXEngine. No trunking threads(s)!");
     for (; listenThreadCount; listenThreadCount--)
-	(new YIAXListener(this,"YIAXListener thread",s_priority))->startup();
+	(new YIAXListener(this,"YIAX Listener",s_priority))->startup();
     for (; eventThreadCount; eventThreadCount--)
-	(new YIAXGetEvent(this,"YIAXGetEvent thread"))->startup();
+	(new YIAXGetEvent(this,"YIAX GetEvent"))->startup();
     for (; trunkThreadCount; trunkThreadCount--)
-	(new YIAXTrunking(this,"YIAXTrunking thread",s_priority))->startup();
+	(new YIAXTrunking(this,"YIAX Trunking",s_priority))->startup();
     m_threadsCreated = true;
 }
 
@@ -1066,8 +1068,13 @@ bool YIAXEngine::userreg(IAXTransaction* tr, bool regrel)
 	String data = "iax/iax2:";
 	data << tr->username() << "@";
 	data << tr->remoteAddr().host() << ":" << tr->remoteAddr().port();
-	// TODO: support number != username
-	data << "/" << tr->username();
+	const char* called = tr->calledNo();
+	const char* context = tr->calledContext();
+	if (called || context) {
+	    data << "/" << called;
+	    if (context)
+		data << "@" << context;
+	}
 	msg.addParam("data",data);
 	msg.addParam("expires",String((unsigned int)tr->expire()));
     }
@@ -1135,17 +1142,15 @@ void YIAXDriver::initialize()
     // If desired codec is disabled fall back to last in list
     if (!m_defaultCodec)
 	m_defaultCodec = fallback;
-    // Other options
-    m_port = s_cfg.getIntValue("general","port",4569);
-    String iface = s_cfg.getValue("general","addr");
-    bool authReq = s_cfg.getBoolValue("registrar","auth_required",true);
-    s_priority = Thread::priority(s_cfg.getValue("general","thread"),s_priority);
-    DDebug(this,DebugInfo,"Default thread priority set to '%s'",Thread::priority(s_priority));
     unlock();
+    // Setup driver if this is the first call
+    if (m_iaxEngine)
+	return;
     setup();
-    // We need channels to be dropped on shutdown
     installRelay(Halt);
     installRelay(Route);
+    installRelay(Progress);
+    Engine::install(new YIAXRegDataHandler);
     // Init IAX engine
     u_int16_t transListCount = 64;
     u_int16_t retransCount = 5;
@@ -1154,17 +1159,17 @@ void YIAXDriver::initialize()
     u_int16_t transTimeout = 10;
     u_int16_t maxFullFrameDataLen = 1400;
     u_int32_t trunkSendInterval = 10;
-    if (!m_iaxEngine) {
-	Engine::install(new YIAXRegDataHandler);
-	m_iaxEngine = new YIAXEngine(iface,m_port,transListCount,retransCount,retransInterval,authTimeout,
-		transTimeout,maxFullFrameDataLen,trunkSendInterval,authReq);
-	m_iaxEngine->debugChain(this);
-	int tos = s_cfg.getIntValue("general","tos",dict_tos,0);
-	if (tos) {
-	    if (!m_iaxEngine->socket().setTOS(tos))
-		Debug(this,DebugWarn,"Could not set IP TOS to 0x%02x",tos);
-	}
-    }
+    m_port = s_cfg.getIntValue("general","port",4569);
+    String iface = s_cfg.getValue("general","addr");
+    bool authReq = s_cfg.getBoolValue("registrar","auth_required",true);
+    m_iaxEngine = new YIAXEngine(iface,m_port,transListCount,retransCount,retransInterval,authTimeout,
+	transTimeout,maxFullFrameDataLen,trunkSendInterval,authReq);
+    m_iaxEngine->debugChain(this);
+    int tos = s_cfg.getIntValue("general","tos",dict_tos,0);
+    if (tos && !m_iaxEngine->socket().setTOS(tos))
+	Debug(this,DebugWarn,"Could not set IP TOS to 0x%02x",tos);
+    s_priority = Thread::priority(s_cfg.getValue("general","thread"),s_priority);
+    DDebug(this,DebugInfo,"Default thread priority set to '%s'",Thread::priority(s_priority));
     int readThreadCount = s_cfg.getIntValue("general","read_threads",Engine::clientMode() ? 1 : 3);
     if (readThreadCount < 1)
 	readThreadCount = 1;
@@ -1177,23 +1182,18 @@ void YIAXDriver::initialize()
     m_iaxEngine->start(readThreadCount,eventThreadCount,trunkingThreadCount);
 }
 
+// Check if we have a line
+bool YIAXDriver::hasLine(const String& line) const
+{
+    return line && s_lines.hasLine(line);
+}
+
 // Route calls that use a line owned by this driver
 bool YIAXDriver::msgRoute(Message& msg)
 {
-    String called = msg.getValue("called");
-    if (!isE164(called))
+    if (!isE164(msg.getValue("called")))
 	return false;
-    String line = msg.getValue("line");
-    if (line.null())
-	line = msg.getValue("account");
-    if (line.null())
-	return false;
-    if (s_lines.hasLine(line)) {
-	msg.setParam("line",line);
-	msg.retValue() = prefix() + called;
-	return true;
-    }
-    return false;
+    return Driver::msgRoute(msg);
 }
 
 bool YIAXDriver::msgExecute(Message& msg, String& dest)
@@ -1353,13 +1353,16 @@ YIAXConsumer::~YIAXConsumer()
 {
 }
 
-void YIAXConsumer::Consume(const DataBlock& data, unsigned long tStamp)
+unsigned long YIAXConsumer::Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags)
 {
     if (m_connection && !m_connection->mutedOut()) {
 	m_total += data.length();
-	if (m_connection->transaction())
+	if (m_connection->transaction()) {
 	    m_connection->transaction()->sendMedia(data,m_format);
+	    return invalidStamp();
+	}
     }
+    return 0;
 }
 
 /**
@@ -1388,7 +1391,10 @@ void YIAXSource::Forward(const DataBlock& data, unsigned long tStamp)
 YIAXConnection::YIAXConnection(YIAXEngine* iaxEngine, IAXTransaction* transaction, Message* msg)
     : Channel(&iplugin,0,transaction->outgoing()),
       m_iaxEngine(iaxEngine), m_transaction(transaction), m_mutedIn(false), m_mutedOut(false),
-      m_hangup(true), m_mutexTrans(true), m_mutexRefIncreased(true), m_refIncreased(false)
+      m_hangup(true),
+      m_mutexTrans(true,"YIAXConnection::trans"),
+      m_mutexRefIncreased(true,"YIAXConnection::refIncreased"),
+      m_refIncreased(false)
 {
     Debug(this,DebugAll,"Created %s [%p]",isOutgoing()?"outgoing":"incoming",this);
     setMaxcall(msg);
@@ -1438,7 +1444,7 @@ void YIAXConnection::callRejected(const char* error, const char* reason, const M
 	error = reason;
     String s(error);
     Lock lock(m_mutexTrans);
-    if (m_transaction && s == "noauth")
+    if (m_transaction && s == "noauth") {
 	if (safeRefIncrease()) {
 	    Debug(this,DebugInfo,"callRejected. Requesting authentication [%p]",this);
 	    m_transaction->sendAuth();
@@ -1446,6 +1452,7 @@ void YIAXConnection::callRejected(const char* error, const char* reason, const M
 	}
 	else
 	    error = "temporary-failure";
+    }
     lock.drop();
     Debug(this,DebugCall,"callRejected. Reason: '%s' [%p]",error,this);
     hangup(reason,true);
@@ -1462,6 +1469,18 @@ bool YIAXConnection::callRouted(Message& msg)
     return true;
 }
 
+bool YIAXConnection::msgProgress(Message& msg)
+{
+    Lock lock(&m_mutexTrans);
+    if (m_transaction) {
+	m_transaction->sendProgress();
+	// only start audio output for early media
+	startAudioOut();
+	return Channel::msgProgress(msg);
+    }
+    return false;
+}
+
 bool YIAXConnection::msgRinging(Message& msg)
 {
     Lock lock(&m_mutexTrans);
@@ -1546,7 +1565,8 @@ void YIAXConnection::handleEvent(IAXEvent* event)
 	    DDebug(this,DebugCall,"DTMF: %s [%p]",dtmf.safe(),this);
 	    Message* m = message("chan.dtmf");
 	    m->addParam("text",dtmf);
-	    Engine::enqueue(m);
+	    m->addParam("detected","iax-event");
+	    dtmfEnqueue(m);
 	    }
 	    break;
 	case IAXEvent::Noise:
@@ -1554,6 +1574,7 @@ void YIAXConnection::handleEvent(IAXEvent* event)
 	    break;
 	case IAXEvent::Progressing:
 	    DDebug(this,DebugInfo,"CALL PROGRESSING [%p]",this);
+	    Engine::enqueue(message("call.progress",false,true));
 	    break;
 	case IAXEvent::Accept:
 	    DDebug(this,DebugCall,"ACCEPT [%p]",this);
@@ -1587,8 +1608,10 @@ void YIAXConnection::handleEvent(IAXEvent* event)
 		event->getList().getString(IAXInfoElement::CAUSE,m_reason);
 		DDebug(this,DebugCall,"REJECT/HANGUP: '%s' [%p]",m_reason.c_str(),this);
 	    }
+#ifdef DEBUG
 	    else
-		DDebug(this,DebugCall,"REJECT/HANGUP [%p]",this);
+		Debug(this,DebugCall,"REJECT/HANGUP [%p]",this);
+#endif
 	    break;
 	case IAXEvent::Timeout:
 	    DDebug(this,DebugNote,"TIMEOUT. Transaction: %u,%u, Frame: %u,%u [%p]",
