@@ -51,8 +51,9 @@ public:
     /**
      * Create a configuration from a file
      * @param filename Name of file to initialize from
+     * @param warn True to warn if the configuration could not be loaded
      */
-    Configuration(const char* filename);
+    Configuration(const char* filename, bool warn = true);
 
     /**
      * Assignment from string operator
@@ -189,9 +190,10 @@ public:
 
     /**
      * Load the configuration from file
+     * @param warn True to also warn if the configuration could not be loaded
      * @return True if successfull, false for failure
      */
-    bool load();
+    bool load(bool warn = true);
 
     /**
      * Save the configuration to file
@@ -208,6 +210,7 @@ private:
 };
 
 class MessageDispatcher;
+class MessageRelay;
 
 /**
  * This class holds the messages that are moved around in the engine.
@@ -375,6 +378,7 @@ private:
 class YATE_API MessageHandler : public String
 {
     friend class MessageDispatcher;
+    friend class MessageRelay;
 public:
     /**
      * Creates a new message handler.
@@ -433,9 +437,18 @@ public:
      */
     void clearFilter();
 
-private:
+protected:    
+    /**
+     * Remove the handler from its dispatcher, remove any installed filter.
+     * This method is called internally from destruct and the destructor
+     */
     void cleanup();
+
+private:
+    virtual bool receivedInternal(Message& msg);
+    void safeNow();
     unsigned m_priority;
+    int m_unsafe;
     MessageDispatcher* m_dispatcher;
     NamedString* m_filter;
 };
@@ -474,14 +487,23 @@ public:
 	: MessageHandler(name,priority), m_receiver(receiver), m_id(id) { }
 
     /**
-     * This method is called whenever the registered name matches the message.
+     * This method is not called from MessageHandler through polymorphism
+     *  and should not be used or reimplemented.
      * @param msg The received message
-     * @return True to stop processing, false to try other handlers
+     * @return True if the receiver exists and has handled the message
      */
     virtual bool received(Message& msg)
-	{ return m_receiver ? m_receiver->received(msg,m_id) : false; }
+	{ return m_receiver && m_receiver->received(msg,m_id); }
+
+    /**
+     * Get the ID of this message relay
+     * @return Numeric identifier passed to receiver
+     */
+    inline int id() const
+	{ return m_id; }
 
 private:
+    virtual bool receivedInternal(Message& msg);
     MessageReceiver* m_receiver;
     int m_id;
 };
@@ -524,7 +546,7 @@ class YATE_API MessagePostHook : public GenObject, public MessageNotifier
  *  messages that are typically dispatched by a separate thread.
  * @short A message dispatching hub
  */
-class YATE_API MessageDispatcher : public GenObject
+class YATE_API MessageDispatcher : public GenObject, public Mutex
 {
 public:
     /**
@@ -539,6 +561,9 @@ public:
 
     /**
      * Installs a handler in the dispatcher.
+     * The handlers are installed in ascending order of their priorities.
+     * There is NO GUARANTEE on the order of handlers with equal priorities
+     *  although for avoiding uncertainity such handlers are sorted by address.
      * @param handler A pointer to the handler to install
      * @return True on success, false on failure
      */
@@ -552,7 +577,11 @@ public:
     bool uninstall(MessageHandler* handler);
 
     /**
-     * Synchronously dispatch a message to the installed handlers
+     * Synchronously dispatch a message to the installed handlers.
+     * Handlers matching the message name and filter parameter are called in
+     *  their installed order (based on priority) until one returns true.
+     * Note that in some cases when a handler is removed from the list
+     *  other handlers with equal priority may be called twice.
      * @param msg The message to dispatch
      * @return True if one handler accepted it, false if all ignored
      */
@@ -612,7 +641,6 @@ private:
     ObjList m_handlers;
     ObjList m_messages;
     ObjList m_hooks;
-    Mutex m_mutex;
     unsigned int m_changes;
     u_int64_t m_warnTime;
 };
@@ -633,8 +661,9 @@ public:
     /**
      * Creates a new Plugin container.
      * @param name the undecorated name of the library that contains the plugin
+     * @param earlyInit True to initialize the plugin early
      */
-    Plugin(const char* name);
+    Plugin(const char* name, bool earlyInit = false);
 
     /**
      * Creates a new Plugin container.
@@ -667,6 +696,16 @@ public:
      */
     virtual bool isBusy() const
 	{ return false; }
+
+    /**
+     * Check if the module is to be initialized early
+     * @return True if the module should be initialized before regular ones
+     */
+    bool earlyInit() const
+	{ return m_early; }
+
+private:
+    bool m_early;
 };
 
 #if 0 /* for documentation generator */
@@ -675,9 +714,55 @@ public:
  * @param pclass Class of the plugin to create
  */
 void INIT_PLUGIN(class pclass);
+
+/**
+ * Macro to create the unloading function
+ * @param unloadNow True if asked to unload immediately, false if just checking
+ * @return True if the plugin can be unloaded, false if not
+ */
+bool UNLOAD_PLUGIN(bool unloadNow);
 #endif
 
 #define INIT_PLUGIN(pclass) static pclass __plugin
+#ifdef DISABLE_UNLOAD
+#define UNLOAD_PLUGIN(arg) static bool _unused_unload(bool arg)
+#else
+#ifdef _WINDOWS
+#define UNLOAD_PLUGIN(arg) extern "C" __declspec(dllexport) bool _unload(bool arg)
+#else
+#define UNLOAD_PLUGIN(arg) extern "C" bool _unload(bool arg)
+#endif
+#endif
+
+/**
+ * Base class for engine running stage checkers.
+ * Descendants may check specific conditions and decide to stop the engine.
+ * There should be only one (static) instance of an engine checker
+ * @short Engine checker interface
+ */
+class YATE_API EngineCheck
+{
+public:
+    /**
+     * Do-nothing destructor of base class
+     */
+    virtual ~EngineCheck()
+	{ }
+
+    /**
+     * Check running conditions. This method is called by the engine in the initialize process
+     * @param cmds Optional list of strings containing extra command line parameters
+     *  (not parsed by the engine)
+     * @return False to stop the program
+     */
+    virtual bool check(const ObjList* cmds) = 0;
+
+    /**
+     * Set the current engine checker
+     * @param ptr The new engine checker. May be 0 to reset it
+     */
+    static void setChecker(EngineCheck* ptr = 0);
+};
 
 /**
  * This class holds global information about the engine.
@@ -688,6 +773,7 @@ void INIT_PLUGIN(class pclass);
 class YATE_API Engine
 {
     friend class EnginePrivate;
+    friend class EngineCommand;
 public:
     /**
      * Running modes - run the engine as Console, Client or Server.
@@ -695,8 +781,21 @@ public:
     enum RunMode {
 	Stopped = 0,
 	Console = 1,
-	Client = 2,
-	Server = 3,
+	Server = 2,
+	Client = 3,
+	ClientProxy = 4,
+    };
+
+    /**
+     * Plugin load and initialization modes.
+     * Default is LoadLate that initailizes the plugin after others.
+     * LoadEarly will move the plugin to the front of the init order.
+     * LoadFail causes the plugin to be unloaded.
+     */
+    enum PluginMode {
+	LoadFail = 0,
+	LoadLate,
+	LoadEarly
     };
 
     /**
@@ -742,7 +841,7 @@ public:
      * @return True if the engine is running in client mode
      */
     inline static bool clientMode()
-	{ return s_mode == Client; }
+	{ return (s_mode == Client) || (s_mode == ClientProxy); }
 
     /**
      * Register or unregister a plugin to the engine.
@@ -752,6 +851,20 @@ public:
      */
     static bool Register(const Plugin* plugin, bool reg = true);
 
+    /**
+     * Get the server node name, should be unique in a cluster
+     * @return Node identifier string, defaults to host name
+     */
+    inline static const String& nodeName()
+	{ return s_node; }
+
+    /**
+     * Get the application's shared directory path
+     * @return The base path for shared files and directories
+     */
+    inline static const String& sharedPath()
+	{ return s_shrpath; }
+
     /**
      * Get the filename for a specific configuration
      * @param name Name of the configuration requested
@@ -761,23 +874,23 @@ public:
     static String configFile(const char* name, bool user = false);
 
     /**
-     * Get the system configuration directory path
-     * @return The directory path for system configuration files
+     * Get the system or user configuration directory path
+     * @param user True to get the user settings path
+     * @return The directory path for system or user configuration files
      */
-    inline static String& configPath()
-	{ return s_cfgpath; }
+    static const String& configPath(bool user = false);
 
     /**
      * Get the configuration file suffix
      * @return The suffix for configuration files
      */
-    inline static String& configSuffix()
+    inline static const String& configSuffix()
 	{ return s_cfgsuffix; }
 
     /**
      * The module loading path
      */
-    inline static String& modulePath()
+    inline static const String& modulePath()
 	{ return s_modpath; }
 
     /**
@@ -787,11 +900,19 @@ public:
      */
     static void extraPath(const String& path);
 
+    /**
+     * Set the per user application data path. This method must be called
+     *  by a main program before calling @ref main() or @ref help()
+     * Path separators are not allowed. The default is taken from CFG_DIR.
+     * @param path Single relative path component to write user specific data
+     */
+    static void userPath(const String& path);
+
     /**
      * Get the module filename suffix
      * @return The suffix for module files
      */
-    inline static String& moduleSuffix()
+    inline static const String& moduleSuffix()
 	{ return s_modsuffix; }
 
     /**
@@ -815,6 +936,13 @@ public:
      */
     static unsigned int runId();
 
+    /**
+     * Get the engine parameters specific to this run.
+     * @return A reference to the list of run specific parameters
+     */
+    inline static const NamedList& runParams()
+	{ return s_params; }
+
     /**
      * Reinitialize the plugins
      */
@@ -928,6 +1056,12 @@ public:
      */
     bool loadPluginDir(const String& relPath);
 
+    /**
+     * Set the load and init mode of the currently loading @ref Plugin
+     * @param mode Load and init mode, default LoadLate
+     */
+    static void pluginMode(PluginMode mode);
+
 protected:
     /**
      * Destroys the engine and everything. You must not call it directly,
@@ -939,9 +1073,10 @@ protected:
      * Loads one plugin from a shared object file
      * @param file Name of the plugin file to load
      * @param local Attempt to keep symbols local if supported by the system
+     * @param nounload Never unload the module from memory, finalize if possible
      * @return True if success, false on failure
      */
-    bool loadPlugin(const char* file, bool local = false);
+    bool loadPlugin(const char* file, bool local = false, bool nounload = false);
 
     /**
      * Loads the plugins from the plugins directory
@@ -958,11 +1093,13 @@ private:
     ObjList m_libs;
     MessageDispatcher m_dispatcher;
     static Engine* s_self;
-    static String s_cfgpath;
+    static String s_node;
+    static String s_shrpath;
     static String s_cfgsuffix;
     static String s_modpath;
     static String s_modsuffix;
     static ObjList s_extramod;
+    static NamedList s_params;
     static int s_haltcode;
     static RunMode s_mode;
 };
