@@ -130,6 +130,7 @@ struct YATE_API FormatInfo {
 	{ }
 };
 
+class DataEndpoint;
 class CallEndpoint;
 class Driver;
 
@@ -291,7 +292,22 @@ private:
  */
 class YATE_API DataNode : public RefObject
 {
+    friend class DataEndpoint;
+
 public:
+    /**
+     * Flags associated with the DataBlocks forwarded between nodes
+     */
+    enum DataFlags {
+	DataStart   = 0x0001,
+	DataEnd     = 0x0002,
+	DataMark    = 0x0004,
+	DataSilent  = 0x0008,
+	DataMissed  = 0x0010,
+	DataError   = 0x0020,
+	DataPrivate = 0x0100
+    };
+
     /**
      * Construct a DataNode
      * @param format Description of the data format, default none
@@ -330,6 +346,21 @@ public:
     inline unsigned long timeStamp() const
 	{ return m_timestamp; }
 
+    /**
+     * Check if this data node is still valid
+     * @return True if still valid, false if node should be removed
+     */
+    virtual bool valid() const
+	{ return true; }
+
+    /**
+     * Modify node parameters
+     * @param params The list of parameters to change
+     * @return True if processed
+     */
+    virtual bool control(NamedList& params)
+	{ return false; }
+
     /**
      * Get the internal representation of an invalid or unknown timestamp
      * @return Invalid timestamp - unsigned long conversion of -1
@@ -338,6 +369,14 @@ public:
 	{ return (unsigned long)-1; }
 
 protected:
+    /**
+     * Owner attach and detach notification.
+     * This method is called with @ref DataEndpoint::commonMutex() held
+     * @param added True if a new owner was added, false if it was removed
+     */
+    virtual void attached(bool added)
+	{ }
+
     DataFormat m_format;
     unsigned long m_timestamp;
 };
@@ -379,10 +418,14 @@ public:
 
     /**
      * Consumes the data sent to it from a source
-     * @param data The raw data block to process; an empty block ends data
+     * @param data The raw data block to process
      * @param tStamp Timestamp of data - typically samples
+     * @param flags Indicator flags associated with the data block
+     * @return Number of samples actually consumed,
+     *  use invalidStamp() to indicate that all data was consumed,
+     *  return zero for consumers that become invalid
      */
-    virtual void Consume(const DataBlock& data, unsigned long tStamp) = 0;
+    virtual unsigned long Consume(const DataBlock& data, unsigned long tStamp, unsigned long flags) = 0;
 
     /**
      * Get the data source of this object if it's connected
@@ -414,7 +457,8 @@ protected:
     virtual bool synchronize(DataSource* source);
 
 private:
-    void Consume(const DataBlock& data, unsigned long tStamp, DataSource* source);
+    unsigned long Consume(const DataBlock& data, unsigned long tStamp,
+	unsigned long flags, DataSource* source);
     DataSource* m_source;
     DataSource* m_override;
     long m_regularTsDelta;
@@ -425,7 +469,7 @@ private:
 /**
  * A data source
  */
-class YATE_API DataSource : public DataNode
+class YATE_API DataSource : public DataNode, public Mutex
 {
     friend class DataTranslator;
 
@@ -435,7 +479,8 @@ public:
      * @param format Name of the data format, default "slin" (Signed Linear)
      */
     inline DataSource(const char* format = "slin")
-	: DataNode(format), m_nextStamp(invalidStamp()), m_translator(0) { }
+	: DataNode(format), Mutex(false,"DataSource"),
+	  m_nextStamp(invalidStamp()), m_translator(0) { }
 
     /**
      * Source's destruct notification - detaches all consumers
@@ -450,11 +495,20 @@ public:
     virtual void* getObject(const String& name) const;
     
     /**
-     * Forwards the data to its consumers
-     * @param data The raw data block to forward; an empty block ends data
-     * @param tStamp Timestamp of data - typically samples
+     * Check if this data source is still valid
+     * @return True if still valid, false if node should be removed
      */
-    void Forward(const DataBlock& data, unsigned long tStamp = invalidStamp());
+    virtual bool valid() const;
+
+    /**
+     * Forwards the data to its consumers
+     * @param data The raw data block to forward
+     * @param tStamp Timestamp of data - typically samples
+     * @param flags Indicator flags associated with the data block
+     * @return Number of samples actually forwarded to all consumers
+     */
+    unsigned long Forward(const DataBlock& data, unsigned long tStamp = invalidStamp(),
+	unsigned long flags = 0);
 
     /**
      * Attach a data consumer
@@ -476,13 +530,6 @@ public:
      */
     void clear();
 
-    /**
-     * Get the mutex that serializes access to this data source
-     * @return Pointer to DataSource's mutex object
-     */
-    inline Mutex* mutex()
-	{ return &m_mutex; }
-
     /**
      * Get the master translator object if this source is part of a translator
      * @return A pointer to the DataTranslator object or NULL
@@ -507,7 +554,6 @@ protected:
     unsigned long m_nextStamp;
     DataTranslator* m_translator;
     ObjList m_consumers;
-    Mutex m_mutex;
 private:
     inline void setTranslator(DataTranslator* translator)
 	{ m_translator = translator; }
@@ -516,13 +562,14 @@ private:
 
 /**
  * A data source with a thread of its own
+ * @short Data source with own thread
  */
 class YATE_API ThreadedSource : public DataSource
 {
     friend class ThreadedSourcePrivate;
 public:
     /**
-     * The destruction notification, stops the thread
+     * The destruction notification, checks that the thread is gone
      */
     virtual void destroyed();
 
@@ -551,34 +598,15 @@ public:
      */
     bool running() const;
 
-    /**
-     * Get the current status of the asynchronous deletion flag
-     */
-    inline bool asyncDelete() const
-	{ return m_asyncDelete; }
-
 protected:
     /**
      * Threaded Source constructor
      * @param format Name of the data format, default "slin" (Signed Linear)
      */
     inline ThreadedSource(const char* format = "slin")
-	: DataSource(format), m_thread(0), m_asyncDelete(false)
+	: DataSource(format), m_thread(0)
 	{ }
 
-    /**
-     * Derived classes should call this method to let the source to be
-     *  destroyed asynchronously in the data thread
-     */
-    inline void asyncDelete(bool async)
-	{ m_asyncDelete = async; }
-
-    /**
-     * Clear the worker thread pointer
-     */
-    inline void clearThread()
-	{ m_thread = 0; }
-
     /**
      * The worker method. You have to reimplement it as you need
      */
@@ -591,15 +619,14 @@ protected:
     virtual void cleanup();
 
     /**
-     * Override so destruction can be delayed after all references were lost
-     *  to let the data pumping thread end normally
-     * @return True to delete the source right away, false to defer
+     * Check if the calling thread should keep looping the worker method
+     * @param runConsumers True to keep running as long consumers are attached
+     * @return True if the calling thread should remain in the run() method
      */
-    virtual bool zeroRefsTest();
+    bool looping(bool runConsumers = false) const;
 
 private:
     ThreadedSourcePrivate* m_thread;
-    bool m_asyncDelete;
 };
 
 /**
@@ -638,6 +665,13 @@ public:
      */
     virtual void* getObject(const String& name) const;
 
+    /**
+     * Check if the data translator has a valid source
+     * @return True if still valid, false if node should be removed
+     */
+    virtual bool valid() const
+	{ return m_tsource && m_tsource->valid(); }
+
     /**
      * Get the data source of a translator object
      * @return A pointer to the DataSource object or NULL
@@ -954,6 +988,13 @@ public:
     inline DataConsumer* getCallRecord() const
 	{ return m_callRecord; }
 
+    /**
+     * Clear a data node from any slot of this object
+     * @param node Pointer to DataSource or DataConsumer to clear
+     * @return True if the node was removed from at least one slot
+     */
+    bool clearData(DataNode* node);
+
     /**
      * Adds a data consumer to the list of sniffers of the local call data
      * @param sniffer Pointer to the DataConsumer to add to sniffer list
@@ -1002,6 +1043,21 @@ public:
     inline const String& name() const
 	{ return m_name; }
 
+    /**
+     * Clear the owner call endpoint.
+     * Works only if the caller provides the correct owner pointer
+     * @param call Pointer to the call endpoint that is to be cleared
+     */
+    inline void clearCall(const CallEndpoint* call)
+	{ if (call == m_call) m_call = 0; }
+
+    /**
+     * Modify data parameters
+     * @param params The list of parameters to change
+     * @return True if processed
+     */
+    virtual bool control(NamedList& params);
+
 protected:
     /**
      * Attempt to connect the endpoint to a peer of the same type
@@ -1105,10 +1161,20 @@ public:
      * Disconnect from the connected peer call endpoint.
      * @param reason Text that describes disconnect reason.
      * @param notify Call disconnected() notification method on old peer
+     * @param params Optional pointer to extra parameters for disconnect cause
      * @return True if the object was deleted, false if it still exists
      */
-    inline bool disconnect(const char* reason = 0, bool notify = true)
-	{ return disconnect(false,reason,notify); }
+    inline bool disconnect(const char* reason = 0, bool notify = true, const NamedList* params = 0)
+	{ return disconnect(false,reason,notify,params); }
+
+    /**
+     * Disconnect from the connected peer call endpoint and notify old peer.
+     * @param reason Text that describes disconnect reason.
+     * @param params Extra parameters for disconnect cause
+     * @return True if the object was deleted, false if it still exists
+     */
+    inline bool disconnect(const char* reason, const NamedList& params)
+	{ return disconnect(false,reason,true,&params); }
 
     /**
      * Get a data endpoint of this object
@@ -1158,6 +1224,14 @@ public:
      */
     DataConsumer* getConsumer(const char* type = "audio") const;
 
+    /**
+     * Clear a data node from any slot of a DataEndpoint of this object
+     * @param node Pointer to DataSource or DataConsumer to clear
+     * @param type Type of data node: "audio", "video", "text"
+     * @return True if the node was removed from at least one slot
+     */
+    bool clearData(DataNode* node, const char* type = "audio");
+
 protected:
     /**
      * Constructor
@@ -1177,13 +1251,20 @@ protected:
      */
     virtual void disconnected(bool final, const char* reason) { }
 
+    /**
+     * Set disconnect parameters
+     * @param params Pointer to disconnect cause parameters, NULL to reset them
+     */
+    virtual void setDisconnect(const NamedList* params) { }
+
     /**
      * Set the peer call endpoint pointer.
      * @param peer A pointer to the new peer or NULL.
      * @param reason Text describing the reason in case of disconnect.
      * @param notify Call notification methods - connected() or disconnected()
+     * @param params Optional pointer to extra parameters for disconnect cause
      */
-    void setPeer(CallEndpoint* peer, const char* reason = 0, bool notify = true);
+    void setPeer(CallEndpoint* peer, const char* reason = 0, bool notify = true, const NamedList* params = 0);
 
     /**
      * Set a foreign data endpoint in this object
@@ -1198,7 +1279,7 @@ protected:
     virtual void setId(const char* newId);
 
 private:
-    bool disconnect(bool final, const char* reason, bool notify);
+    bool disconnect(bool final, const char* reason, bool notify, const NamedList* params);
 };
 
 /**
@@ -1272,6 +1353,15 @@ public:
      */
     bool filterDebug(const String& item) const;
 
+    /**
+     * Helper function to complete just one item on a command line
+     * @param itemList Tab separated list of possible values to complete
+     * @param item Item to possibly insert in the list
+     * @param partWord Partial word to complete, may be empty
+     * @return True if the item was added to list, false if it didn't match
+     */
+    static bool itemComplete(String& itemList, const String& item, const String& partWord);
+
 protected:
     /**
      * IDs of the installed relays
@@ -1298,6 +1388,10 @@ protected:
 	Progress   = 0x00010000,
 	Update     = 0x00020000,
 	Transfer   = 0x00040000,
+	Control	   = 0x00080000,
+	// Instant messenging related
+	ImRoute    = 0x00100000,
+	ImExecute  = 0x00200000,
 	// Last possible public ID
 	PubLast    = 0x0fffffff,
 	// Private messages base ID
@@ -1315,8 +1409,9 @@ protected:
      * Constructor
      * @param name Plugin name of this driver
      * @param type Type of the driver: "misc", "route", etc.
+     * @param earlyInit True to attempt to initialize module before others
      */
-    Module(const char* name, const char* type = 0);
+    Module(const char* name, const char* type = 0, bool earlyInit = false);
 
     /**
      * Destructor
@@ -1349,6 +1444,44 @@ protected:
      */
     bool installRelay(const char* name, unsigned priority = 100);
 
+    /**
+     * Install a custom message relay
+     * @param id RelayID of the new relay to create
+     * @param name Name of the custom relay to create
+     * @param priority Priority of the handler, 0 = top
+     * @return True if installed or already was one installed
+     */
+    bool installRelay(int id, const char* name, unsigned priority = 100);
+
+    /**
+     * Install a custom message relay
+     * @param relay Custom message relay
+     * @return True if installed, false if there was already one with same ID
+     */
+    bool installRelay(MessageRelay* relay);
+
+    /**
+     * Uninstall a message relay
+     * @param relay Pointer to message relay
+     * @param delRelay True to delete the relay after removing it
+     * @return True if uninstalled, false if if was not present
+     */
+    bool uninstallRelay(MessageRelay* relay, bool delRelay = true);
+
+    /**
+     * Uninstall a message relay
+     * @param id RelayID to uninstall, relay will be deleted
+     * @param delRelay True to delete the relay after removing it
+     * @return True if uninstalled, false if if was not present
+     */
+    bool uninstallRelay(int id, bool delRelay = true);
+
+    /**
+     * Uninstall all installed relays in preparation for unloading
+     * @return True if all relays were uninstalled, false if something wrong
+     */
+    bool uninstallRelays();
+
     /**
      * Message receiver handler
      * @param msg The received message
@@ -1435,7 +1568,7 @@ protected:
 private:
     Module(); // no default constructor please
     static TokenDict s_messages[];
-    bool installRelay(const char* name, int id, unsigned priority);
+    ObjList m_relayList;
 };
 
 /**
@@ -1448,11 +1581,16 @@ class YATE_API Channel : public CallEndpoint, public DebugEnabler
     friend class Router;
 
 private:
+    NamedList m_parameters;
     Driver* m_driver;
     bool m_outgoing;
     u_int64_t m_timeout;
     u_int64_t m_maxcall;
-    mutable unsigned int m_sequence;
+    u_int64_t m_dtmfTime;
+    unsigned int m_dtmfSeq;
+    String m_dtmfText;
+    String m_dtmfDetected;
+    String m_lastPeerId;
 
 protected:
     String m_status;
@@ -1587,6 +1725,13 @@ public:
      */
     virtual void msgStatus(Message& msg);
 
+    /**
+     * Control message handler that is invoked only for messages to this channel
+     * @param msg Control message
+     * @return True to stop processing the message, false to let it flow
+     */
+    virtual bool msgControl(Message& msg);
+
     /**
      * Timer check method, by default handles channel timeouts
      * @param msg Timer message
@@ -1741,6 +1886,13 @@ public:
     inline const String& billid() const
 	{ return m_billid; }
 
+    /**
+     * Get the last connected peer id
+     * @return The last connected peer id
+     */
+    const String& lastPeerId() const
+	{ return m_lastPeerId; }
+
     /**
      * Start a routing thread for this channel, dereference dynamic channels
      * @param msg Pointer to message to route, typically a "call.route", will be
@@ -1802,12 +1954,25 @@ protected:
      */
     virtual void disconnected(bool final, const char* reason);
 
+    /**
+     * Set disconnect parameters
+     * @param params Pointer to disconnect cause parameters, NULL to reset them
+     */
+    virtual void setDisconnect(const NamedList* params);
+
     /**
      * Set a new ID for this channel
      * @param newId New ID to set to this channel
      */
     virtual void setId(const char* newId);
 
+    /**
+     * Create a properly populated chan.disconnect message
+     * @param reason Channel disconnect reason if available
+     * @return A new allocated and parameter filled chan.disconnected message
+     */
+    virtual Message* getDisconnect(const char* reason);
+
     /**
      * Set the current status of the channel.
      * Note that a value of "answered" will set the answered flag
@@ -1828,6 +1993,20 @@ protected:
     inline void setOutgoing(bool outgoing = true)
 	{ m_outgoing = outgoing; }
 
+    /**
+     * Add sequence number to chan.dtmf message, check for duplicates
+     * @param msg chan.dtmf message to apply sequence number
+     * @return True if the message is a duplicate (same tone, different method)
+     */
+    bool dtmfSequence(Message& msg);
+
+    /**
+     * Add sequence number to chan.dtmf and enqueue it, delete if duplicate
+     * @param msg chan.dtmf message to sequence and enqueue
+     * @return True if the message was enqueued, false if was a duplicate
+     */
+    bool dtmfEnqueue(Message* msg);
+
     /**
      * Attempt to install an override data source to send DTMF inband.
      * Needs a tone generator module capable to override with "tone/dtmfstr/xyz"
@@ -1844,6 +2023,20 @@ protected:
      */
     bool toneDetect(const char* sniffer = 0);
 
+    /**
+     * Get the disconnect parameters list
+     * @return Reference to disconnect parameters
+     */
+    inline NamedList& parameters()
+	{ return m_parameters; }
+
+    /**
+     * Get the disconnect parameters list
+     * @return Constant teference to disconnect parameters
+     */
+    inline const NamedList& parameters() const
+	{ return m_parameters; }
+
 private:
     void init();
     Channel(); // no default constructor please
@@ -1870,6 +2063,7 @@ private:
     int m_timeout;
     int m_maxroute;
     int m_maxchans;
+    bool m_dtmfDups;
 
 public:
     /**
@@ -2007,6 +2201,22 @@ protected:
      */
     virtual void genUpdate(Message& msg);
 
+    /**
+     * Check if driver owns a client line (registered to an external server)
+     * @param line Name of the line to check
+     * @return True if this driver owns  line with the specified name
+     *
+     */
+    virtual bool hasLine(const String& line) const;
+
+    /**
+     * Routing message handler. The default implementation routes to this
+     *  driver if it owns a line named in the "account" or "line" parameter.
+     * @param msg Call routing message
+     * @return True to stop processing the message, false to try other handlers
+     */
+    virtual bool msgRoute(Message& msg);
+
     /**
      * Create an outgoing calling channel
      * @param msg Call execute message
@@ -2082,6 +2292,13 @@ protected:
     inline void maxChans(int ncalls)
 	{ m_maxchans = ncalls; }
 
+    /**
+     * Set the DTMF duplicates allowed flag
+     * @param duplicates True to allow DTMF duplicate messages
+     */
+    inline void dtmfDups(bool duplicates)
+	{ m_dtmfDups = duplicates; }
+
 private:
     Driver(); // no default constructor please
 };
